@@ -34,12 +34,17 @@
 
 #include <opm/grid/CpGrid.hpp>
 #include <opm/grid/polyhedralgrid.hh>
+#if HAVE_DUNE_ALUGRID
+#include <dune/alugrid/grid.hh>
+#include <dune/alugrid/3d/gridview.hh>
+#include "alucartesianindexmapper.hh"
+#endif // HAVE_DUNE_ALUGRID
 
 #if HAVE_DUNE_FEM
 #include <dune/fem/gridpart/adaptiveleafgridpart.hh>
 #include <dune/fem/gridpart/common/gridpart2gridview.hh>
 #include <ebos/femcpgridcompat.hh>
-#endif
+#endif // HAVE_DUNE_FEM
 
 #include <boost/date_time.hpp>
 
@@ -237,6 +242,41 @@ readRockCompactionParameters_()
     }
 }
 
+template<class GridView, class FluidSystem, class Scalar>
+Scalar EclGenericProblem<GridView,FluidSystem,Scalar>::
+rockCompressibility(unsigned globalSpaceIdx) const
+{
+    if (this->rockParams_.empty())
+        return 0.0;
+
+    unsigned tableIdx = 0;
+    if (!this->rockTableIdx_.empty()) {
+        tableIdx = this->rockTableIdx_[globalSpaceIdx];
+    }
+    return this->rockParams_[tableIdx].compressibility;
+}
+
+template<class GridView, class FluidSystem, class Scalar>
+Scalar EclGenericProblem<GridView,FluidSystem,Scalar>::
+rockReferencePressure(unsigned globalSpaceIdx) const
+{
+    if (this->rockParams_.empty())
+        return 1e5;
+
+    unsigned tableIdx = 0;
+    if (!this->rockTableIdx_.empty()) {
+        tableIdx = this->rockTableIdx_[globalSpaceIdx];
+    }
+    return this->rockParams_[tableIdx].referencePressure;
+}
+
+template<class GridView, class FluidSystem, class Scalar>
+Scalar EclGenericProblem<GridView,FluidSystem,Scalar>::
+porosity(unsigned globalSpaceIdx, unsigned timeIdx) const
+{
+    return this->referencePorosity_[timeIdx][globalSpaceIdx];
+}
+
 template<class GridView, class FluidSystem, class Scalar>
 template<class T>
 void EclGenericProblem<GridView,FluidSystem,Scalar>::
@@ -282,6 +322,15 @@ updatePlmixnum_()
     updateNum("PLMIXNUM", plmixnum_);
 }
 
+template<class GridView, class FluidSystem, class Scalar>
+void EclGenericProblem<GridView,FluidSystem,Scalar>::
+updateKrnum_()
+{
+    updateNum("KRNUMX", krnumx_);
+    updateNum("KRNUMY", krnumy_);
+    updateNum("KRNUMZ", krnumz_);
+}
+
 template<class GridView, class FluidSystem, class Scalar>
 bool EclGenericProblem<GridView,FluidSystem,Scalar>::
 vapparsActive(int episodeIdx) const
@@ -572,6 +621,21 @@ solventSaturation(unsigned elemIdx) const
     return solventSaturation_[elemIdx];
 }
 
+template<class GridView, class FluidSystem, class Scalar>
+Scalar EclGenericProblem<GridView,FluidSystem,Scalar>::
+drsdtcon(unsigned elemIdx, int episodeIdx) const
+{
+    if (convectiveDrs_.empty())
+        return 0;
+
+    // The episode index is set to -1 in the initialization phase.
+    // Output drsdt value for index 0
+    episodeIdx = std::max(episodeIdx, 0);
+    const auto& oilVaporizationControl = schedule_[episodeIdx].oilvap();
+    int pvtRegionIdx = pvtRegionIndex(elemIdx);
+    return oilVaporizationControl.getMaxDRSDT(pvtRegionIdx)*convectiveDrs_[elemIdx];
+}
+
 template<class GridView, class FluidSystem, class Scalar>
 Scalar EclGenericProblem<GridView,FluidSystem,Scalar>::
 polymerConcentration(unsigned elemIdx) const
@@ -729,12 +793,46 @@ template class EclGenericProblem<Dune::Fem::GridPart2GridViewImpl<
                                  Opm::BlackOilFluidSystem<
                                      double,
                                      Opm::BlackOilDefaultIndexTraits>,
+                                                        double>;
+#if HAVE_DUNE_ALUGRID
+
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+
+template class EclGenericProblem<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>,
+                                 BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,
                                  double>;
+template class EclGenericProblem<Dune::Fem::GridPart2GridViewImpl<
+                                     Dune::Fem::AdaptiveLeafGridPart<
+                                        ALUGrid3CN,
+                                         Dune::PartitionIteratorType(4),
+                                         false>>,
+                                 Opm::BlackOilFluidSystem<
+                                     double,
+                                     Opm::BlackOilDefaultIndexTraits>,
+                                                        double>;
+
+#endif //HAVE_DUNE_ALUGRID                              
 #else
 template class EclGenericProblem<Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>,
                                  BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,
                                  double>;
-#endif
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = const Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = const Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+
+template class EclGenericProblem<Dune::GridView<Dune::ALU3dLeafGridViewTraits<ALUGrid3CN, Dune::PartitionIteratorType(4)>>,
+                                 BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,
+                                 double>;
+
+#endif //HAVE_DUNE_ALUGRID
+#endif //HAVE_DUNE_FEM
 
 template class EclGenericProblem<Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>,
                                  BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,
