@@ -33,7 +33,6 @@
 #include <opm/input/eclipse/EclipseState/EclipseState.hpp>
 #include <opm/input/eclipse/EclipseState/Runspec.hpp>
 #include <opm/input/eclipse/EclipseState/Tables/TracerVdTable.hpp>
-#include <opm/simulators/linalg/FlexibleSolver.hpp>
 
 #include <dune/istl/operators.hh>
 #include <dune/istl/solvers.hh>
@@ -45,7 +44,13 @@
 #include <dune/fem/gridpart/adaptiveleafgridpart.hh>
 #include <dune/fem/gridpart/common/gridpart2gridview.hh>
 #include <ebos/femcpgridcompat.hh>
-#endif
+#endif // HAVE_DUNE_FEM
+
+#if HAVE_DUNE_ALUGRID
+#include <dune/alugrid/grid.hh>
+#include <dune/alugrid/3d/gridview.hh>
+#include "alucartesianindexmapper.hh"
+#endif // HAVE_DUNE_ALUGRID
 
 #include <fmt/format.h>
 #include <iostream>
@@ -63,7 +68,7 @@ struct TracerSolverSelector
 {
     using Comm = Dune::OwnerOverlapCopyCommunication<int, int>;
     using TracerOperator = Dune::OverlappingSchwarzOperator<M, V, V, Comm>;
-    using type = Dune::FlexibleSolver<M, V>;
+    using type = Dune::FlexibleSolver<TracerOperator>;
 };
 template<class Vector, class Grid, class Matrix>
 std::tuple<std::unique_ptr<Dune::OverlappingSchwarzOperator<Matrix,Vector,Vector,
@@ -83,7 +88,7 @@ createParallelFlexibleSolver(const Dune::CpGrid& grid, const Matrix& M, const Pr
 {
         using TracerOperator = Dune::OverlappingSchwarzOperator<Matrix,Vector,Vector,
                                                                 Dune::OwnerOverlapCopyCommunication<int,int>>;
-        using TracerSolver = Dune::FlexibleSolver<Matrix, Vector>;
+        using TracerSolver = Dune::FlexibleSolver<TracerOperator>;
         const auto& cellComm = grid.cellCommunication();
         auto op = std::make_unique<TracerOperator>(M, cellComm);
         auto dummyWeights = [](){ return Vector();};
@@ -209,17 +214,14 @@ doInit(bool rst, size_t numGridDof,
     tracerResidual_.resize(numGridDof);
 
     // allocate matrix for storing the Jacobian of the tracer residual
-    tracerMatrix_ = new TracerMatrix(numGridDof, numGridDof, TracerMatrix::random);
+    tracerMatrix_ = std::make_unique<TracerMatrix>(numGridDof, numGridDof, TracerMatrix::random);
 
     // find the sparsity pattern of the tracer matrix
     using NeighborSet = std::set<unsigned>;
     std::vector<NeighborSet> neighbors(numGridDof);
 
     Stencil stencil(gridView_, dofMapper_);
-    auto elemIt = gridView_.template begin<0>();
-    const auto elemEndIt = gridView_.template end<0>();
-    for (; elemIt != elemEndIt; ++elemIt) {
-        const auto& elem = *elemIt;
+    for (const auto& elem : elements(gridView_)) {
         stencil.update(elem);
 
         for (unsigned primaryDofIdx = 0; primaryDofIdx < stencil.numPrimaryDof(); ++primaryDofIdx) {
@@ -395,13 +397,49 @@ template class EclGenericTracerModel<Dune::CpGrid,
                                                                   Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false> >,
                                                       false, false>,
                                      double>;
+#if HAVE_DUNE_ALUGRID
+
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+                                    
+template class EclGenericTracerModel<ALUGrid3CN, Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>, Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>>, Opm::EcfvStencil<double,Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>,false,false>,
+                                     double>;
+
+template class EclGenericTracerModel<ALUGrid3CN,
+Dune::Fem::GridPart2GridViewImpl<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false> >,
+                                     Dune::MultipleCodimMultipleGeomTypeMapper<
+                                         Dune::Fem::GridPart2GridViewImpl<
+Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false> > >,
+                                     Opm::EcfvStencil<double, Dune::Fem::GridPart2GridViewImpl<
+                                                                  Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false> >,
+                                                      false, false>,
+                                     double>;                                     
+#endif //HAVE_DUNE_ALUGRID                                     
 #else
 template class EclGenericTracerModel<Dune::CpGrid,
                                      Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>,
                                      Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>>,
                                      Opm::EcfvStencil<double,Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>,false,false>,
                                      double>;
-#endif
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+
+template class EclGenericTracerModel<ALUGrid3CN,
+                                     Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN, Dune::PartitionIteratorType(4)>>,
+                                     Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN,
+                                         Dune::PartitionIteratorType(4)>>>,
+                                     Opm::EcfvStencil<double,Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN,
+                                        Dune::PartitionIteratorType(4)>>,false,false>,
+                                     double>;
+#endif //HAVE_DUNE_ALUGRID
+#endif //HAVE_DUNE_FEM
 
 template class EclGenericTracerModel<Dune::PolyhedralGrid<3,3,double>,
                                      Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>,
