@@ -28,6 +28,11 @@
 #include <opm/grid/cpgrid/GridHelpers.hpp>
 #include <opm/grid/polyhedralgrid.hh>
 #include <opm/grid/utility/cartesianToCompressed.hpp>
+#if HAVE_DUNE_ALUGRID
+#include "eclalugridvanguard.hh"
+#include <dune/alugrid/grid.hh>
+#include <dune/alugrid/3d/gridview.hh>
+#endif // HAVE_DUNE_ALUGRID
 
 #include <opm/output/eclipse/EclipseIO.hpp>
 #include <opm/output/eclipse/RestartValue.hpp>
@@ -43,11 +48,15 @@
 
 #include <dune/grid/common/mcmgmapper.hh>
 
+#if HAVE_MPI
+#include <ebos/eclmpiserializer.hh>
+#endif
+
 #if HAVE_DUNE_FEM
 #include <dune/fem/gridpart/adaptiveleafgridpart.hh>
 #include <dune/fem/gridpart/common/gridpart2gridview.hh>
 #include <ebos/femcpgridcompat.hh>
-#endif
+#endif // HAVE_DUNE_FEM
 
 #if HAVE_MPI
 #include <mpi.h>
@@ -242,21 +251,20 @@ eclIO() const
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
 void EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
-writeInit()
+writeInit(const std::function<unsigned int(unsigned int)>& map)
 {
     if (collectToIORank_.isIORank()) {
         std::map<std::string, std::vector<int> > integerVectors;
         if (collectToIORank_.isParallel())
             integerVectors.emplace("MPI_RANK", collectToIORank_.globalRanks());
-        auto cartMap = cartesianToCompressed(equilGrid_->size(0),
-                                             UgGridHelpers::globalCell(*equilGrid_));
-        eclIO_->writeInitial(computeTrans_(cartMap), integerVectors, exportNncStructure_(cartMap));
+        auto cartMap = cartesianToCompressed(equilGrid_->size(0), UgGridHelpers::globalCell(*equilGrid_));
+        eclIO_->writeInitial(computeTrans_(cartMap, map), integerVectors, exportNncStructure_(cartMap, map));
     }
 }
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
 data::Solution EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
-computeTrans_(const std::unordered_map<int,int>& cartesianToActive) const
+computeTrans_(const std::unordered_map<int,int>& cartesianToActive, const std::function<unsigned int(unsigned int)>& map) const
 {
     const auto& cartMapper = *equilCartMapper_;
     const auto& cartDims = cartMapper.cartesianDimensions();
@@ -277,16 +285,8 @@ computeTrans_(const std::unordered_map<int,int>& cartesianToActive) const
     using GlobElementMapper = Dune::MultipleCodimMultipleGeomTypeMapper<GlobalGridView>;
     GlobElementMapper globalElemMapper(globalGridView, Dune::mcmgElementLayout());
 
-    auto elemIt = globalGridView.template begin</*codim=*/0>();
-    const auto& elemEndIt = globalGridView.template end</*codim=*/0>();
-    for (; elemIt != elemEndIt; ++ elemIt) {
-        const auto& elem = *elemIt;
-
-        auto isIt = globalGridView.ibegin(elem);
-        const auto& isEndIt = globalGridView.iend(elem);
-        for (; isIt != isEndIt; ++ isIt) {
-            const auto& is = *isIt;
-
+    for (const auto& elem : elements(globalGridView)) {
+        for (const auto& is : intersections(globalGridView, elem)) {
             if (!is.neighbor())
                 continue; // intersection is on the domain boundary
 
@@ -304,6 +304,12 @@ computeTrans_(const std::unordered_map<int,int>& cartesianToActive) const
             int gc1 = std::min(cartIdx1, cartIdx2);
             int gc2 = std::max(cartIdx1, cartIdx2);
 
+            // Re-ordering in case of non-empty mapping between equilGrid to grid
+            if (map) {
+                c1 = map(c1); // equilGridToGrid map
+                c2 = map(c2);
+            }
+
             if (gc2 - gc1 == 1 && cartDims[0] > 1 ) {
                 tranx.data[gc1] = globalTrans().transmissibility(c1, c2);
                 continue; // skip other if clauses as they are false, last one needs some computation
@@ -327,7 +333,7 @@ computeTrans_(const std::unordered_map<int,int>& cartesianToActive) const
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
 std::vector<NNCdata> EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
-exportNncStructure_(const std::unordered_map<int,int>& cartesianToActive) const
+exportNncStructure_(const std::unordered_map<int,int>& cartesianToActive, const std::function<unsigned int(unsigned int)>& map) const
 {
     std::size_t nx = eclState_.getInputGrid().getNX();
     std::size_t ny = eclState_.getInputGrid().getNY();
@@ -359,16 +365,8 @@ exportNncStructure_(const std::unordered_map<int,int>& cartesianToActive) const
     // Cartesian index mapper for the serial I/O grid
     const auto& equilCartMapper =  *equilCartMapper_;
     const auto& cartDims = cartMapper_.cartesianDimensions();
-    auto elemIt = globalGridView.template begin</*codim=*/0>();
-    const auto& elemEndIt = globalGridView.template end</*codim=*/0>();
-    for (; elemIt != elemEndIt; ++ elemIt) {
-        const auto& elem = *elemIt;
-
-        auto isIt = globalGridView.ibegin(elem);
-        const auto& isEndIt = globalGridView.iend(elem);
-        for (; isIt != isEndIt; ++ isIt) {
-            const auto& is = *isIt;
-
+    for (const auto& elem : elements(globalGridView)) {
+        for (const auto& is : intersections(globalGridView, elem)) {
             if (!is.neighbor())
                 continue; // intersection is on the domain boundary
 
@@ -378,14 +376,20 @@ exportNncStructure_(const std::unordered_map<int,int>& cartesianToActive) const
             if (c1 > c2)
                 continue; // we only need to handle each connection once, thank you.
 
-            std::size_t cc1 = equilCartMapper.cartesianIndex( c1 ); //globalIOGrid_.globalCell()[c1];
-            std::size_t cc2 = equilCartMapper.cartesianIndex( c2 ); //globalIOGrid_.globalCell()[c2];
+            std::size_t cc1 = equilCartMapper.cartesianIndex( c1 );
+            std::size_t cc2 = equilCartMapper.cartesianIndex( c2 );
 
             if ( cc2 < cc1 )
                 std::swap(cc1, cc2);
 
             auto cellDiff = cc2 - cc1;
 
+            // Re-ordering in case of non-empty mapping between equilGrid to grid
+            if (map) {
+                c1 = map(c1); // equilGridToGrid map
+                c2 = map(c2);
+            }
+
             if (cellDiff != 1 &&
                 cellDiff != nx &&
                 cellDiff != nx*ny &&
@@ -430,9 +434,10 @@ doWriteOutput(const int                     reportStepNum,
               bool doublePrecision)
 {
     const auto isParallel = this->collectToIORank_.isParallel();
+    const bool needsReordering = this->collectToIORank_.doesNeedReordering();
 
     RestartValue restartValue {
-        isParallel ? this->collectToIORank_.globalCellData()
+        (isParallel || needsReordering) ? this->collectToIORank_.globalCellData()
                    : std::move(localCellData),
 
         isParallel ? this->collectToIORank_.globalWellData()
@@ -489,7 +494,6 @@ evalSummary(const int reportStepNum,
             SummaryState& summaryState,
             UDQState& udqState)
 {
-    std::vector<char> buffer;
     if (collectToIORank_.isIORank()) {
         const auto& summary = eclIO_->summary();
         auto wbp_calculators = summary.wbp_calculators(reportStepNum);
@@ -530,23 +534,13 @@ evalSummary(const int reportStepNum,
         auto udq_step = reportStepNum - 1;
         const auto& udq_config = schedule_.getUDQConfig(udq_step);
         udq_config.eval( udq_step, schedule_.wellMatcher(udq_step), summaryState, udqState);
-
-        buffer = summaryState.serialize();
     }
+#if HAVE_MPI
     if (collectToIORank_.isParallel()) {
-#ifdef HAVE_MPI
-        unsigned long buffer_size = buffer.size();
-        grid_.comm().broadcast(&buffer_size, 1, collectToIORank_.ioRank);
-        if (!collectToIORank_.isIORank())
-            buffer.resize( buffer_size );
-
-        grid_.comm().broadcast(buffer.data(), buffer_size, collectToIORank_.ioRank);
-        if (!collectToIORank_.isIORank()) {
-            SummaryState& st = summaryState;
-            st.deserialize(buffer);
-        }
-#endif
+        EclMpiSerializer ser(grid_.comm());
+        ser.append(summaryState);
     }
+#endif
 }
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
@@ -561,8 +555,7 @@ globalTrans() const
 #if HAVE_DUNE_FEM
 template class EclGenericWriter<Dune::CpGrid,
                                 Dune::CpGrid,
-                                Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>,
-                                Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>>,
+                                Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>, Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>>,
                                 double>;
 template class EclGenericWriter<Dune::CpGrid,
                                 Dune::CpGrid,
@@ -578,18 +571,59 @@ template class EclGenericWriter<Dune::CpGrid,
                                             Dune::PartitionIteratorType(4),
                                             false>>>,
                                 double>;
+
+
+#ifdef HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
 #else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+                                                               
+template class EclGenericWriter<ALUGrid3CN,
+                                Dune::CpGrid,
+                                Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>, Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>>,
+                                double>;
+                                
+template class EclGenericWriter<ALUGrid3CN,
+                                Dune::CpGrid,
+                                Dune::Fem::GridPart2GridViewImpl<
+                                    Dune::Fem::AdaptiveLeafGridPart<
+                                        ALUGrid3CN,
+                                        Dune::PartitionIteratorType(4),
+                                        false>>,
+                                Dune::MultipleCodimMultipleGeomTypeMapper<
+                                    Dune::Fem::GridPart2GridViewImpl<
+                                        Dune::Fem::AdaptiveLeafGridPart<
+                                            ALUGrid3CN,
+                                            Dune::PartitionIteratorType(4),
+                                            false>>>,
+                                double>;                                
+#endif // HAVE_DUNE_ALUGRID
+
+#else // !HAVE_DUNE_FEM
 template class EclGenericWriter<Dune::CpGrid,
                                 Dune::CpGrid,
                                 Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>,
                                 Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>>,
                                 double>;
-#endif
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+template class EclGenericWriter<ALUGrid3CN,
+                                Dune::CpGrid,
+                                Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN,Dune::PartitionIteratorType(4)>>,  Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN, Dune::PartitionIteratorType(4)>>>,
+                                double>;
+
+#endif // HAVE_DUNE_ALUGRID
+#endif // !HAVE_DUNE_FEM
 
 template class EclGenericWriter<Dune::PolyhedralGrid<3,3,double>,
                                 Dune::PolyhedralGrid<3,3,double>,
-                                Dune::GridView<Dune::PolyhedralGridViewTraits<3, 3, double, Dune::PartitionIteratorType(4)>>,
-                                Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>>,
+                                Dune::GridView<Dune::PolyhedralGridViewTraits<3, 3, double, Dune::PartitionIteratorType(4)>>,                              Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>>,
                                 double>;
-
+                                                                  
 } // namespace Opm
