@@ -30,6 +30,12 @@
 #include <opm/grid/CpGrid.hpp>
 #include <opm/grid/polyhedralgrid.hh>
 
+#if HAVE_DUNE_ALUGRID
+#include <dune/alugrid/grid.hh>
+#include <dune/alugrid/3d/gridview.hh>
+#include "alucartesianindexmapper.hh"
+#endif // HAVE_DUNE_ALUGRID
+
 #include <opm/input/eclipse/EclipseState/EclipseState.hpp>
 #include <opm/input/eclipse/EclipseState/Grid/FaceDir.hpp>
 #include <opm/input/eclipse/EclipseState/Grid/FieldPropsManager.hpp>
@@ -82,11 +88,11 @@ std::uint64_t directionalIsId(std::uint32_t elemIdx1, std::uint32_t elemIdx2)
 
 namespace Opm {
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 EclTransmissibility(const EclipseState& eclState,
                     const GridView& gridView,
-                    const Dune::CartesianIndexMapper<Grid>& cartMapper,
+                    const CartesianIndexMapper& cartMapper,
                     const Grid& grid,
                     std::function<std::array<double,dimWorld>(int)> centroids,
                     bool enableEnergy,
@@ -103,36 +109,36 @@ EclTransmissibility(const EclipseState& eclState,
     transmissibilityThreshold_  = unitSystem.parse("Transmissibility").getSIScaling() * 1e-6;
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-Scalar EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+Scalar EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 transmissibility(unsigned elemIdx1, unsigned elemIdx2) const
 {
     return trans_.at(isId(elemIdx1, elemIdx2));
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-Scalar EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+Scalar EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 transmissibilityBoundary(unsigned elemIdx, unsigned boundaryFaceIdx) const
 {
     return transBoundary_.at(std::make_pair(elemIdx, boundaryFaceIdx));
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-Scalar EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+Scalar EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 thermalHalfTrans(unsigned insideElemIdx, unsigned outsideElemIdx) const
 {
     return thermalHalfTrans_.at(directionalIsId(insideElemIdx, outsideElemIdx));
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-Scalar EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+Scalar EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 thermalHalfTransBoundary(unsigned insideElemIdx, unsigned boundaryFaceIdx) const
 {
     return thermalHalfTransBoundary_.at(std::make_pair(insideElemIdx, boundaryFaceIdx));
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-Scalar EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+Scalar EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 diffusivity(unsigned elemIdx1, unsigned elemIdx2) const
 {
     if (diffusivity_.empty())
@@ -142,9 +148,9 @@ diffusivity(unsigned elemIdx1, unsigned elemIdx2) const
 
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
-update(bool global)
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
+update(bool global, const std::function<unsigned int(unsigned int)>& map)
 {
     const auto& cartDims = cartMapper_.cartesianDimensions();
     auto& transMult = eclState_.getTransMult();
@@ -155,8 +161,11 @@ update(bool global)
     const std::vector<double>& ntg = eclState_.fieldProps().get_double("NTG");
     const bool updateDiffusivity = eclState_.getSimulationConfig().isDiffusive() && enableDiffusivity_;
     unsigned numElements = elemMapper.size();
-
-    extractPermeability_();
+    
+    if (map)
+        extractPermeability_(map);
+    else
+        extractPermeability_();
 
     // calculate the axis specific centroids of all elements
     std::array<std::vector<DimVector>, dimWorld> axisCentroids;
@@ -164,11 +173,7 @@ update(bool global)
     for (unsigned dimIdx = 0; dimIdx < dimWorld; ++dimIdx)
         axisCentroids[dimIdx].resize(numElements);
 
-    auto elemIt = gridView_.template begin</*codim=*/ 0>();
-    const auto& elemEndIt = gridView_.template end</*codim=*/ 0>();
-    size_t centroidIdx = 0;
-    for (; elemIt != elemEndIt; ++elemIt, ++centroidIdx) {
-        const auto& elem = *elemIt;
+    for (const auto& elem : elements(gridView_)) {
         unsigned elemIdx = elemMapper.index(elem);
 
         // compute the axis specific "centroids" used for the transmissibilities. for
@@ -218,9 +223,7 @@ update(bool global)
     }
 
     // compute the transmissibilities for all intersections
-    elemIt = gridView_.template begin</*codim=*/ 0>();
-    for (; elemIt != elemEndIt; ++elemIt) {
-        const auto& elem = *elemIt;
+    for (const auto& elem : elements(gridView_)) {
         unsigned elemIdx = elemMapper.index(elem);
 
         auto isIt = gridView_.ibegin(elem);
@@ -467,10 +470,8 @@ update(bool global)
     std::unordered_map<std::size_t,int> globalToLocal;
 
     // loop over all elements (global grid) and store Cartesian index
-    elemIt = grid_.leafGridView().template begin<0>();
-
-    for (; elemIt != elemEndIt; ++elemIt) {
-        int elemIdx = elemMapper.index(*elemIt);
+    for (const auto& elem : elements(grid_.leafGridView())) {
+        int elemIdx = elemMapper.index(elem);
         int cartElemIdx = cartMapper_.cartesianIndex(elemIdx);
         globalToLocal[cartElemIdx] = elemIdx;
     }
@@ -481,8 +482,8 @@ update(bool global)
     removeSmallNonCartesianTransmissibilities_();
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 extractPermeability_()
 {
     unsigned numElem = gridView_.size(/*codim=*/0);
@@ -523,8 +524,51 @@ extractPermeability_()
                                "(The PERM{X,Y,Z} keywords are missing)");
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
+extractPermeability_(const std::function<unsigned int(unsigned int)>& map)
+{
+    unsigned numElem = gridView_.size(/*codim=*/0);
+    permeability_.resize(numElem);
+
+    // read the intrinsic permeabilities from the eclState. Note that all arrays
+    // provided by eclState are one-per-cell of "uncompressed" grid, whereas the
+    // simulation grid might remove a few elements. (e.g. because it is distributed
+    // over several processes.)
+    const auto& fp = eclState_.fieldProps();
+    if (fp.has_double("PERMX")) {
+        const std::vector<double>& permxData = fp.get_double("PERMX");
+
+        std::vector<double> permyData;
+        if (fp.has_double("PERMY"))
+            permyData = fp.get_double("PERMY");
+        else
+            permyData = permxData;
+
+        std::vector<double> permzData;
+        if (fp.has_double("PERMZ"))
+            permzData = fp.get_double("PERMZ");
+        else
+            permzData = permxData;
+
+        for (size_t dofIdx = 0; dofIdx < numElem; ++ dofIdx) {
+            permeability_[dofIdx] = 0.0;
+            size_t inputDofIdx = map(dofIdx);
+            permeability_[dofIdx][0][0] = permxData[inputDofIdx];
+            permeability_[dofIdx][1][1] = permyData[inputDofIdx];
+            permeability_[dofIdx][2][2] = permzData[inputDofIdx];
+        }
+
+        // for now we don't care about non-diagonal entries
+
+    }
+    else
+        throw std::logic_error("Can't read the intrinsic permeability from the ecl state. "
+                               "(The PERM{X,Y,Z} keywords are missing)");
+}
+
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 extractPorosity_()
 {
     // read the intrinsic porosity from the eclState. Note that all arrays
@@ -540,8 +584,8 @@ extractPorosity_()
                                "(The PORO keywords are missing)");
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 removeSmallNonCartesianTransmissibilities_()
 {
     const auto& cartDims = cartMapper_.cartesianDimensions();
@@ -562,8 +606,8 @@ removeSmallNonCartesianTransmissibilities_()
     }
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper, Scalar>::
 applyAllZMultipliers_(Scalar& trans,
                       unsigned insideFaceIdx,
                       unsigned outsideFaceIdx,
@@ -601,8 +645,8 @@ applyAllZMultipliers_(Scalar& trans,
     }
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 updateFromEclState_(bool global)
 {
     const FieldPropsManager* fp =
@@ -633,9 +677,9 @@ updateFromEclState_(bool global)
     resetTransmissibilityFromArrays_(is_tran, trans);
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
 std::array<std::vector<double>,3>
-EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 createTransmissibilityArrays_(const std::array<bool,3>& is_tran)
 {
     const auto& cartDims = cartMapper_.cartesianDimensions();
@@ -648,16 +692,9 @@ createTransmissibilityArrays_(const std::array<bool,3>& is_tran)
           std::vector<double>(is_tran[2] ? numElem : 0, 0)};
 
     // compute the transmissibilities for all intersections
-    auto elemIt = gridView_.template begin</*codim=*/ 0>();
-    const auto& elemEndIt = gridView_.template end</*codim=*/ 0>();
-
-    for (; elemIt != elemEndIt; ++elemIt) {
-        const auto& elem = *elemIt;
-        auto isIt = gridView_.ibegin(elem);
-        const auto& isEndIt = gridView_.iend(elem);
-        for (; isIt != isEndIt; ++ isIt) {
+    for (const auto& elem : elements(gridView_)) {
+        for (const auto& intersection : intersections(gridView_, elem)) {
             // store intersection, this might be costly
-            const auto& intersection = *isIt;
             if (!intersection.neighbor())
                 continue; // intersection is on the domain boundary
 
@@ -700,8 +737,8 @@ createTransmissibilityArrays_(const std::array<bool,3>& is_tran)
     return trans;
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 resetTransmissibilityFromArrays_(const std::array<bool,3>& is_tran,
                                  const std::array<std::vector<double>,3>& trans)
 {
@@ -709,16 +746,8 @@ resetTransmissibilityFromArrays_(const std::array<bool,3>& is_tran,
     ElementMapper elemMapper(gridView_, Dune::mcmgElementLayout());
 
     // compute the transmissibilities for all intersections
-    auto elemIt = gridView_.template begin</*codim=*/ 0>();
-    const auto& elemEndIt = gridView_.template end</*codim=*/ 0>();
-
-    for (; elemIt != elemEndIt; ++elemIt) {
-        const auto& elem = *elemIt;
-        auto isIt = gridView_.ibegin(elem);
-        const auto& isEndIt = gridView_.iend(elem);
-        for (; isIt != isEndIt; ++ isIt) {
-            // store intersection, this might be costly
-            const auto& intersection = *isIt;
+    for (const auto& elem : elements(gridView_)) {
+        for (const auto& intersection : intersections(gridView_, elem)) {
             if (!intersection.neighbor())
                 continue; // intersection is on the domain boundary
 
@@ -758,9 +787,9 @@ resetTransmissibilityFromArrays_(const std::array<bool,3>& is_tran,
     }
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
 template<class Intersection>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 computeFaceProperties(const Intersection& intersection,
                       const int,
                       const int,
@@ -781,9 +810,9 @@ computeFaceProperties(const Intersection& intersection,
 }
 
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
 template<class Intersection>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 computeFaceProperties(const Intersection& intersection,
                       const int insideElemIdx,
                       const int insideFaceIdx,
@@ -800,9 +829,9 @@ computeFaceProperties(const Intersection& intersection,
     faceAreaNormal = grid_.faceAreaNormalEcl(faceIdx);
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
 std::tuple<std::vector<NNCdata>, std::vector<NNCdata>>
-EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 applyNncToGridTrans_(const std::unordered_map<std::size_t,int>& cartesianToCompressed)
 {
     // First scale NNCs with EDITNNC.
@@ -853,8 +882,8 @@ applyNncToGridTrans_(const std::unordered_map<std::size_t,int>& cartesianToCompr
     return std::make_tuple(processedNnc, unprocessedNnc);
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 applyEditNncToGridTrans_(const std::unordered_map<std::size_t,int>& globalToLocal)
 {
     const auto& nnc_input = eclState_.getInputNNC();
@@ -926,8 +955,8 @@ applyEditNncToGridTrans_(const std::unordered_map<std::size_t,int>& globalToLoca
     }
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 computeHalfTrans_(Scalar& halfTrans,
                   const DimVector& areaNormal,
                   int faceIdx, // in the reference element that contains the intersection
@@ -947,8 +976,8 @@ computeHalfTrans_(Scalar& halfTrans,
     halfTrans /= distance.two_norm2();
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 computeHalfDiffusivity_(Scalar& halfDiff,
                         const DimVector& areaNormal,
                         const DimVector& distance,
@@ -963,9 +992,9 @@ computeHalfDiffusivity_(Scalar& halfDiff,
     halfDiff /= distance.two_norm2();
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-typename EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::DimVector
-EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+typename EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::DimVector
+EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 distanceVector_(const DimVector& center,
                 int faceIdx, // in the reference element that contains the intersection
                 unsigned elemIdx,
@@ -980,8 +1009,8 @@ distanceVector_(const DimVector& center,
     return x;
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 applyMultipliers_(Scalar& trans,
                   unsigned faceIdx,
                   unsigned cartElemIdx,
@@ -1014,8 +1043,8 @@ applyMultipliers_(Scalar& trans,
     }
 }
 
-template<class Grid, class GridView, class ElementMapper, class Scalar>
-void EclTransmissibility<Grid,GridView,ElementMapper,Scalar>::
+template<class Grid, class GridView, class ElementMapper, class CartesianIndexMapper, class Scalar>
+void EclTransmissibility<Grid,GridView,ElementMapper,CartesianIndexMapper,Scalar>::
 applyNtg_(Scalar& trans,
           unsigned faceIdx,
           unsigned elemIdx,
@@ -1047,6 +1076,7 @@ applyNtg_(Scalar& trans,
 template class EclTransmissibility<Dune::CpGrid,
                                    Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>,
                                    Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<Dune::CpGrid, Dune::PartitionIteratorType(4), false>>>>,
+                                   Dune::CartesianIndexMapper<Dune::CpGrid>,
                                    double>;
 template class EclTransmissibility<Dune::CpGrid,
                                    Dune::Fem::GridPart2GridViewImpl<
@@ -1060,17 +1090,65 @@ template class EclTransmissibility<Dune::CpGrid,
                                                Dune::CpGrid,
                                                Dune::PartitionIteratorType(4),
                                                false> > >,
+                                   Dune::CartesianIndexMapper<Dune::CpGrid>,             
                                    double>;
-#else
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else    
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI 
+                                                                     
+template class EclTransmissibility<ALUGrid3CN, Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>,
+                                   Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::Fem::GridPart2GridViewTraits<Dune::Fem::AdaptiveLeafGridPart<ALUGrid3CN, Dune::PartitionIteratorType(4), false>>>>,
+                                   Dune::CartesianIndexMapper<ALUGrid3CN>,
+                                   double>;
+template class EclTransmissibility<ALUGrid3CN,
+                                  Dune::Fem::GridPart2GridViewImpl<
+                                       Dune::Fem::AdaptiveLeafGridPart<
+                                           ALUGrid3CN,
+                                           Dune::PartitionIteratorType(4),
+                                           false> >,
+                                   Dune::MultipleCodimMultipleGeomTypeMapper<
+                                       Dune::Fem::GridPart2GridViewImpl<
+                                           Dune::Fem::AdaptiveLeafGridPart<
+                                               ALUGrid3CN,
+                                               Dune::PartitionIteratorType(4),
+                                               false> > >,
+                                   Dune::CartesianIndexMapper<ALUGrid3CN>,
+                                  double>;                                   
+#endif //HAVE_DUNE_ALUGRID
+
+#else // !DUNE_FEM
 template class EclTransmissibility<Dune::CpGrid,
                                    Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>,
                                    Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::DefaultLeafGridViewTraits<Dune::CpGrid>>>,
+                                   Dune::CartesianIndexMapper<Dune::CpGrid>,
                                    double>;
-#endif
+
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>;
+#else
+    using ALUGrid3CN = Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>;
+#endif //HAVE_MPI
+
+template class
+EclTransmissibility<ALUGrid3CN,
+                    Dune::GridView<
+                        Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN, Dune::PartitionIteratorType(4)>
+                    >,
+                    Dune::MultipleCodimMultipleGeomTypeMapper<
+                        Dune::GridView<Dune::ALU3dLeafGridViewTraits<const ALUGrid3CN, Dune::PartitionIteratorType(4)>>
+                    >,
+                    Dune::CartesianIndexMapper<ALUGrid3CN>,
+                    double>;
+#endif //HAVE_DUNE_ALUGRID
+#endif //HAVE_DUNE_FEM
 
 template class EclTransmissibility<Dune::PolyhedralGrid<3,3,double>,
-                                   Dune::GridView<Dune::PolyhedralGridViewTraits<3, 3, double, Dune::PartitionIteratorType(4)>>,
-                                   Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>>,
+Dune::GridView<Dune::PolyhedralGridViewTraits<3, 3, double, Dune::PartitionIteratorType(4)>>, Dune::MultipleCodimMultipleGeomTypeMapper<Dune::GridView<Dune::PolyhedralGridViewTraits<3,3,double,Dune::PartitionIteratorType(4)>>>,
+Dune::CartesianIndexMapper<Dune::PolyhedralGrid<3,3,double>>,
                                    double>;
 
 } // namespace Opm
