@@ -21,16 +21,6 @@
 
 namespace Opm {
 
-class Deck;
-class EclipseState;
-class Schedule;
-class SummaryConfig;
-
-void flowEbosBrineSaltPrecipitationSetDeck(double setupTime, std::shared_ptr<Deck> deck,
-                          std::shared_ptr<EclipseState> eclState,
-                          std::shared_ptr<Schedule> schedule,
-                          std::shared_ptr<SummaryConfig> summaryConfig);
-
 //! \brief Main function used in flow binary.
 int flowEbosBrineSaltPrecipitationMain(int argc, char** argv, bool outputCout, bool outputFiles);
 
