@@ -22,27 +22,14 @@
 #ifndef FLOW_ONEPHASE_ENERGY_HPP
 #define FLOW_ONEPHASE_ENERGY_HPP
 
-#include <memory>
-
 namespace Opm {
 
-class Deck;
-class EclipseState;
-template<class TypeTag> class FlowMainEbos;
-class Schedule;
-class SummaryConfig;
-class UDQState;
-class WellTestState;
-namespace Action {
-class State;
-}
-
-void flowEbosWaterOnlyEnergySetDeck(double setupTime, std::shared_ptr<Deck> deck,
-                                    std::shared_ptr<EclipseState> eclState,
-                                    std::shared_ptr<Schedule> schedule,
-                                    std::shared_ptr<SummaryConfig> summaryConfig);
+//! \brief Main functon used in main flow binary.
 int flowEbosWaterOnlyEnergyMain(int argc, char** argv, bool outputCout, bool outputFiles);
+
+//! \brief Main function used in flow_onephase_energy binary.
 int flowEbosWaterOnlyEnergyMainStandalone(int argc, char** argv);
+
 }
 
 #endif
