@@ -17,6 +17,10 @@
   along with OPM.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#if HAVE_CONFIG_H
+#include "config.h"
+#endif // HAVE_CONFIG_H
+
 #include <opm/core/props/satfunc/RelpermDiagnostics.hpp>
 #include <opm/core/props/phaseUsageFromDeck.hpp>
 
@@ -38,7 +42,10 @@
 
 #include <opm/grid/CpGrid.hpp>
 #include <opm/grid/polyhedralgrid.hh>
-
+#ifdef HAVE_DUNE_ALUGRID
+#include <dune/alugrid/grid.hh>
+#include <dune/alugrid/3d/gridview.hh>
+#endif // HAVE_DUNE_ALUGRID
 namespace Opm{
 
     bool RelpermDiagnostics::phaseCheck_(const EclipseState& es)
@@ -112,13 +119,16 @@ namespace Opm{
         const TableContainer& sof2Tables = tableManager.getSof2Tables();
         const TableContainer& sgwfnTables= tableManager.getSgwfnTables();
 
+        const SwofletTable& swofletTable = tableManager.getSwofletTable();
+        const SgofletTable& sgofletTable = tableManager.getSgofletTable();
+
         // Family I test.
         bool family1 = pu.phase_used[BlackoilPhases::Liquid];
         if (pu.phase_used[BlackoilPhases::Aqua]) {
-            family1 = family1 && !swofTables.empty();
+            family1 = family1 && (!swofTables.empty() || !swofletTable.empty());
         }
         if (pu.phase_used[BlackoilPhases::Vapour]) {
-            family1 = family1 && (!sgofTables.empty() || !slgofTables.empty());
+            family1 = family1 && ((!sgofTables.empty() || !sgofletTable.empty()) || !slgofTables.empty());
         }
 
         // Family II test.
@@ -664,7 +674,9 @@ namespace Opm{
             satfunc::getRawFunctionValues(tables, phases, rtep);
 
         const TableContainer&  swofTables = tables.getSwofTables();
+        const SwofletTable&  swofletTables = tables.getSwofletTable();
         const TableContainer&  sgofTables = tables.getSgofTables();
+        const SgofletTable&  sgofletTables = tables.getSgofletTable();
         const TableContainer& slgofTables = tables.getSlgofTables();
         const TableContainer&  sof3Tables = tables.getSof3Tables();
 
@@ -693,14 +705,19 @@ namespace Opm{
                      if (!sgofTables.empty()) {
                          const auto& table = sgofTables.getTable<SgofTable>(satnumIdx);
                          krog_value = table.evaluate( "KROG" , unscaledEpsInfo_[satnumIdx].Sgl );
+                     } else if (!sgofletTables.empty()) {
+                         krog_value = sgofletTables[satnumIdx].krt2_relperm;
                      } else {
                          assert(!slgofTables.empty());
                          const auto& table = slgofTables.getTable<SlgofTable>(satnumIdx);
                          krog_value = table.evaluate( "KROG" , unscaledEpsInfo_[satnumIdx].Sgl );
                      }
-                     {
+                     if (!swofTables.empty()) {
                          const auto& table = swofTables.getTable<SwofTable>(satnumIdx);
                          krow_value = table.evaluate("KROW" , unscaledEpsInfo_[satnumIdx].Swl);
+                     } else {
+                         assert(!swofletTables.empty());
+                         krow_value = swofletTables[satnumIdx].krt2_relperm;
                      }
                  }
                  if (satFamily_ == SaturationFunctionFamily::FamilyII) {
@@ -799,4 +816,11 @@ namespace Opm{
 
     INSTANCE_DIAGNOSIS(Dune::CpGrid)
     INSTANCE_DIAGNOSIS(Dune::PolyhedralGrid<3,3>)
+#if HAVE_DUNE_ALUGRID
+#if HAVE_MPI
+    INSTANCE_DIAGNOSIS(Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridMPIComm>)
+#else    
+    INSTANCE_DIAGNOSIS(Dune::ALUGrid<3, 3, Dune::cube, Dune::nonconforming, Dune::ALUGridNoComm>)
+#endif //HAVE_MPI
+#endif //HAVE_DUNE_ALUGRID
 } //namespace Opm
