@@ -29,6 +29,7 @@
 #include <opm/simulators/utils/ParallelFileMerger.hpp>
 #include <opm/simulators/utils/moduleVersion.hpp>
 #include <opm/simulators/utils/ParallelEclipseState.hpp>
+#include <flow/flow_ebos_blackoil.hpp>
 
 #include <opm/input/eclipse/EclipseState/EclipseState.hpp>
 #include <opm/input/eclipse/EclipseState/IOConfig/IOConfig.hpp>
@@ -160,7 +161,6 @@ namespace Opm
 
             // flow also does not use the eWoms Newton method
             EWOMS_HIDE_PARAM(TypeTag, NewtonMaxError);
-            EWOMS_HIDE_PARAM(TypeTag, NewtonMaxIterations);
             EWOMS_HIDE_PARAM(TypeTag, NewtonTolerance);
             EWOMS_HIDE_PARAM(TypeTag, NewtonTargetIterations);
             EWOMS_HIDE_PARAM(TypeTag, NewtonVerbose);
@@ -192,7 +192,7 @@ namespace Opm
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteSaturations);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteTemperature);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteViscosities);
-                EWOMS_HIDE_PARAM(TypeTag, VtkWriteWaterFormationVolumeFactor);                
+                EWOMS_HIDE_PARAM(TypeTag, VtkWriteWaterFormationVolumeFactor);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteGasDissolutionFactor);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteGasFormationVolumeFactor);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteGasSaturationPressure);
@@ -204,7 +204,7 @@ namespace Opm
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteDofIndex);
                 EWOMS_HIDE_PARAM(TypeTag, VtkWriteMobilities);
                 //}
-            EWOMS_HIDE_PARAM(TypeTag, VtkWriteAverageMolarMasses);           
+            EWOMS_HIDE_PARAM(TypeTag, VtkWriteAverageMolarMasses);
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteFugacities);
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteFugacityCoeffs);
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteMassFractions);
@@ -212,7 +212,11 @@ namespace Opm
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteMoleFractions);
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteTotalMassFractions);
             EWOMS_HIDE_PARAM(TypeTag, VtkWriteTotalMoleFractions);
-            
+
+            EWOMS_HIDE_PARAM(TypeTag, VtkWriteTortuosities);
+            EWOMS_HIDE_PARAM(TypeTag, VtkWriteDiffusionCoefficients);
+            EWOMS_HIDE_PARAM(TypeTag, VtkWriteEffectiveDiffusionCoefficients);
+
             EWOMS_END_PARAM_REGISTRATION(TypeTag);
 
             int mpiRank = comm.rank();
@@ -387,6 +391,10 @@ namespace Opm
             return ebosSimulator_.get();
         }
 
+        SimulatorTimer* getSimTimer() {
+            return simtimer_.get();
+        }
+
     private:
         // called by execute() or executeInitStep()
         int execute_(int (FlowMainEbos::* runOrInitFunc)(), bool cleanup)
