@@ -23,8 +23,10 @@
 #define OPM_MAIN_HEADER_INCLUDED
 
 #include <flow/flow_ebos_blackoil.hpp>
+#include <flow/flow_ebos_blackoil_legacyassembly.hpp>
 
 #include <flow/flow_ebos_gasoil.hpp>
+#include <flow/flow_ebos_gasoildiffuse.hpp>
 #include <flow/flow_ebos_gasoil_energy.hpp>
 #include <flow/flow_ebos_oilwater.hpp>
 #include <flow/flow_ebos_gaswater.hpp>
@@ -35,6 +37,7 @@
 #include <flow/flow_ebos_brine.hpp>
 #include <flow/flow_ebos_brine_saltprecipitation.hpp>
 #include <flow/flow_ebos_gaswater_saltprec_vapwat.hpp>
+#include <flow/flow_ebos_brine_precsalt_vapwat.hpp>
 #include <flow/flow_ebos_onephase.hpp>
 #include <flow/flow_ebos_onephase_energy.hpp>
 #include <flow/flow_ebos_oilwater_brine.hpp>
@@ -71,6 +74,10 @@
 #include <opm/simulators/utils/ParallelEclipseState.hpp>
 #endif
 
+#if HAVE_DAMARIS
+#include <opm/simulators/utils/DamarisOutputModule.hpp>
+#endif
+
 #include <cassert>
 #include <cstdlib>
 #include <filesystem>
@@ -95,20 +102,6 @@ struct FlowEarlyBird {
 
 namespace Opm {
 
-template <class TypeTag>
-void flowEbosSetDeck(std::shared_ptr<Deck> deck,
-                     std::shared_ptr<EclipseState> eclState,
-                     std::shared_ptr<Schedule> schedule,
-                     std::shared_ptr<SummaryConfig> summaryConfig)
-{
-    using Vanguard = GetPropType<TypeTag, Properties::Vanguard>;
-
-    Vanguard::setExternalDeck(deck);
-    Vanguard::setExternalEclState(eclState);
-    Vanguard::setExternalSchedule(schedule);
-    Vanguard::setExternalSummaryConfig(summaryConfig);
-}
-
 // ----------------- Main program -----------------
 template <class TypeTag>
 int flowEbosMain(int argc, char** argv, bool outputCout, bool outputFiles)
@@ -130,9 +123,6 @@ int flowEbosMain(int argc, char** argv, bool outputCout, bool outputFiles)
 //   usage scenarios, we refactored the original run() in flow.cpp into this class.
 class Main
 {
-private:
-    using FlowMainEbosType = FlowMainEbos<Properties::TTag::EclFlowProblem>;
-
 public:
     Main(int argc, char** argv) : argc_(argc), argv_(argv)  { initMPI();  }
 
@@ -180,6 +170,22 @@ public:
 
         EclGenericVanguard::setCommunication(nullptr);
 
+#if HAVE_DAMARIS
+        if (enableDamarisOutput_) {
+            int err;
+            if (isSimulationRank_) {
+                err = damaris_stop();
+                if (err != DAMARIS_OK) {
+                    std::cerr << "ERROR: Damaris library produced an error result for damaris_stop()" << std::endl;
+                }
+            }
+            err = damaris_finalize();
+            if (err != DAMARIS_OK) {
+                std::cerr << "ERROR: Damaris library produced an error result for damaris_finalize()" << std::endl;
+            }
+        }
+#endif // HAVE_DAMARIS
+
 #if HAVE_MPI && !HAVE_DUNE_FEM
         MPI_Finalize();
 #endif
@@ -249,6 +255,7 @@ public:
         return exitCode;
     }
 
+    using FlowMainEbosType = FlowMainEbos<Properties::TTag::EclFlowProblemTPFA>;
     // To be called from the Python interface code. Only do the
     // initialization and then return a pointer to the FlowEbosMain
     // object that can later be accessed directly from the Python interface
@@ -259,7 +266,7 @@ public:
         if (initialize_<Properties::TTag::FlowEarlyBird>(exitCode)) {
             // TODO: check that this deck really represents a blackoil
             // case. E.g. check that number of phases == 3
-            flowEbosBlackoilSetDeck(
+            EclGenericVanguard::setParams(
                 setupTime_,
                 deck_,
                 eclipseState_,
@@ -268,7 +275,7 @@ public:
                 std::move(this->actionState_),
                 std::move(this->wtestState_),
                 summaryConfig_);
-            return flowEbosBlackoilMainInit(
+            return flowEbosBlackoilTpfaMainInit(
                 argc_, argv_, outputCout_, outputFiles_);
         } else {
             //NOTE: exitCode was set by initialize_() above;
@@ -282,10 +289,20 @@ private:
         const auto& rspec = this->eclipseState_->runspec();
         const auto& phases = rspec.phases();
 
+        EclGenericVanguard::setParams(this->setupTime_,
+                                      this->deck_,
+                                      this->eclipseState_,
+                                      this->schedule_,
+                                      std::move(this->udqState_),
+                                      std::move(this->actionState_),
+                                      std::move(this->wtestState_),
+                                      this->summaryConfig_);
+
         // run the actual simulator
         //
         // TODO: make sure that no illegal combinations like thermal and
         //       twophase are requested.
+        const bool thermal = eclipseState_->getSimulationConfig().isThermal();
 
         // Single-phase case
         if (rspec.micp()) {
@@ -293,17 +310,17 @@ private:
         }
 
         // water-only case
-        else if(phases.size() == 1 && phases.active(Phase::WATER) && !eclipseState_->getSimulationConfig().isThermal()) {
+        else if(phases.size() == 1 && phases.active(Phase::WATER) && !thermal) {
             return this->runWaterOnly(phases);
         }
 
         // water-only case with energy
-        else if(phases.size() == 2 && phases.active(Phase::WATER) && eclipseState_->getSimulationConfig().isThermal()) {
+        else if(phases.size() == 2 && phases.active(Phase::WATER) && thermal) {
             return this->runWaterOnlyEnergy(phases);
         }
 
         // Twophase cases
-        else if (phases.size() == 2 && !eclipseState_->getSimulationConfig().isThermal()) {
+        else if (phases.size() == 2 && !thermal) {
             return this->runTwoPhase(phases);
         }
 
@@ -333,7 +350,7 @@ private:
         }
 
         // Energy case
-        else if (eclipseState_->getSimulationConfig().isThermal()) {
+        else if (thermal) {
             return this->runThermal(phases);
         }
 
@@ -356,8 +373,14 @@ private:
     template <class TypeTag>
     int dispatchStatic_()
     {
-        flowEbosSetDeck<TypeTag>(
-            deck_, eclipseState_, schedule_, summaryConfig_);
+        EclGenericVanguard::setParams(this->setupTime_,
+                                      this->deck_,
+                                      this->eclipseState_,
+                                      this->schedule_,
+                                      std::move(this->udqState_),
+                                      std::move(this->actionState_),
+                                      std::move(this->wtestState_),
+                                      this->summaryConfig_);
         return flowEbosMain<TypeTag>(argc_, argv_, outputCout_, outputFiles_);
     }
 
@@ -374,11 +397,6 @@ private:
         externalSetupTimer.start();
 
         handleVersionCmdLine_(argc_, argv_);
-#if HAVE_DUNE_FEM
-        int mpiRank = Dune::Fem::MPIManager::rank();
-#else
-        int mpiRank = EclGenericVanguard::comm().rank();
-#endif
 
         // we always want to use the default locale, and thus spare us the trouble
         // with incorrect locale settings.
@@ -407,11 +425,6 @@ private:
             return false; //  Whether to run the simulator
         }
 
-        FileOutputMode outputMode = FileOutputMode::OUTPUT_NONE;
-        outputCout_ = false;
-        if (mpiRank == 0)
-            outputCout_ = EWOMS_GET_PARAM(PreTypeTag, bool, EnableTerminalOutput);
-
         std::string deckFilename;
         std::string outputDir;
         if ( eclipseState_ ) {
@@ -420,8 +433,43 @@ private:
         }
         else {
             deckFilename = EWOMS_GET_PARAM(PreTypeTag, std::string, EclDeckFileName);
+            outputDir = EWOMS_GET_PARAM(PreTypeTag, std::string, OutputDir);
         }
 
+#if HAVE_DAMARIS
+        enableDamarisOutput_ = EWOMS_GET_PARAM(PreTypeTag, bool, EnableDamarisOutput);
+        if (enableDamarisOutput_) {
+            if (!outputDir.empty()) {
+                ensureOutputDirExists(outputDir);
+            }
+            // By default EnableDamarisOutputCollective is true so all simulation results will
+            // be written into one single file for each iteration using Parallel HDF5.
+            // It set to false, FilePerCore mode is used in Damaris, then simulation results in each
+            // node are aggregated by dedicated Damaris cores and stored to separate files per Damaris core.
+            // Irrespective of mode, output is written asynchronously at the end of each timestep.
+            const bool enableDamarisOutputCollective = EWOMS_GET_PARAM(PreTypeTag, bool, EnableDamarisOutputCollective);
+            // Using the ModifyModel class to set the XML file for Damaris.
+            DamarisOutput::initializeDamaris(EclGenericVanguard::comm(), EclGenericVanguard::comm().rank(), outputDir, enableDamarisOutputCollective);
+            int is_client;
+            MPI_Comm new_comm;
+            int err = damaris_start(&is_client);
+            isSimulationRank_ = (is_client > 0);
+            if (isSimulationRank_ && err == DAMARIS_OK) {
+                damaris_client_comm_get(&new_comm);
+                EclGenericVanguard::setCommunication(std::make_unique<Parallel::Communication>(new_comm));
+            } else {
+                return false;
+            }
+        }
+#endif // HAVE_DAMARIS
+
+        int mpiRank = EclGenericVanguard::comm().rank();
+        FileOutputMode outputMode = FileOutputMode::OUTPUT_NONE;
+        outputCout_ = false;
+        if (mpiRank == 0)
+            outputCout_ = EWOMS_GET_PARAM(PreTypeTag, bool, EnableTerminalOutput);
+
+
         if (deckFilename.empty()) {
             if (mpiRank == 0) {
                 std::cerr << "No input case given. Try '--help' for a usage description.\n";
@@ -451,9 +499,6 @@ private:
         try {
             auto python = std::make_shared<Python>();
             const bool init_from_restart_file = !EWOMS_GET_PARAM(PreTypeTag, bool, SchedRestart);
-            if (outputDir.empty())
-                outputDir = EWOMS_GET_PARAM(PreTypeTag, std::string, OutputDir);
-
             const bool allRanksDbgPrtLog = EWOMS_GET_PARAM(PreTypeTag, bool,
                                                       EnableLoggingFalloutWarning);
             outputMode = setupLogging(mpiRank,
@@ -583,12 +628,6 @@ private:
             return EXIT_FAILURE;
         }
 
-        flowEbosMICPSetDeck(this->setupTime_,
-                            this->deck_,
-                            this->eclipseState_,
-                            this->schedule_,
-                            this->summaryConfig_);
-
         return flowEbosMICPMain(this->argc_,
                                 this->argv_,
                                 this->outputCout_,
@@ -597,24 +636,36 @@ private:
 
     int runTwoPhase(const Phases& phases)
     {
+        const bool diffusive = eclipseState_->getSimulationConfig().isDiffusive();
+
         // oil-gas
         if (phases.active( Phase::OIL ) && phases.active( Phase::GAS )) {
-            flowEbosGasOilSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-            return flowEbosGasOilMain(argc_, argv_, outputCout_, outputFiles_);
+            if (diffusive) {
+                return flowEbosGasOilDiffuseMain(argc_, argv_, outputCout_, outputFiles_);
+            } else {
+                return flowEbosGasOilMain(argc_, argv_, outputCout_, outputFiles_);
+            }
         }
 
         // oil-water
         else if ( phases.active( Phase::OIL ) && phases.active( Phase::WATER ) ) {
-            flowEbosOilWaterSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
+            if (diffusive) {
+                if (outputCout_) {
+                    std::cerr << "The DIFFUSE option is not available for the two-phase water/oil model." << std::endl;
+                }
+                return EXIT_FAILURE;
+            }
             return flowEbosOilWaterMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
         // gas-water
         else if ( phases.active( Phase::GAS ) && phases.active( Phase::WATER ) ) {
-            flowEbosGasWaterSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
+            if (diffusive) {
+                if (outputCout_) {
+                    std::cerr << "The DIFFUSE option is not available for the two-phase gas/water model." << std::endl;
+                }
+                return EXIT_FAILURE;
+            }
             return flowEbosGasWaterMain(argc_, argv_, outputCout_, outputFiles_);
         }
         else {
@@ -645,22 +696,15 @@ private:
         }
 
         if (phases.size() == 3) { // oil water polymer case
-            flowEbosOilWaterPolymerSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
             return flowEbosOilWaterPolymerMain(argc_, argv_, outputCout_, outputFiles_);
         }
         else {
-            flowEbosPolymerSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
             return flowEbosPolymerMain(argc_, argv_, outputCout_, outputFiles_);
         }
     }
 
     int runFoam()
     {
-        flowEbosFoamSetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-
         return flowEbosFoamMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
@@ -673,8 +717,6 @@ private:
 
             return EXIT_FAILURE;
         }
-        flowEbosWaterOnlySetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
 
         return flowEbosWaterOnlyMain(argc_, argv_, outputCout_, outputFiles_);
     }
@@ -688,8 +730,6 @@ private:
 
             return EXIT_FAILURE;
         }
-        flowEbosWaterOnlyEnergySetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
 
         return flowEbosWaterOnlyEnergyMain(argc_, argv_, outputCout_, outputFiles_);
     }
@@ -707,33 +747,29 @@ private:
         if (phases.size() == 3) { 
 
             if (phases.active(Phase::OIL)){ // oil water brine case
-                flowEbosOilWaterBrineSetDeck(
-                    setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
                 return flowEbosOilWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
             }
             if (phases.active(Phase::GAS)){ // gas water brine case
                 if (eclipseState_->getSimulationConfig().hasPRECSALT() &&
                     eclipseState_->getSimulationConfig().hasVAPWAT()) {
                     //case with water vaporization into gas phase and salt precipitation
-                    flowEbosGasWaterSaltprecVapwatSetDeck(
-                        setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-                    return flowEbosGasWaterSaltprecVapwatMain(argc_, argv_, outputCout_, outputFiles_); 
+                    return flowEbosGasWaterSaltprecVapwatMain(argc_, argv_, outputCout_, outputFiles_);
                 } 
                 else {
-                    flowEbosGasWaterBrineSetDeck(
-                        setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
                     return flowEbosGasWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
                 }
             }
         }
         else if (eclipseState_->getSimulationConfig().hasPRECSALT()) {
-            flowEbosBrineSaltPrecipitationSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-            return flowEbosBrineSaltPrecipitationMain(argc_, argv_, outputCout_, outputFiles_);
+            if (eclipseState_->getSimulationConfig().hasVAPWAT()) {
+                    //case with water vaporization into gas phase and salt precipitation
+                    return flowEbosBrinePrecsaltVapwatMain(argc_, argv_, outputCout_, outputFiles_);
+            }
+            else {
+                return flowEbosBrineSaltPrecipitationMain(argc_, argv_, outputCout_, outputFiles_);
+            }
         }
         else {
-            flowEbosBrineSetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
             return flowEbosBrineMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
@@ -742,17 +778,11 @@ private:
 
     int runSolvent()
     {
-        flowEbosSolventSetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-
         return flowEbosSolventMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runExtendedBlackOil()
     {
-        flowEbosExtboSetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-
         return flowEbosExtboMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
@@ -760,29 +790,22 @@ private:
     {
         // oil-gas-thermal
         if (!phases.active( Phase::WATER ) && phases.active( Phase::OIL ) && phases.active( Phase::GAS )) {
-            flowEbosGasOilEnergySetDeck(
-                setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
             return flowEbosGasOilEnergyMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
-        flowEbosEnergySetDeck(
-            setupTime_, deck_, eclipseState_, schedule_, summaryConfig_);
-
         return flowEbosEnergyMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runBlackOil()
     {
-        flowEbosBlackoilSetDeck(this->setupTime_,
-                                this->deck_,
-                                this->eclipseState_,
-                                this->schedule_,
-                                std::move(this->udqState_),
-                                std::move(this->actionState_),
-                                std::move(this->wtestState_),
-                                this->summaryConfig_);
-
-        return flowEbosBlackoilMain(argc_, argv_, outputCout_, outputFiles_);
+        const bool diffusive = eclipseState_->getSimulationConfig().isDiffusive();
+        if (diffusive) {
+            // Use the traditional linearizer, as the TpfaLinearizer does not
+            // support the diffusion module yet.
+            return flowEbosBlackoilMain(argc_, argv_, outputCout_, outputFiles_);
+        } else {
+            return flowEbosBlackoilTpfaMain(argc_, argv_, outputCout_, outputFiles_);
+        }
     }
 
     int argc_{0};
@@ -806,6 +829,9 @@ private:
     // To demonstrate run with non_world_comm
     bool test_split_comm_ = false;
     bool isSimulationRank_ = true;
+#if HAVE_DAMARIS
+    bool enableDamarisOutput_ = false;
+#endif
 };
 
 } // namespace Opm
