@@ -22,22 +22,37 @@
 #ifndef OPM_ISTLSOLVER_EBOS_HEADER_INCLUDED
 #define OPM_ISTLSOLVER_EBOS_HEADER_INCLUDED
 
+#include <dune/istl/owneroverlapcopy.hh>
+
+#include <ebos/eclbasevanguard.hh>
+
+#include <opm/common/ErrorMacros.hpp>
+
+#include <opm/models/discretization/common/fvbaseproperties.hh>
+#include <opm/models/common/multiphasebaseproperties.hh>
 #include <opm/models/utils/parametersystem.hh>
 #include <opm/models/utils/propertysystem.hh>
+#include <opm/simulators/flow/BlackoilModelParametersEbos.hpp>
 #include <opm/simulators/linalg/ExtractParallelGridInformationToISTL.hpp>
-#include <opm/simulators/linalg/FlexibleSolver.hpp>
-#include <opm/simulators/linalg/MatrixBlock.hpp>
-#include <opm/simulators/linalg/ParallelIstlInformation.hpp>
+#include <opm/simulators/linalg/FlowLinearSolverParameters.hpp>
+#include <opm/simulators/linalg/matrixblock.hh>
+#include <opm/simulators/linalg/istlsparsematrixadapter.hh>
+#include <opm/simulators/linalg/PreconditionerWithUpdate.hpp>
 #include <opm/simulators/linalg/WellOperators.hpp>
 #include <opm/simulators/linalg/WriteSystemMatrixHelper.hpp>
 #include <opm/simulators/linalg/findOverlapRowsAndColumns.hpp>
 #include <opm/simulators/linalg/getQuasiImpesWeights.hpp>
 #include <opm/simulators/linalg/setupPropertyTree.hpp>
 
-#if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
-#include <opm/simulators/linalg/bda/BdaBridge.hpp>
-#include <opm/simulators/linalg/bda/WellContributions.hpp>
-#endif
+#include <any>
+#include <cstddef>
+#include <functional>
+#include <memory>
+#include <set>
+#include <sstream>
+#include <string>
+#include <tuple>
+#include <vector>
 
 namespace Opm::Properties {
 
@@ -58,10 +73,10 @@ struct SparseMatrixAdapter<TypeTag, TTag::FlowIstlSolver>
 private:
     using Scalar = GetPropType<TypeTag, Properties::Scalar>;
     enum { numEq = getPropValue<TypeTag, Properties::NumEq>() };
-    typedef MatrixBlock<Scalar, numEq, numEq> Block;
+    using Block = MatrixBlock<Scalar, numEq, numEq>;
 
 public:
-    typedef typename Linear::IstlSparseMatrixAdapter<Block> type;
+    using type = typename Linear::IstlSparseMatrixAdapter<Block>;
 };
 
 } // namespace Opm::Properties
@@ -69,6 +84,110 @@ public:
 namespace Opm
 {
 
+#if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
+template<class Matrix, class Vector, int block_size> class BdaBridge;
+class WellContributions;
+#endif
+
+namespace detail
+{
+
+template<class Matrix, class Vector, class Comm>
+struct FlexibleSolverInfo
+{
+    using AbstractSolverType = Dune::InverseOperator<Vector, Vector>;
+    using AbstractOperatorType = Dune::AssembledLinearOperator<Matrix, Vector, Vector>;
+    using AbstractPreconditionerType = Dune::PreconditionerWithUpdate<Vector, Vector>;
+
+    void create(const Matrix& matrix,
+                bool parallel,
+                const PropertyTree& prm,
+                size_t pressureIndex,
+                std::function<Vector()> trueFunc,
+                Comm& comm);
+
+    std::unique_ptr<AbstractSolverType> solver_;
+    std::unique_ptr<AbstractOperatorType> op_;
+    std::unique_ptr<LinearOperatorExtra<Vector,Vector>> wellOperator_;
+    AbstractPreconditionerType* pre_ = nullptr;
+    size_t interiorCellNum_ = 0;
+};
+
+#if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
+template<class Matrix, class Vector>
+struct BdaSolverInfo
+{
+  using WellContribFunc = std::function<void(WellContributions&)>;
+  using Bridge = BdaBridge<Matrix,Vector,Matrix::block_type::rows>;
+
+  BdaSolverInfo(const std::string& accelerator_mode,
+                const std::string& fpga_bitstream,
+                const int linear_solver_verbosity,
+                const int maxit,
+                const double tolerance,
+                const int platformID,
+                const int deviceID,
+                const std::string& opencl_ilu_reorder,
+                const std::string& linsolver);
+
+  ~BdaSolverInfo();
+
+  template<class Grid>
+  void prepare(const Grid& grid,
+               const Dune::CartesianIndexMapper<Grid>& cartMapper,
+               const std::vector<Well>& wellsForConn,
+               const std::vector<int>& cellPartition,
+               const size_t nonzeroes,
+               const bool useWellConn);
+
+  bool apply(Vector& rhs,
+             const bool useWellConn,
+             WellContribFunc getContribs,
+             const int rank,
+             Matrix& matrix,
+             Vector& x,
+             Dune::InverseOperatorResult& result);
+
+  int numJacobiBlocks_ = 0;
+
+private:
+  /// Create sparsity pattern for block-Jacobi matrix based on partitioning of grid.
+  /// Do not initialize the values, that is done in copyMatToBlockJac()
+  template<class Grid>
+  void blockJacobiAdjacency(const Grid& grid,
+                            const std::vector<int>& cell_part,
+                            size_t nonzeroes);
+
+  void copyMatToBlockJac(const Matrix& mat, Matrix& blockJac);
+
+  std::unique_ptr<Bridge> bridge_;
+  std::string accelerator_mode_;
+  std::unique_ptr<Matrix> blockJacobiForGPUILU0_;
+  std::vector<std::set<int>> wellConnectionsGraph_;
+};
+#endif
+
+#ifdef HAVE_MPI
+/// Copy values in parallel.
+void copyParValues(std::any& parallelInformation, size_t size,
+                   Dune::OwnerOverlapCopyCommunication<int,int>& comm);
+#endif
+
+/// Zero out off-diagonal blocks on rows corresponding to overlap cells
+/// Diagonal blocks on ovelap rows are set to diag(1.0).
+template<class Matrix>
+void makeOverlapRowsInvalid(Matrix& matrix,
+                            const std::vector<int>& overlapRows);
+
+/// Create sparsity pattern for block-Jacobi matrix based on partitioning of grid.
+/// Do not initialize the values, that is done in copyMatToBlockJac()
+template<class Matrix, class Grid>
+std::unique_ptr<Matrix> blockJacobiAdjacency(const Grid& grid,
+                                             const std::vector<int>& cell_part,
+                                             size_t nonzeroes,
+                                             const std::vector<std::set<int>>& wellConnectionsGraph);
+}
+
     /// This class solves the fully implicit black-oil system by
     /// solving the reduced system (after eliminating well variables)
     /// as a block-structured matrix (one block for all cell variables) for a fixed
@@ -87,21 +206,17 @@ namespace Opm
         using Matrix = typename SparseMatrixAdapter::IstlMatrix;
         using ThreadManager = GetPropType<TypeTag, Properties::ThreadManager>;
         using ElementContext = GetPropType<TypeTag, Properties::ElementContext>;
-        using FlexibleSolverType = Dune::FlexibleSolver<Matrix, Vector>;
+        using AbstractSolverType = Dune::InverseOperator<Vector, Vector>;
         using AbstractOperatorType = Dune::AssembledLinearOperator<Matrix, Vector, Vector>;
+        using AbstractPreconditionerType = Dune::PreconditionerWithUpdate<Vector, Vector>;
         using WellModelOperator = WellModelAsLinearOperator<WellModel, Vector, Vector>;
         using ElementMapper = GetPropType<TypeTag, Properties::ElementMapper>;
         constexpr static std::size_t pressureIndex = GetPropType<TypeTag, Properties::Indices>::pressureSwitchIdx;
 
-#if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
-        static const unsigned int block_size = Matrix::block_type::rows;
-        std::unique_ptr<BdaBridge<Matrix, Vector, block_size>> bdaBridge;
-#endif
-
 #if HAVE_MPI
         using CommunicationType = Dune::OwnerOverlapCopyCommunication<int,int>;
 #else
-        using CommunicationType = Dune::CollectiveCommunication< int >;
+        using CommunicationType = Dune::CollectiveCommunication<int>;
 #endif
 
     public:
@@ -118,6 +233,7 @@ namespace Opm
         explicit ISTLSolverEbos(const Simulator& simulator)
             : simulator_(simulator),
               iterations_( 0 ),
+              calls_( 0 ),
               converged_(false),
               matrix_()
         {
@@ -146,8 +262,16 @@ namespace Opm
                 const std::string opencl_ilu_reorder = EWOMS_GET_PARAM(TypeTag, std::string, OpenclIluReorder);
                 const int linear_solver_verbosity = parameters_.linear_solver_verbosity_;
                 std::string fpga_bitstream = EWOMS_GET_PARAM(TypeTag, std::string, FpgaBitstream);
-                std::string linsolver = EWOMS_GET_PARAM(TypeTag, std::string, Linsolver);
-                bdaBridge.reset(new BdaBridge<Matrix, Vector, block_size>(accelerator_mode, fpga_bitstream, linear_solver_verbosity, maxit, tolerance, platformID, deviceID, opencl_ilu_reorder, linsolver));
+                std::string linsolver = EWOMS_GET_PARAM(TypeTag, std::string, LinearSolver);
+                bdaBridge = std::make_unique<detail::BdaSolverInfo<Matrix,Vector>>(accelerator_mode,
+                                                                                   fpga_bitstream,
+                                                                                   linear_solver_verbosity,
+                                                                                   maxit,
+                                                                                   tolerance,
+                                                                                   platformID,
+                                                                                   deviceID,
+                                                                                   opencl_ilu_reorder,
+                                                                                   linsolver);
             }
 #else
             if (EWOMS_GET_PARAM(TypeTag, std::string, AcceleratorMode) != "none") {
@@ -157,11 +281,10 @@ namespace Opm
             extractParallelGridInformationToISTL(simulator_.vanguard().grid(), parallelInformation_);
 
             // For some reason simulator_.model().elementMapper() is not initialized at this stage
-            // Hence const auto& elemMapper = simulator_.model().elementMapper(); does not work.
+            //const auto& elemMapper = simulator_.model().elementMapper(); //does not work.
             // Set it up manually
             ElementMapper elemMapper(simulator_.vanguard().gridView(), Dune::mcmgElementLayout());
             detail::findOverlapAndInterior(simulator_.vanguard().grid(), elemMapper, overlapRows_, interiorRows_);
-
             useWellConn_ = EWOMS_GET_PARAM(TypeTag, bool, MatrixAddWellContributions);
 #if HAVE_FPGA
             // check usage of MatrixAddWellContributions: for FPGA they must be included
@@ -178,7 +301,7 @@ namespace Opm
                 OPM_THROW_NOLOG(std::runtime_error, msg);
             }
 
-            interiorCellNum_ = detail::numMatrixRowsToUseInSolver(simulator_.vanguard().grid(), true);
+            flexibleSolver_.interiorCellNum_ = detail::numMatrixRowsToUseInSolver(simulator_.vanguard().grid(), true);
 
             // Print parameters to PRT/DBG logs.
             if (on_io_rank) {
@@ -190,19 +313,17 @@ namespace Opm
         }
 
         // nothing to clean here
-        void eraseMatrix() {
+        void eraseMatrix()
+        {
         }
 
         void prepare(const SparseMatrixAdapter& M, Vector& b)
         {
             static bool firstcall = true;
 #if HAVE_MPI
-            if (firstcall && parallelInformation_.type() == typeid(ParallelISTLInformation)) {
-                // Parallel case.
-                const ParallelISTLInformation* parinfo = std::any_cast<ParallelISTLInformation>(&parallelInformation_);
-                assert(parinfo);
+            if (firstcall) {
                 const size_t size = M.istlMatrix().N();
-                parinfo->copyValuesTo(comm_->indexSet(), comm_->remoteIndices(), size, 1);
+                detail::copyParValues(parallelInformation_, size, *comm_);
             }
 #endif
 
@@ -212,6 +333,17 @@ namespace Opm
                 // to the original one with a deleter that does nothing.
                 // Outch! We need to be able to scale the linear system! Hence const_cast
                 matrix_ = const_cast<Matrix*>(&M.istlMatrix());
+
+                useWellConn_ = EWOMS_GET_PARAM(TypeTag, bool, MatrixAddWellContributions);
+                // setup sparsity pattern for jacobi matrix for preconditioner (only used for openclSolver)
+#if HAVE_OPENCL
+                bdaBridge->numJacobiBlocks_ = EWOMS_GET_PARAM(TypeTag, int, NumJacobiBlocks);
+                bdaBridge->prepare(simulator_.vanguard().grid(),
+                                   simulator_.vanguard().cartesianIndexMapper(),
+                                   simulator_.vanguard().schedule().getWellsatEnd(),
+                                   simulator_.vanguard().cellPartition(),
+                                   getMatrix().nonzeroes(), useWellConn_);
+#endif
             } else {
                 // Pointers should not change
                 if ( &(M.istlMatrix()) != matrix_ ) {
@@ -222,26 +354,31 @@ namespace Opm
             rhs_ = &b;
 
             if (isParallel() && prm_.get<std::string>("preconditioner.type") != "ParOverILU0") {
-                makeOverlapRowsInvalid(getMatrix());
+                detail::makeOverlapRowsInvalid(getMatrix(), overlapRows_);
             }
             prepareFlexibleSolver();
             firstcall = false;
         }
 
 
-        void setResidual(Vector& /* b */) {
+        void setResidual(Vector& /* b */)
+        {
             // rhs_ = &b; // Must be handled in prepare() instead.
         }
 
-        void getResidual(Vector& b) const {
+        void getResidual(Vector& b) const
+        {
             b = *rhs_;
         }
 
-        void setMatrix(const SparseMatrixAdapter& /* M */) {
+        void setMatrix(const SparseMatrixAdapter& /* M */)
+        {
             // matrix_ = &M.istlMatrix(); // Must be handled in prepare() instead.
         }
 
-        bool solve(Vector& x) {
+        bool solve(Vector& x)
+        {
+            calls_ += 1;
             // Write linear system if asked for.
             const int verbosity = prm_.get<int>("verbosity", 0);
             const bool write_matrix = verbosity > 10;
@@ -254,47 +391,23 @@ namespace Opm
 
             // Solve system.
             Dune::InverseOperatorResult result;
-            bool accelerator_was_used = false;
 
             // Use GPU if: available, chosen by user, and successful.
             // Use FPGA if: support compiled, chosen by user, and successful.
 #if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
-            bool use_gpu = bdaBridge->getUseGpu();
-            bool use_fpga = bdaBridge->getUseFpga();
-            if (use_gpu || use_fpga) {
-                const std::string accelerator_mode = EWOMS_GET_PARAM(TypeTag, std::string, AcceleratorMode);
-                auto wellContribs = WellContributions::create(accelerator_mode, useWellConn_);
-                bdaBridge->initWellContributions(*wellContribs);
-
-                // the WellContributions can only be applied separately with CUDA or OpenCL, not with an FPGA or amgcl
-#if HAVE_CUDA || HAVE_OPENCL
-                if (!useWellConn_) {
-                    simulator_.problem().wellModel().getWellContributions(*wellContribs);
-                }
+            std::function<void(WellContributions&)> getContribs =
+                [this](WellContributions& w)
+                {
+                    this->simulator_.problem().wellModel().getWellContributions(w);
+                };
+            if (!bdaBridge->apply(*rhs_, useWellConn_, getContribs,
+                                  simulator_.gridView().comm().rank(),
+                                  const_cast<Matrix&>(this->getMatrix()),
+                                  x, result))
 #endif
-
-                // Const_cast needed since the CUDA stuff overwrites values for better matrix condition..
-                bdaBridge->solve_system(const_cast<Matrix*>(&getMatrix()), *rhs_, *wellContribs, result);
-                if (result.converged) {
-                    // get result vector x from non-Dune backend, iff solve was successful
-                    bdaBridge->get_result(x);
-                    accelerator_was_used = true;
-                } else {
-                    // warn about CPU fallback
-                    // BdaBridge might have disabled its BdaSolver for this simulation due to some error
-                    // in that case the BdaBridge is disabled and flexibleSolver is always used
-                    // or maybe the BdaSolver did not converge in time, then it will be used next linear solve
-                    if (simulator_.gridView().comm().rank() == 0) {
-                        OpmLog::warning(bdaBridge->getAccleratorName() + " did not converge, now trying Dune to solve current linear system...");
-                    }
-                }
-            }
-#endif
-
-            // Otherwise, use flexible istl solver.
-            if (!accelerator_was_used) {
-                assert(flexibleSolver_);
-                flexibleSolver_->apply(x, *rhs_, result);
+            {
+                assert(flexibleSolver_.solver_);
+                flexibleSolver_.solver_->apply(x, *rhs_, result);
             }
 
             // Check convergence, iterations etc.
@@ -317,19 +430,8 @@ namespace Opm
         const std::any& parallelInformation() const { return parallelInformation_; }
 
     protected:
-        // 3x3 matrix block inversion was unstable at least 2.3 until and including
-        // 2.5.0. There may still be some issue with the 4x4 matrix block inversion
-        // we therefore still use the block inversion in OPM
-        typedef ParallelOverlappingILU0<Dune::BCRSMatrix<Dune::MatrixBlock<typename Matrix::field_type,
-                                                                            Matrix::block_type::rows,
-                                                                            Matrix::block_type::cols> >,
-                                                                            Vector, Vector> SeqPreconditioner;
-
 #if HAVE_MPI
-        typedef Dune::OwnerOverlapCopyCommunication<int, int> Comm;
-        // 3x3 matrix block inversion was unstable from at least 2.3 until and
-        // including 2.5.0
-        typedef ParallelOverlappingILU0<Matrix,Vector,Vector,Comm> ParPreconditioner;
+        using Comm = Dune::OwnerOverlapCopyCommunication<int, int>;
 #endif
 
         void checkConvergence( const Dune::InverseOperatorResult& result ) const
@@ -357,42 +459,28 @@ namespace Opm
         void prepareFlexibleSolver()
         {
 
-            std::function<Vector()> weightsCalculator = getWeightsCalculator();
-
             if (shouldCreateSolver()) {
-                if (isParallel()) {
-#if HAVE_MPI
-                    if (useWellConn_) {
-                        using ParOperatorType = Dune::OverlappingSchwarzOperator<Matrix, Vector, Vector, Comm>;
-                        linearOperatorForFlexibleSolver_ = std::make_unique<ParOperatorType>(getMatrix(), *comm_);
-                        flexibleSolver_ = std::make_unique<FlexibleSolverType>(*linearOperatorForFlexibleSolver_, *comm_, prm_, weightsCalculator,
-                                                                               pressureIndex);
-                    } else {
-                        using ParOperatorType = WellModelGhostLastMatrixAdapter<Matrix, Vector, Vector, true>;
-                        wellOperator_ = std::make_unique<WellModelOperator>(simulator_.problem().wellModel());
-                        linearOperatorForFlexibleSolver_ = std::make_unique<ParOperatorType>(getMatrix(), *wellOperator_, interiorCellNum_);
-                        flexibleSolver_ = std::make_unique<FlexibleSolverType>(*linearOperatorForFlexibleSolver_, *comm_, prm_, weightsCalculator,
-                                                                               pressureIndex);
-                    }
-#endif
-                } else {
-                    if (useWellConn_) {
-                        using SeqOperatorType = Dune::MatrixAdapter<Matrix, Vector, Vector>;
-                        linearOperatorForFlexibleSolver_ = std::make_unique<SeqOperatorType>(getMatrix());
-                        flexibleSolver_ = std::make_unique<FlexibleSolverType>(*linearOperatorForFlexibleSolver_, prm_, weightsCalculator,
-                                                                               pressureIndex);
-                    } else {
-                        using SeqOperatorType = WellModelMatrixAdapter<Matrix, Vector, Vector, false>;
-                        wellOperator_ = std::make_unique<WellModelOperator>(simulator_.problem().wellModel());
-                        linearOperatorForFlexibleSolver_ = std::make_unique<SeqOperatorType>(getMatrix(), *wellOperator_);
-                        flexibleSolver_ = std::make_unique<FlexibleSolverType>(*linearOperatorForFlexibleSolver_, prm_, weightsCalculator,
-                                                                               pressureIndex);
-                    }
+                std::function<Vector()> trueFunc =
+                    [this]
+                    {
+                        return this->getTrueImpesWeights(pressureIndex);
+                    };
+
+                if (!useWellConn_) {
+                    auto wellOp = std::make_unique<WellModelOperator>(simulator_.problem().wellModel());
+                    flexibleSolver_.wellOperator_ = std::move(wellOp);
                 }
+
+                flexibleSolver_.create(getMatrix(),
+                                       isParallel(),
+                                       prm_,
+                                       pressureIndex,
+                                       trueFunc,
+                                       *comm_);
             }
             else
             {
-                flexibleSolver_->preconditioner().update();
+                flexibleSolver_.pre_->update();
             }
         }
 
@@ -403,7 +491,7 @@ namespace Opm
         {
             // Decide if we should recreate the solver or just do
             // a minimal preconditioner update.
-            if (!flexibleSolver_) {
+            if (!flexibleSolver_.solver_) {
                 return true;
             }
             if (this->parameters_.cpr_reuse_setup_ == 0) {
@@ -419,45 +507,28 @@ namespace Opm
                 // Recreate solver if the last solve used more than 10 iterations.
                 return this->iterations() > 10;
             }
-
-            // Otherwise, do not recreate solver.
-            assert(this->parameters_.cpr_reuse_setup_ == 3);
-
-            return false;
-        }
-
-
-        /// Return an appropriate weight function if a cpr preconditioner is asked for.
-        std::function<Vector()> getWeightsCalculator() const
-        {
-            std::function<Vector()> weightsCalculator;
-
-            using namespace std::string_literals;
-
-            auto preconditionerType = prm_.get("preconditioner.type"s, "cpr"s);
-            if (preconditionerType == "cpr" || preconditionerType == "cprt") {
-                const bool transpose = preconditionerType == "cprt";
-                const auto weightsType = prm_.get("preconditioner.weight_type"s, "quasiimpes"s);
-                if (weightsType == "quasiimpes") {
-                    // weights will be created as default in the solver
-                    // assignment p = pressureIndex prevent compiler warning about
-                    // capturing variable with non-automatic storage duration
-                    weightsCalculator = [this, transpose, p = pressureIndex]() {
-                        return Amg::getQuasiImpesWeights<Matrix, Vector>(this->getMatrix(), p, transpose);
-                    };
-                } else if (weightsType == "trueimpes") {
-                    // assignment p = pressureIndex prevent compiler warning about
-                    // capturing variable with non-automatic storage duration
-                    weightsCalculator = [this, p = pressureIndex]() {
-                        return this->getTrueImpesWeights(p);
-                    };
-                } else {
-                    OPM_THROW(std::invalid_argument,
-                              "Weights type " << weightsType << "not implemented for cpr."
-                              << " Please use quasiimpes or trueimpes.");
-                }
+            if (this->parameters_.cpr_reuse_setup_ == 3) {
+                // Recreate solver if the last solve used more than 10 iterations.
+                return false;
             }
-            return weightsCalculator;
+            if (this->parameters_.cpr_reuse_setup_ == 4) {
+                // Recreate solver every 'step' solve calls.
+                const int step = this->parameters_.cpr_reuse_interval_;
+                const bool create = ((calls_ % step) == 0);
+                return create;
+            }
+
+            // If here, we have an invalid parameter.
+            const bool on_io_rank = (simulator_.gridView().comm().rank() == 0);
+            std::string msg = "Invalid value: " + std::to_string(this->parameters_.cpr_reuse_setup_)
+                + " for --cpr-reuse-setup parameter, run with --help to see allowed values.";
+            if (on_io_rank) {
+                OpmLog::error(msg);
+            }
+            throw std::runtime_error(msg);
+
+            // Never reached.
+            return false;
         }
 
 
@@ -468,36 +539,13 @@ namespace Opm
         {
             Vector weights(rhs_->size());
             ElementContext elemCtx(simulator_);
-            Amg::getTrueImpesWeights(pressureVarIndex, weights, simulator_.vanguard().gridView(),
+            Amg::getTrueImpesWeights(pressureVarIndex, weights,
+                                     simulator_.vanguard().gridView(),
                                      elemCtx, simulator_.model(),
                                      ThreadManager::threadId());
             return weights;
         }
 
-
-        /// Zero out off-diagonal blocks on rows corresponding to overlap cells
-        /// Diagonal blocks on ovelap rows are set to diag(1.0).
-        void makeOverlapRowsInvalid(Matrix& matrix) const
-        {
-            //value to set on diagonal
-            const int numEq = Matrix::block_type::rows;
-            typename Matrix::block_type diag_block(0.0);
-            for (int eq = 0; eq < numEq; ++eq)
-                diag_block[eq][eq] = 1.0;
-
-            //loop over precalculated overlap rows and columns
-            for (auto row = overlapRows_.begin(); row != overlapRows_.end(); row++ )
-                {
-                    int lcell = *row;
-                    // Zero out row.
-                    matrix[lcell] = 0.0;
-
-                    //diagonal block set to diag(1.0).
-                    matrix[lcell][lcell] = diag_block;
-                }
-        }
-
-
         Matrix& getMatrix()
         {
             return *matrix_;
@@ -510,6 +558,7 @@ namespace Opm
 
         const Simulator& simulator_;
         mutable int iterations_;
+        mutable int calls_;
         mutable bool converged_;
         std::any parallelInformation_;
 
@@ -517,19 +566,18 @@ namespace Opm
         Matrix* matrix_;
         Vector *rhs_;
 
-        std::unique_ptr<FlexibleSolverType> flexibleSolver_;
-        std::unique_ptr<AbstractOperatorType> linearOperatorForFlexibleSolver_;
-        std::unique_ptr<WellModelAsLinearOperator<WellModel, Vector, Vector>> wellOperator_;
+#if HAVE_CUDA || HAVE_OPENCL || HAVE_FPGA || HAVE_AMGCL
+        std::unique_ptr<detail::BdaSolverInfo<Matrix, Vector>> bdaBridge;
+#endif
+
+        detail::FlexibleSolverInfo<Matrix,Vector,CommunicationType> flexibleSolver_;
         std::vector<int> overlapRows_;
         std::vector<int> interiorRows_;
-        std::vector<std::set<int>> wellConnectionsGraph_;
 
         bool useWellConn_;
-        size_t interiorCellNum_;
 
         FlowLinearSolverParameters parameters_;
         PropertyTree prm_;
-        bool scale_variables_;
 
         std::shared_ptr< CommunicationType > comm_;
     }; // end ISTLSolver
