@@ -61,9 +61,9 @@ openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_,
         use_cpr = false;
         use_isai = true;
     } else if (linsolver.compare("cpr_trueimpes") == 0) {
-        OPM_THROW(std::logic_error, "Error openclSolver does not support --linsolver=cpr_trueimpes");
+        OPM_THROW(std::logic_error, "Error openclSolver does not support --linerar-solver=cpr_trueimpes");
     } else {
-        OPM_THROW(std::logic_error, "Error unknown value for argument --linsolver, " + linsolver);
+        OPM_THROW(std::logic_error, "Error unknown value for argument --linear-solver, " + linsolver);
     }
 
     using PreconditionerType = typename Preconditioner<block_size>::PreconditionerType;
@@ -279,7 +279,9 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         OpmLog::info(out.str());
     }
 
-    t_rest.start();
+    if (verbosity >= 3) {
+        t_rest.start();
+    }
     for (it = 0.5; it < maxit; it += 0.5) {
         rhop = rho;
         rho = OpenclKernels::dot(d_rw, d_r, d_tmp, N);
@@ -288,32 +290,47 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
             beta = (rho / rhop) * (alpha / omega);
             OpenclKernels::custom(d_p, d_v, d_r, omega, beta, N);
         }
-        t_rest.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_rest.stop();
+            t_prec.start();
+        }
 
         // pw = prec(p)
-        t_prec.start();
         prec->apply(d_p, d_pw);
-        t_prec.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_prec.stop();
+            t_spmv.start();
+        }
 
         // v = A * pw
-        t_spmv.start();
         OpenclKernels::spmv(d_Avals, d_Acols, d_Arows, d_pw, d_v, Nb, block_size);
-        t_spmv.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_spmv.stop();
+            t_well.start();
+        }
 
         // apply wellContributions
-        t_well.start();
         if(wellContribs.getNumWells() > 0){
             static_cast<WellContributionsOCL&>(wellContribs).apply(d_pw, d_v, d_toOrder);
         }
-        t_well.stop();
+        if(verbosity >= 3) {
+            queue->finish();
+            t_well.stop();
+            t_rest.start();
+        }
 
-        t_rest.start();
         tmp1 = OpenclKernels::dot(d_rw, d_v, d_tmp, N);
         alpha = rho / tmp1;
         OpenclKernels::axpy(d_v, -alpha, d_r, N);      // r = r - alpha * v
         OpenclKernels::axpy(d_pw, alpha, d_x, N);      // x = x + alpha * pw
         norm = OpenclKernels::norm(d_r, d_tmp, N);
-        t_rest.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_rest.stop();
+        }
 
         if (norm < tolerance * norm_0) {
             break;
@@ -322,30 +339,44 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         it += 0.5;
 
         // s = prec(r)
-        t_prec.start();
+        if (verbosity >= 3) {
+            t_prec.start();
+        }
         prec->apply(d_r, d_s);
-        t_prec.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_prec.stop();
+            t_spmv.start();
+        }
 
         // t = A * s
-        t_spmv.start();
         OpenclKernels::spmv(d_Avals, d_Acols, d_Arows, d_s, d_t, Nb, block_size);
-        t_spmv.stop();
+        if(verbosity >= 3){
+            queue->finish();
+            t_spmv.stop();
+            t_well.start();
+        }
 
         // apply wellContributions
-        t_well.start();
         if(wellContribs.getNumWells() > 0){
             static_cast<WellContributionsOCL&>(wellContribs).apply(d_s, d_t, d_toOrder);
         }
-        t_well.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_well.stop();
+            t_rest.start();
+        }
 
-        t_rest.start();
         tmp1 = OpenclKernels::dot(d_t, d_r, d_tmp, N);
         tmp2 = OpenclKernels::dot(d_t, d_t, d_tmp, N);
         omega = tmp1 / tmp2;
         OpenclKernels::axpy(d_s, omega, d_x, N);     // x = x + omega * s
         OpenclKernels::axpy(d_t, -omega, d_r, N);    // r = r - omega * t
         norm = OpenclKernels::norm(d_r, d_tmp, N);
-        t_rest.stop();
+        if (verbosity >= 3) {
+            queue->finish();
+            t_rest.stop();
+        }
 
         if (norm < tolerance * norm_0) {
             break;
@@ -370,7 +401,7 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
             ", time per iteration: " << res.elapsed / it << ", iterations: " << it;
         OpmLog::info(out.str());
     }
-    if (verbosity >= 4) {
+    if (verbosity >= 3) {
         std::ostringstream out;
         out << "openclSolver::prec_apply:  " << t_prec.elapsed() << " s\n";
         out << "wellContributions::apply:  " << t_well.elapsed() << " s\n";
@@ -383,14 +414,21 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
 
 
 template <unsigned int block_size>
-void openclSolverBackend<block_size>::initialize(int N_, int nnz_, int dim, double *vals, int *rows, int *cols) {
-    this->N = N_;
-    this->nnz = nnz_;
-    this->nnzb = nnz_ / block_size / block_size;
+void openclSolverBackend<block_size>::initialize(std::shared_ptr<BlockedMatrix> matrix, std::shared_ptr<BlockedMatrix> jacMatrix) {
+    this->Nb = matrix->Nb;
+    this->N = Nb * block_size;
+    this->nnzb = matrix->nnzbs;
+    this->nnz = nnzb * block_size * block_size;
+
+    if (jacMatrix) {
+        useJacMatrix = true;
+    }
 
-    Nb = (N + dim - 1) / dim;
     std::ostringstream out;
-    out << "Initializing GPU, matrix size: " << N << " blocks, nnzb: " << nnzb << "\n";
+    out << "Initializing GPU, matrix size: " << Nb << " blockrows, nnzb: " << nnzb << "\n";
+    if (useJacMatrix) {
+        out << "Blocks in ILU matrix: " << jacMatrix->nnzbs << "\n";
+    }
     out << "Maxit: " << maxit << std::scientific << ", tolerance: " << tolerance << "\n";
     out << "PlatformID: " << platformID << ", deviceID: " << deviceID << "\n";
     OpmLog::info(out.str());
@@ -401,9 +439,12 @@ void openclSolverBackend<block_size>::initialize(int N_, int nnz_, int dim, doub
         prec->setOpencl(context, queue);
 
 #if COPY_ROW_BY_ROW
-        vals_contiguous.resize(nnz);
+        vals_contiguous = new double[N];
 #endif
-        mat.reset(new BlockedMatrix(Nb, nnzb, block_size, vals, cols, rows));
+        // mat.reset(new BlockedMatrix(Nb, nnzb, block_size, vals, cols, rows));
+        // jacMat.reset(new BlockedMatrix(Nb, jac_nnzb, block_size, vals2, cols2, rows2));
+        mat = matrix;
+        jacMat = jacMatrix;
 
         d_x = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
         d_b = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
@@ -441,7 +482,6 @@ void openclSolverBackend<block_size>::initialize(int N_, int nnz_, int dim, doub
     initialized = true;
 } // end initialize()
 
-
 template <unsigned int block_size>
 void openclSolverBackend<block_size>::finalize() {
     if (opencl_ilu_reorder != ILUReorder::NONE) {
@@ -527,8 +567,11 @@ template <unsigned int block_size>
 bool openclSolverBackend<block_size>::analyze_matrix() {
     Timer t;
 
-    // bool success = bilu0->init(mat.get());
-    bool success = prec->analyze_matrix(mat.get());
+    bool success;
+    if (useJacMatrix)
+        success = prec->analyze_matrix(mat.get(), jacMat.get());
+    else
+        success = prec->analyze_matrix(mat.get());
 
     if (opencl_ilu_reorder == ILUReorder::NONE) {
         rmat = mat.get();
@@ -579,7 +622,11 @@ template <unsigned int block_size>
 bool openclSolverBackend<block_size>::create_preconditioner() {
     Timer t;
 
-    bool result = prec->create_preconditioner(mat.get());
+    bool result;
+    if (useJacMatrix)
+        result = prec->create_preconditioner(mat.get(), jacMat.get());
+    else
+        result = prec->create_preconditioner(mat.get());
 
     if (verbosity > 2) {
         std::ostringstream out;
@@ -639,21 +686,26 @@ void openclSolverBackend<block_size>::get_result(double *x) {
 
 
 template <unsigned int block_size>
-SolverStatus openclSolverBackend<block_size>::solve_system(int N_, int nnz_, int dim, double *vals, int *rows, int *cols, double *b, WellContributions& wellContribs, BdaResult &res) {
+SolverStatus openclSolverBackend<block_size>::solve_system(std::shared_ptr<BlockedMatrix> matrix,
+                                                           double *b,
+                                                           std::shared_ptr<BlockedMatrix> jacMatrix,
+                                                           WellContributions& wellContribs,
+                                                           BdaResult &res)
+{
     if (initialized == false) {
-        initialize(N_, nnz_,  dim, vals, rows, cols);
+        initialize(matrix, jacMatrix);
         if (analysis_done == false) {
             if (!analyze_matrix()) {
                 return SolverStatus::BDA_SOLVER_ANALYSIS_FAILED;
             }
         }
-        update_system(vals, b, wellContribs);
+        update_system(matrix->nnzValues, b, wellContribs);
         if (!create_preconditioner()) {
             return SolverStatus::BDA_SOLVER_CREATE_PRECONDITIONER_FAILED;
         }
         copy_system_to_gpu();
     } else {
-        update_system(vals, b, wellContribs);
+        update_system(matrix->nnzValues, b, wellContribs);
         if (!create_preconditioner()) {
             return SolverStatus::BDA_SOLVER_CREATE_PRECONDITIONER_FAILED;
         }
