@@ -17,6 +17,10 @@
   along with OPM.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#if HAVE_CONFIG_H
+#include "config.h"
+#endif // HAVE_CONFIG_H
+
 #include <opm/simulators/utils/UnsupportedFlowKeywords.hpp>
 
 namespace Opm::FlowKeywordValidation
@@ -320,7 +324,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"IONXSURF", {false, std::nullopt}},
         {"ISOLNUM", {false, std::nullopt}},
         {"JFUNCR", {false, std::nullopt}},
-        {"KRNUM", {false, std::nullopt}},
         {"KRNUMMF", {false, std::nullopt}},
         {"LANGMPL", {false, std::nullopt}},
         {"LANGMUIR", {false, std::nullopt}},
@@ -380,7 +383,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"MESSOPTS", {false, std::nullopt}},
         {"MESSSRVC", {false, std::nullopt}},
         {"MINNNCT", {false, std::nullopt}},
-        {"MINPORV", {false, std::nullopt}},
         {"MLANG", {false, std::nullopt}},
         {"MLANGSLV", {false, std::nullopt}},
         {"MONITOR", {false, std::nullopt}},
@@ -399,7 +401,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"NARROW", {false, std::nullopt}},
         {"NCONSUMP", {false, std::nullopt}},
         {"NEFAC", {false, std::nullopt}},
-        {"NETBALAN", {false, std::nullopt}},
         {"NETCOMPA", {false, std::nullopt}},
         {"NEXT", {false, std::nullopt}},
         {"NEXTSTPL", {false, std::nullopt}},
@@ -550,7 +551,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"RSCONST", {false, std::nullopt}},
         {"RSCONSTT", {false, std::nullopt}},
         {"RSSPEC", {false, std::nullopt}},
-        {"RTEMPA", {false, std::nullopt}},
         {"RWGSALT", {false, std::nullopt}},
         {"SAMG", {false, std::nullopt}},
         {"SAVE", {false, std::nullopt}},
@@ -624,7 +624,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"TEMP", {false, std::nullopt}},
         {"TEMPNODE", {false, std::nullopt}},
         {"TEMPTVD", {false, std::nullopt}},
-        {"TEMPVD", {false, std::nullopt}},
         {"TIGHTEN", {false, std::nullopt}},
         {"TIGHTENP", {false, std::nullopt}},
         {"TIME", {false, std::nullopt}},
@@ -762,7 +761,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WTHPMAX", {false, std::nullopt}},
         {"WTMULT", {false, std::nullopt}},
         {"WVFPDP", {false, std::nullopt}},
-        {"WVFPEXP", {false, std::nullopt}},
         {"WWPAVE", {false, std::nullopt}},
         {"ZIPPY2", {false, std::nullopt}},
         {"ZIPP2OFF", {false, std::nullopt}},
