@@ -93,23 +93,6 @@ namespace {
             ->setMessageLimiter(std::make_shared<Opm::MessageLimiter>(10, limits));
     }
 
-    void ensureOutputDirExists_(const std::string& cmdline_output_dir)
-    {
-        namespace fs = std::filesystem;
-
-        if (! fs::is_directory(cmdline_output_dir)) {
-            try {
-                fs::create_directories(cmdline_output_dir);
-            }
-            catch (...) {
-                throw std::runtime_error {
-                    fmt::format("Creation of output directory '{}' failed",
-                                cmdline_output_dir)
-                };
-            }
-        }
-    }
-
     void loadObjectsFromRestart(const Opm::Deck&                     deck,
                                 const Opm::Parser&                   parser,
                                 const Opm::ParseContext&             parseContext,
@@ -376,6 +359,24 @@ namespace {
 
 // ---------------------------------------------------------------------------
 
+
+void Opm::ensureOutputDirExists(const std::string& cmdline_output_dir)
+{
+    namespace fs = std::filesystem;
+
+    if (! fs::is_directory(cmdline_output_dir)) {
+        try {
+            fs::create_directories(cmdline_output_dir);
+        }
+        catch (...) {
+            throw std::runtime_error {
+                fmt::format("Creation of output directory '{}' failed",
+                            cmdline_output_dir)
+                    };
+        }
+    }
+}
+
 // Setup the OpmLog backends
 Opm::FileOutputMode
 Opm::setupLogging(const int          mpi_rank_,
@@ -387,7 +388,7 @@ Opm::setupLogging(const int          mpi_rank_,
                   const bool         allRanksDbgLog)
 {
     if (!cmdline_output_dir.empty()) {
-        ensureOutputDirExists_(cmdline_output_dir);
+        ensureOutputDirExists(cmdline_output_dir);
     }
 
     // create logFile
@@ -465,7 +466,11 @@ Opm::setupLogging(const int          mpi_rank_,
     if (mpi_rank_ == 0) {
         std::shared_ptr<Opm::StreamLog> streamLog = std::make_shared<Opm::StreamLog>(std::cout, Opm::Log::StdoutMessageTypes);
         Opm::OpmLog::addBackend(stdout_log_id, streamLog);
-
+        // Set a tag limit of 10 (no category limit). Will later in
+        // the run be replaced by calling setupMessageLimiter(), after
+        // the deck is read and the (possibly user-set) category
+        // limits are known.
+        streamLog->setMessageLimiter(std::make_shared<Opm::MessageLimiter>(10));
         bool use_color_coding = OpmLog::stdoutIsTerminal();
         streamLog->setMessageFormatter(std::make_shared<Opm::SimpleMessageFormatter>(use_color_coding));
     }
@@ -548,7 +553,7 @@ void Opm::readDeck(Opm::Parallel::Communication    comm,
 
     if (! parseSuccess) {
         if (comm.rank() == 0) {
-            OpmLog::error("Unrecoverable errors were encountered while loading input");
+            OpmLog::error(fmt::format("Unrecoverable errors while loading input: {}", failureMessage));
         }
 
 #if HAVE_MPI
