@@ -97,6 +97,7 @@ namespace Opm {
             typedef BlackoilModelParametersEbos<TypeTag> ModelParameters;
 
             using Grid = GetPropType<TypeTag, Properties::Grid>;
+            using EquilGrid = GetPropType<TypeTag, Properties::EquilGrid>;
             using FluidSystem = GetPropType<TypeTag, Properties::FluidSystem>;
             using ElementContext = GetPropType<TypeTag, Properties::ElementContext>;
             using Indices = GetPropType<TypeTag, Properties::Indices>;
@@ -112,7 +113,7 @@ namespace Opm {
                 typename BlackoilWellModelGeneric::GLiftWellStateMap;
             using GLiftEclWells = typename GasLiftGroupInfo::GLiftEclWells;
             using GLiftSyncGroups = typename GasLiftSingleWellGeneric::GLiftSyncGroups;
-
+            constexpr static std::size_t pressureVarIndex = GetPropType<TypeTag, Properties::Indices>::pressureSwitchIdx;
             typedef typename BaseAuxiliaryModule<TypeTag>::NeighborSet NeighborSet;
 
             static const int numEq = Indices::numEq;
@@ -127,8 +128,6 @@ namespace Opm {
             typedef Dune::FieldVector<Scalar, numEq    > VectorBlockType;
             typedef Dune::BlockVector<VectorBlockType> BVector;
 
-            typedef Dune::FieldMatrix<Scalar, numEq, numEq > MatrixBlockType;
-
             typedef BlackOilPolymerModule<TypeTag> PolymerModule;
             typedef BlackOilMICPModule<TypeTag> MICPModule;
 
@@ -143,7 +142,7 @@ namespace Opm {
             BlackoilWellModel(Simulator& ebosSimulator);
 
             void init();
-            void initWellContainer() override;
+            void initWellContainer(const int reportStepIdx) override;
 
             /////////////
             // <eWoms auxiliary module stuff>
@@ -210,6 +209,9 @@ namespace Opm {
                 endReportStep();
             }
 
+            void computeTotalRatesForDof(RateVector& rate,
+                                         unsigned globalIdx) const;
+
             template <class Context>
             void computeTotalRatesForDof(RateVector& rate,
                                          const Context& context,
@@ -224,14 +226,14 @@ namespace Opm {
             {
                 initFromRestartFile(restartValues,
                                     this->ebosSimulator_.vanguard().transferWTestState(),
-                                    UgGridHelpers::numCells(grid()),
+                                    grid().size(0),
                                     param_.use_multisegment_well_);
             }
 
             data::Wells wellData() const
             {
                 auto wsrpt = this->wellState()
-                    .report(UgGridHelpers::globalCell(this->grid()),
+                    .report(ebosSimulator_.vanguard().globalCell().data(),
                             [this](const int well_index) -> bool
                 {
                     return this->wasDynamicallyShutThisTimeStep(well_index);
@@ -251,25 +253,18 @@ namespace Opm {
             // subtract B*inv(D)*C * x from A*x
             void apply(const BVector& x, BVector& Ax) const;
 
-#if HAVE_CUDA || HAVE_OPENCL
             // accumulate the contributions of all Wells in the WellContributions object
             void getWellContributions(WellContributions& x) const;
-#endif
 
             // apply well model with scaling of alpha
             void applyScaleAdd(const Scalar alpha, const BVector& x, BVector& Ax) const;
 
             // Check if well equations is converged.
-            ConvergenceReport getWellConvergence(const std::vector<Scalar>& B_avg, const bool checkGroupConvergence = false) const;
+            ConvergenceReport getWellConvergence(const std::vector<Scalar>& B_avg, const bool checkWellGroupControls = false) const;
 
             const SimulatorReportSingle& lastReport() const;
 
-            void addWellContributions(SparseMatrixAdapter& jacobian) const
-            {
-                for ( const auto& well: well_container_ ) {
-                    well->addWellContributions(jacobian);
-                }
-            }
+            void addWellContributions(SparseMatrixAdapter& jacobian) const;
 
             // called at the beginning of a report step
             void beginReportStep(const int time_step);
@@ -285,23 +280,41 @@ namespace Opm {
             // at the beginning of each time step (Not report step)
             void prepareTimeStep(DeferredLogger& deferred_logger);
             void initPrimaryVariablesEvaluation() const;
-            void updateWellControls(DeferredLogger& deferred_logger, const bool checkGroupControls);
+            bool shouldBalanceNetwork(const int reportStepIndex, const int iterationIdx) const;
+            std::tuple<bool, bool, double> updateWellControls(DeferredLogger& deferred_logger);
 
             void updateAndCommunicate(const int reportStepIdx,
                                       const int iterationIdx,
                                       DeferredLogger& deferred_logger);
 
+            bool updateGroupControls(const Group& group,
+                                    DeferredLogger& deferred_logger,
+                                    const int reportStepIdx,
+                                    const int iterationIdx);
+
             WellInterfacePtr getWell(const std::string& well_name) const;
             bool hasWell(const std::string& well_name) const;
 
+            using PressureMatrix = Dune::BCRSMatrix<Opm::MatrixBlock<double, 1, 1>>;
+
+            int numLocalWellsEnd() const;
+
+            void addWellPressureEquations(PressureMatrix& jacobian, const BVector& weights,const bool use_well_weights) const;
+
+            std::vector<std::vector<int>> getMaxWellConnections() const;
+
+            void addWellPressureEquationsStruct(PressureMatrix& jacobian) const;
+
             void initGliftEclWellMap(GLiftEclWells &ecl_well_map);
 
             /// \brief Get list of local nonshut wells
-            const std::vector<WellInterfacePtr>& localNonshutWells()
+            const std::vector<WellInterfacePtr>& localNonshutWells() const
             {
                 return well_container_;
             }
 
+            int numLocalNonshutWells() const;
+
         protected:
             Simulator& ebosSimulator_;
 
@@ -350,6 +363,9 @@ namespace Opm {
             const Grid& grid() const
             { return ebosSimulator_.vanguard().grid(); }
 
+            const EquilGrid& equilGrid() const
+            { return ebosSimulator_.vanguard().equilGrid(); }
+
             const EclipseState& eclState() const
             { return ebosSimulator_.vanguard().eclState(); }
 
@@ -357,6 +373,10 @@ namespace Opm {
             // and in the well equations.
             void assemble(const int iterationIdx,
                           const double dt);
+            bool assembleImpl(const int iterationIdx,
+                              const double dt,
+                              const std::size_t recursion_level,
+                              DeferredLogger& local_deferredLogger);
 
             // called at the end of a time step
             void timeStepSucceeded(const double& simulationTime, const double dt);
