@@ -386,13 +386,6 @@ wellsActive() const
     return wells_active_;
 }
 
-bool
-BlackoilWellModelGeneric::
-localWellsActive() const
-{
-    return numLocalWells() > 0;
-}
-
 bool
 BlackoilWellModelGeneric::
 anyMSWellOpenLocal() const
@@ -639,7 +632,7 @@ initFromRestartFile(const RestartValue& restartValues,
     this->local_parallel_well_info_ = createLocalParallelWellInfo(wells_ecl_);
 
     this->initializeWellProdIndCalculators();
-    this->initializeWellPerfData();
+    initializeWellPerfData();
 
     if (! this->wells_ecl_.empty()) {
         handle_ms_well &= anyMSWellOpenLocal();
@@ -670,13 +663,6 @@ initFromRestartFile(const RestartValue& restartValues,
     initial_step_ = false;
 }
 
-void
-BlackoilWellModelGeneric::
-setWellsActive(const bool wells_active)
-{
-    wells_active_ = wells_active;
-}
-
 std::vector<Well>
 BlackoilWellModelGeneric::
 getLocalWells(const int timeStepIdx) const
@@ -723,53 +709,53 @@ initializeWellPerfData()
     well_perf_data_.resize(wells_ecl_.size());
     int well_index = 0;
     for (const auto& well : wells_ecl_) {
-        int completion_index = 0;
+        int connection_index = 0;
         // INVALID_ECL_INDEX marks no above perf available
-        int completion_index_above = ParallelWellInfo::INVALID_ECL_INDEX;
+        int connection_index_above = ParallelWellInfo::INVALID_ECL_INDEX;
         well_perf_data_[well_index].clear();
         well_perf_data_[well_index].reserve(well.getConnections().size());
         CheckDistributedWellConnections checker(well, local_parallel_well_info_[well_index].get());
-        bool hasFirstPerforation = false;
-        bool firstOpenCompletion = true;
+        bool hasFirstConnection = false;
+        bool firstOpenConnection = true;
         auto& parallelWellInfo = this->local_parallel_well_info_[well_index].get();
         parallelWellInfo.beginReset();
 
-        for (const auto& completion : well.getConnections()) {
-            const int active_index = compressedIndexForInterior(completion.global_index());
-            if (completion.state() == Connection::State::OPEN) {
+        for (const auto& connection : well.getConnections()) {
+            const int active_index = compressedIndexForInterior(connection.global_index());
+            if (connection.state() == Connection::State::OPEN) {
                 if (active_index >= 0) {
-                    if (firstOpenCompletion)
+                    if (firstOpenConnection)
                     {
-                        hasFirstPerforation = true;
+                        hasFirstConnection = true;
                     }
-                    checker.connectionFound(completion_index);
+                    checker.connectionFound(connection_index);
                     PerforationData pd;
                     pd.cell_index = active_index;
-                    pd.connection_transmissibility_factor = completion.CF();
-                    pd.satnum_id = completion.satTableId();
-                    pd.ecl_index = completion_index;
+                    pd.connection_transmissibility_factor = connection.CF();
+                    pd.satnum_id = connection.satTableId();
+                    pd.ecl_index = connection_index;
                     well_perf_data_[well_index].push_back(pd);
-                    parallelWellInfo.pushBackEclIndex(completion_index_above,
-                                                      completion_index);
+                    parallelWellInfo.pushBackEclIndex(connection_index_above,
+                                                      connection_index);
                 }
-                firstOpenCompletion = false;
-                // Next time this index is the one above as each open completion is
+                firstOpenConnection = false;
+                // Next time this index is the one above as each open connection is
                 // is stored somehwere.
-                completion_index_above = completion_index;
+                connection_index_above = connection_index;
             } else {
-                checker.connectionFound(completion_index);
-                if (completion.state() != Connection::State::SHUT) {
+                checker.connectionFound(connection_index);
+                if (connection.state() != Connection::State::SHUT) {
                     OPM_THROW(std::runtime_error,
-                              "Completion state: " << Connection::State2String(completion.state()) << " not handled");
+                              "Connection state: " << Connection::State2String(connection.state()) << " not handled");
                 }
             }
             // Note: we rely on the connections being filtered! I.e. there are only connections
             // to active cells in the global grid.
-            ++completion_index;
+            ++connection_index;
         }
         parallelWellInfo.endReset();
         checker.checkAllConnectionsFound();
-        parallelWellInfo.communicateFirstPerforation(hasFirstPerforation);
+        parallelWellInfo.communicateFirstPerforation(hasFirstConnection);
         ++well_index;
     }
 }
@@ -1010,7 +996,17 @@ checkGroupProductionConstraints(const Group& group,
             // sum over all nodes
             current_rate = comm_.sum(current_rate);
 
-            if (controls.liquid_target < current_rate  ) {
+            bool skip = false;
+            if (controls.liquid_target == controls.oil_target) {
+                double current_water_rate = WellGroupHelpers::sumWellSurfaceRates(group, schedule(), well_state, reportStepIdx, phase_usage_.phase_pos[BlackoilPhases::Aqua], false);
+                current_water_rate = comm_.sum(current_water_rate);
+                if (std::abs(current_water_rate) < 1e-12) {
+                    skip = true;
+                    deferred_logger.debug("LRAT_ORAT_GROUP", "GROUP " + group.name() + " The LRAT target is equal the ORAT target and the water rate is zero, skip checking LRAT");
+                }
+            }
+
+            if (!skip && controls.liquid_target < current_rate ) {
                 double scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.liquid_target / current_rate;
@@ -1172,8 +1168,7 @@ bool
 BlackoilWellModelGeneric::
 checkGroupHigherConstraints(const Group& group,
                             DeferredLogger& deferred_logger,
-                            const int reportStepIdx,
-                            std::set<std::string>& switched_groups)
+                            const int reportStepIdx)
 {
     // Set up coefficients for RESV <-> surface rate conversion.
     // Use the pvtRegionIdx from the top cell of the first well.
@@ -1203,9 +1198,8 @@ checkGroupHigherConstraints(const Group& group,
 
     std::vector<double> rates(phase_usage_.num_phases, 0.0);
 
-    const bool skip = switched_groups.count(group.name()) || group.name() == "FIELD";
-
-    if (!skip && group.isInjectionGroup()) {
+    bool isField = group.name() == "FIELD";
+    if (!isField && group.isInjectionGroup()) {
         // Obtain rates for group.
         std::vector<double> resv_coeff_inj(phase_usage_.num_phases, 0.0);
         calcInjRates(fipnum, pvtreg, resv_coeff_inj);
@@ -1221,33 +1215,33 @@ checkGroupHigherConstraints(const Group& group,
             auto currentControl = this->groupState().injection_control(group.name(), phase);
             if (currentControl != Group::InjectionCMode::FLD && group.injectionGroupControlAvailable(phase)) {
                 const Group& parentGroup = schedule().getGroup(group.parent(), reportStepIdx);
-                const std::pair<bool, double> changed_this = WellGroupHelpers::checkGroupConstraintsInj(
-                    group.name(),
-                    group.parent(),
-                    parentGroup,
-                    this->wellState(),
-                    this->groupState(),
-                    reportStepIdx,
-                    &guideRate_,
-                    rates.data(),
-                    phase,
-                    phase_usage_,
-                    group.getGroupEfficiencyFactor(),
-                    schedule(),
-                    summaryState_,
-                    resv_coeff_inj,
-                    deferred_logger);
-                if (changed_this.first) {
-                    switched_groups.insert(group.name());
+                const auto [is_changed, scaling_factor] = WellGroupHelpers::checkGroupConstraintsInj(
+                group.name(),
+                group.parent(),
+                parentGroup,
+                this->wellState(),
+                this->groupState(),
+                reportStepIdx,
+                &guideRate_,
+                rates.data(),
+                phase,
+                phase_usage_,
+                group.getGroupEfficiencyFactor(),
+                schedule(),
+                summaryState_,
+                resv_coeff_inj,
+                deferred_logger);
+                if (is_changed) {
+                    switched_inj_groups_.insert({ {group.name(), phase}, Group::InjectionCMode2String(Group::InjectionCMode::FLD)});
                     actionOnBrokenConstraints(group, Group::InjectionCMode::FLD, phase, deferred_logger);
-                    WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second, group, schedule(), reportStepIdx, /* isInjector */ true, this->groupState(), this->wellState());
+                    WellGroupHelpers::updateWellRatesFromGroupTargetScale(scaling_factor, group, schedule(), reportStepIdx, /* isInjector */ true, this->groupState(), this->wellState());
                     changed = true;
                 }
             }
         }
     }
 
-    if (!skip && group.isProductionGroup()) {
+    if (!isField && group.isProductionGroup()) {
         // Obtain rates for group.
         for (int phasePos = 0; phasePos < phase_usage_.num_phases; ++phasePos) {
             const double local_current_rate = WellGroupHelpers::sumWellSurfaceRates(group, schedule(), this->wellState(), reportStepIdx, phasePos, /* isInjector */ false);
@@ -1258,37 +1252,32 @@ checkGroupHigherConstraints(const Group& group,
         calcRates(fipnum, pvtreg, resv_coeff);
         // Check higher up only if under individual (not FLD) control.
         const Group::ProductionCMode& currentControl = this->groupState().production_control(group.name());
-            if (currentControl != Group::ProductionCMode::FLD && group.productionGroupControlAvailable()) {
-                const Group& parentGroup = schedule().getGroup(group.parent(), reportStepIdx);
-                const std::pair<bool, double> changed_this = WellGroupHelpers::checkGroupConstraintsProd(
-                    group.name(),
-                    group.parent(),
-                    parentGroup,
-                    this->wellState(),
-                    this->groupState(),
-                    reportStepIdx,
-                    &guideRate_,
-                    rates.data(),
-                    phase_usage_,
-                    group.getGroupEfficiencyFactor(),
-                    schedule(),
-                    summaryState_,
-                    resv_coeff,
-                    deferred_logger);
-                if (changed_this.first) {
-                    switched_groups.insert(group.name());
-                    const auto exceed_action = group.productionControls(summaryState_).exceed_action;
-                    actionOnBrokenConstraints(group, exceed_action, Group::ProductionCMode::FLD, deferred_logger);
-                    WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second, group, schedule(), reportStepIdx, /* isInjector */ false, this->groupState(), this->wellState());
-                    changed = true;
-                }
+        if (currentControl != Group::ProductionCMode::FLD && group.productionGroupControlAvailable()) {
+            const Group& parentGroup = schedule().getGroup(group.parent(), reportStepIdx);
+            const auto [is_changed, scaling_factor] = WellGroupHelpers::checkGroupConstraintsProd(
+                group.name(),
+                group.parent(),
+                parentGroup,
+                this->wellState(),
+                this->groupState(),
+                reportStepIdx,
+                &guideRate_,
+                rates.data(),
+                phase_usage_,
+                group.getGroupEfficiencyFactor(),
+                schedule(),
+                summaryState_,
+                resv_coeff,
+                deferred_logger);
+            if (is_changed) {
+                switched_prod_groups_.insert({group.name(), Group::ProductionCMode2String(Group::ProductionCMode::FLD)});
+                const auto exceed_action = group.productionControls(summaryState_).exceed_action;
+                actionOnBrokenConstraints(group, exceed_action, Group::ProductionCMode::FLD, deferred_logger);
+                WellGroupHelpers::updateWellRatesFromGroupTargetScale(scaling_factor, group, schedule(), reportStepIdx, /* isInjector */ false, this->groupState(), this->wellState());
+                changed = true;
             }
-    }
-
-    // call recursively down the group hiearchy
-    for (const std::string& groupName : group.groups()) {
-        bool changed_this = checkGroupHigherConstraints( schedule().getGroup(groupName, reportStepIdx), deferred_logger, reportStepIdx, switched_groups);
-        changed = changed || changed_this;
+        }
+        
     }
 
     return changed;
@@ -1298,12 +1287,10 @@ bool
 BlackoilWellModelGeneric::
 updateGroupIndividualControl(const Group& group,
                              DeferredLogger& deferred_logger,
-                             const int reportStepIdx,
-                             std::set<std::string>& switched_groups)
+                             const int reportStepIdx)
 {
     bool changed = false;
-    const bool skip = switched_groups.count(group.name());
-    if (!skip && group.isInjectionGroup())
+    if (group.isInjectionGroup())
     {
         const Phase all[] = {Phase::WATER, Phase::OIL, Phase::GAS};
         for (Phase phase : all) {
@@ -1313,62 +1300,28 @@ updateGroupIndividualControl(const Group& group,
             const auto& changed_this = checkGroupInjectionConstraints(group, reportStepIdx, phase);
             if (changed_this.first != Group::InjectionCMode::NONE)
             {
-                switched_groups.insert(group.name());
+                switched_inj_groups_.insert({{group.name(), phase}, Group::InjectionCMode2String(changed_this.first)});
                 actionOnBrokenConstraints(group, changed_this.first, phase, deferred_logger);
                 WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second, group, schedule(), reportStepIdx, /* isInjector */ false, this->groupState(), this->wellState());
                 changed = true;
             }
         }
     }
-    if (!skip && group.isProductionGroup()) {
+    if (group.isProductionGroup()) {
         const auto& changed_this = checkGroupProductionConstraints(group, reportStepIdx, deferred_logger);
         const auto controls = group.productionControls(summaryState_);
         if (changed_this.first != Group::ProductionCMode::NONE)
         {
-            switched_groups.insert(group.name());
+            switched_prod_groups_.insert({group.name(), Group::ProductionCMode2String(changed_this.first)});
             actionOnBrokenConstraints(group, controls.exceed_action, changed_this.first, deferred_logger);
             WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second, group, schedule(), reportStepIdx, /* isInjector */ false, this->groupState(), this->wellState());
             changed = true;
         }
     }
 
-    // call recursively down the group hiearchy
-    for (const std::string& groupName : group.groups()) {
-        bool changed_this = updateGroupIndividualControl( schedule().getGroup(groupName, reportStepIdx), deferred_logger, reportStepIdx, switched_groups);
-        changed = changed || changed_this;
-    }
-
     return changed;
 }
 
-bool
-BlackoilWellModelGeneric::
-updateGroupIndividualControls(DeferredLogger& deferred_logger,
-                              std::set<std::string>& switched_groups,
-                              const int reportStepIdx,
-                              const int iterationIdx)
-{
-    const int nupcol = schedule()[reportStepIdx].nupcol();
-    // don't switch group control when iterationIdx > nupcol
-    // to avoid oscilations between group controls
-    if (iterationIdx > nupcol)
-        return false;
-
-    const Group& fieldGroup = schedule().getGroup("FIELD", reportStepIdx);
-    return updateGroupIndividualControl(fieldGroup, deferred_logger,
-                                        reportStepIdx, switched_groups);
-}
-
-bool
-BlackoilWellModelGeneric::
-updateGroupHigherControls(DeferredLogger& deferred_logger,
-                          const int reportStepIdx,
-                          std::set<std::string>& switched_groups)
-{
-    const Group& fieldGroup = schedule().getGroup("FIELD", reportStepIdx);
-    return checkGroupHigherConstraints(fieldGroup, deferred_logger, reportStepIdx, switched_groups);
-}
-
 void
 BlackoilWellModelGeneric::
 actionOnBrokenConstraints(const Group& group,
@@ -1418,7 +1371,7 @@ actionOnBrokenConstraints(const Group& group,
 
     Parallel::Communication cc = comm_;
     if (!ss.str().empty() && cc.rank() == 0)
-        deferred_logger.info(ss.str());
+        deferred_logger.debug(ss.str());
 }
 
 void
@@ -2111,14 +2064,14 @@ inferLocalShutWells()
     }
 }
 
-void
+std::pair<bool, double>
 BlackoilWellModelGeneric::
 updateNetworkPressures(const int reportStepIdx)
 {
     // Get the network and return if inactive.
     const auto& network = schedule()[reportStepIdx].network();
     if (!network.active()) {
-        return;
+        return { false, 0.0 };
     }
     node_pressures_ = WellGroupHelpers::computeNetworkPressures(network,
                                                                 this->wellState(),
@@ -2128,6 +2081,8 @@ updateNetworkPressures(const int reportStepIdx)
                                                                 reportStepIdx);
 
     // Set the thp limits of wells
+    bool active_limit_change = false;
+    double network_imbalance = 0.0;
     for (auto& well : well_container_generic_) {
         // Producers only, since we so far only support the
         // "extended" network model (properties defined by
@@ -2137,20 +2092,25 @@ updateNetworkPressures(const int reportStepIdx)
             if (it != node_pressures_.end()) {
                 // The well belongs to a group with has a network pressure constraint,
                 // set the dynamic THP constraint of the well accordingly.
-                well->setDynamicThpLimit(it->second);
+                const double new_limit = it->second;
+                well->setDynamicThpLimit(new_limit);
+                const SingleWellState& ws = this->wellState()[well->indexOfWell()];
+                const bool thp_is_limit = ws.production_cmode == Well::ProducerCMode::THP;
+                const bool will_switch_to_thp = ws.thp < new_limit;
+                if (thp_is_limit || will_switch_to_thp) {
+                    active_limit_change = true;
+                    network_imbalance = std::max(network_imbalance, std::fabs(new_limit - ws.thp));
+                }
             }
         }
     }
+    return { active_limit_change, network_imbalance };
 }
 
 void
 BlackoilWellModelGeneric::
 calculateEfficiencyFactors(const int reportStepIdx)
 {
-    if ( !localWellsActive() ) {
-        return;
-    }
-
     for (auto& well : well_container_generic_) {
         const Well& wellEcl = well->wellEcl();
         double well_efficiency_factor = wellEcl.getEfficiencyFactor();
@@ -2358,7 +2318,7 @@ runWellPIScaling(const int timeStepIdx,
         this->createWellContainer(timeStepIdx);
         this->inferLocalShutWells();
 
-        this->initWellContainer();
+        this->initWellContainer(timeStepIdx);
 
         this->calculateProductivityIndexValues(local_deferredLogger);
         this->calculateProductivityIndexValuesShutWells(timeStepIdx, local_deferredLogger);
@@ -2390,14 +2350,14 @@ guideRateUpdateIsNeeded(const int reportStepIdx) const {
     });
     if (!need_update && this->report_step_starts_) {
         const auto& events = this->schedule()[reportStepIdx].wellgroup_events();
-        const auto effective_events_mask = ScheduleEvents::WELL_STATUS_CHANGE
+        constexpr auto effective_events_mask = ScheduleEvents::WELL_STATUS_CHANGE
             + ScheduleEvents::INJECTION_TYPE_CHANGED
             + ScheduleEvents::WELL_SWITCHED_INJECTOR_PRODUCER
             + ScheduleEvents::NEW_WELL;
 
         need_update = std::any_of(this->well_container_generic_.begin(),
                               this->well_container_generic_.end(),
-            [&events, effective_events_mask](const WellInterfaceGeneric* well)
+            [&events](const WellInterfaceGeneric* well)
         {
             return events.hasEvent(well->name(), effective_events_mask);
         });
