@@ -90,8 +90,7 @@ public:
     /// return true if wells are available in the reservoir
     bool wellsActive() const;
     bool hasWell(const std::string& wname);
-    /// return true if wells are available on this process
-    bool localWellsActive() const;
+
     // whether there exists any multisegment well open on this process
     bool anyMSWellOpenLocal() const;
 
@@ -143,8 +142,6 @@ public:
                              const size_t numCells,
                              bool handle_ms_well);
 
-    void setWellsActive(const bool wells_active);
-
     /*
       Will assign the internal member last_valid_well_state_ to the
       current value of the this->active_well_state_. The state stored
@@ -253,7 +250,7 @@ protected:
 
     bool wasDynamicallyShutThisTimeStep(const int well_index) const;
 
-    void updateNetworkPressures(const int reportStepIdx);
+    std::pair<bool, double> updateNetworkPressures(const int reportStepIdx);
 
     void updateWsolvent(const Group& group,
                         const int reportStepIdx,
@@ -340,22 +337,11 @@ protected:
 
     bool checkGroupHigherConstraints(const Group& group,
                                      DeferredLogger& deferred_logger,
-                                     const int reportStepIdx,
-                                     std::set<std::string>& switched_groups);
+                                     const int reportStepIdx);
 
     bool updateGroupIndividualControl(const Group& group,
                                       DeferredLogger& deferred_logger,
-                                      const int reportStepIdx,
-                                      std::set<std::string>& switched_groups);
-
-    bool updateGroupIndividualControls(DeferredLogger& deferred_logger,
-                                       std::set<std::string>& switched_groups,
-                                       const int reportStepIdx,
-                                       const int iterationIdx);
-
-    bool updateGroupHigherControls(DeferredLogger& deferred_logger,
-                                   const int reportStepIdx,
-                                   std::set<std::string>& switched_groups);
+                                      const int reportStepIdx);
 
     void actionOnBrokenConstraints(const Group& group,
                                    const Group::ExceedAction& exceed_action,
@@ -400,7 +386,7 @@ protected:
 
     // create the well container
     virtual void createWellContainer(const int time_step) = 0;
-    virtual void initWellContainer() = 0;
+    virtual void initWellContainer(const int reportStepIdx) = 0;
 
     virtual void calculateProductivityIndexValuesShutWells(const int reportStepIdx,
                                                            DeferredLogger& deferred_logger) = 0;
@@ -462,6 +448,10 @@ protected:
 
     double last_glift_opt_time_ = -1.0;
 
+    std::map<std::string, std::string> switched_prod_groups_;
+    std::map<std::pair<std::string, Opm::Phase>, std::string> switched_inj_groups_;
+
+
 private:
     WellInterfaceGeneric* getGenWell(const std::string& well_name);
 };
