@@ -22,10 +22,16 @@
 
 #include <opm/simulators/utils/DeferredLoggingErrorHelpers.hpp>
 #include <opm/core/props/phaseUsageFromDeck.hpp>
-
+#include <opm/grid/utility/cartesianToCompressed.hpp>
 #include <opm/input/eclipse/Units/UnitSystem.hpp>
 
 #include <opm/simulators/wells/VFPProperties.hpp>
+#include <opm/simulators/utils/MPIPacker.hpp>
+#include <opm/simulators/linalg/bda/WellContributions.hpp>
+
+#if HAVE_MPI
+#include <ebos/eclmpiserializer.hh>
+#endif
 
 #include <algorithm>
 #include <utility>
@@ -51,14 +57,13 @@ namespace Opm {
         // Number of cells the global grid view
         global_num_cells_ = ebosSimulator_.vanguard().globalNumCells();
 
-        // Set up parallel wells
-        auto& parallel_wells = ebosSimulator.vanguard().parallelWells();
-
-        this->parallel_well_info_.reserve(parallel_wells.size());
-        for( const auto& name_bool: parallel_wells)
         {
-            this->parallel_well_info_.emplace_back(name_bool,
-                                                   ebosSimulator_.gridView().comm());
+            auto& parallel_wells = ebosSimulator.vanguard().parallelWells();
+
+            this->parallel_well_info_.reserve(parallel_wells.size());
+            for( const auto& name_bool : parallel_wells) {
+                this->parallel_well_info_.emplace_back(name_bool, grid().comm());
+            }
         }
 
         this->alternative_well_rate_init_ =
@@ -94,11 +99,15 @@ namespace Opm {
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
-    initWellContainer()
+    initWellContainer(const int reportStepIdx)
     {
+        const uint64_t effective_events_mask = ScheduleEvents::WELL_STATUS_CHANGE
+                        + ScheduleEvents::NEW_WELL;
+        const auto& events = schedule()[reportStepIdx].wellgroup_events();
         for (auto& wellPtr : this->well_container_) {
+            const bool well_opened_this_step = report_step_starts_ && events.hasEvent(wellPtr->name(), effective_events_mask);
             wellPtr->init(&this->phase_usage_, this->depth_, this->gravity_,
-                          this->local_num_cells_, this->B_avg_);
+                          this->local_num_cells_, this->B_avg_, well_opened_this_step);
         }
     }
 
@@ -126,7 +135,6 @@ namespace Opm {
             {
                 const auto& connection = connectionSet.get(c);
                 int compressed_idx = compressedIndexForInterior(connection.global_index());
-
                 if ( compressed_idx >= 0 ) { // Ignore connections in inactive/remote cells.
                     wellCells.push_back(compressed_idx);
                 }
@@ -196,11 +204,6 @@ namespace Opm {
             this->initializeWellPerfData();
             this->initializeWellState(timeStepIdx, summaryState);
 
-            // Wells are active if they are active wells on at least
-            // one process.
-            wells_active_ = localWellsActive() ? 1 : 0;
-            wells_active_ = grid.comm().max(wells_active_);
-
             // handling MS well related
             if (param_.use_multisegment_well_&& anyMSWellOpenLocal()) { // if we use MultisegmentWell model
                 this->wellState().initWellStateMSWell(wells_ecl_, &this->prevWellState());
@@ -224,7 +227,7 @@ namespace Opm {
             {
                 const auto& sched_state = this->schedule()[timeStepIdx];
                 // update VFP properties
-                vfp_properties_.reset(new VFPProperties( sched_state.vfpinj(), sched_state.vfpprod()) );
+                vfp_properties_.reset(new VFPProperties( sched_state.vfpinj(), sched_state.vfpprod(), this->prevWellState()));
                 this->initializeWellProdIndCalculators();
                 if (sched_state.events().hasEvent(ScheduleEvents::Events::WELL_PRODUCTIVITY_INDEX)) {
                     this->runWellPIScaling(timeStepIdx, local_deferredLogger);
@@ -247,6 +250,8 @@ namespace Opm {
         updatePerforationIntensiveQuantities();
         updateAverageFormationFactor();
         DeferredLogger local_deferredLogger;
+        switched_prod_groups_.clear();
+        switched_inj_groups_.clear();
 
         this->resetWGState();
         const int reportStepIdx = ebosSimulator_.episodeIndex();
@@ -262,10 +267,15 @@ namespace Opm {
             // create the well container
             createWellContainer(reportStepIdx);
 
+            // Wells are active if they are active wells on at least one process.
+            const Grid& grid = ebosSimulator_.vanguard().grid();
+            wells_active_ = !this->well_container_.empty();
+            wells_active_ = grid.comm().max(wells_active_);
+
             // do the initialization for all the wells
             // TODO: to see whether we can postpone of the intialization of the well containers to
             // optimize the usage of the following several member variables
-            this->initWellContainer();
+            this->initWellContainer(reportStepIdx);
 
             // update the updated cell flag
             std::fill(is_cell_perforated_.begin(), is_cell_perforated_.end(), false);
@@ -392,7 +402,7 @@ namespace Opm {
 
                 WellInterfacePtr well = createWellForWellTest(well_name, timeStepIdx, deferred_logger);
                 // some preparation before the well can be used
-                well->init(&phase_usage_, depth_, gravity_, local_num_cells_, B_avg_);
+                well->init(&phase_usage_, depth_, gravity_, local_num_cells_, B_avg_, true);
 
                 double well_efficiency_factor = wellEcl.getEfficiencyFactor();
                 WellGroupHelpers::accumulateGroupEfficiencyFactor(schedule().getGroup(wellEcl.groupName(), timeStepIdx),
@@ -460,6 +470,35 @@ namespace Opm {
         for (const auto& well : well_container_) {
             well->reportWellSwitching(this->wellState().well(well->indexOfWell()), local_deferredLogger);
         }
+        // report group switching
+        if (terminal_output_) {
+
+            for (const auto& [name, to] : switched_prod_groups_) {
+                const Group::ProductionCMode& oldControl = this->prevWGState().group_state.production_control(name);
+                std::string from = Group::ProductionCMode2String(oldControl);
+                if (to != from) {
+                    std::string msg = "    Production Group " + name
+                    + " control mode changed from ";
+                    msg += from;
+                    msg += " to " + to;
+                    local_deferredLogger.info(msg);
+                }
+            }
+            for (const auto& [key, to] : switched_inj_groups_) {
+                const std::string& name = key.first;
+                const Opm::Phase& phase = key.second;
+
+                const Group::InjectionCMode& oldControl = this->prevWGState().group_state.injection_control(name, phase);
+                std::string from = Group::InjectionCMode2String(oldControl);
+                if (to != from) {
+                    std::string msg = "    Injection Group " + name
+                    + " control mode changed from ";
+                    msg += from;
+                    msg += " to " + to;
+                    local_deferredLogger.info(msg);
+                }
+            }
+        }
 
         // update the rate converter with current averages pressures etc in
         rateConverter_->template defineState<ElementContext>(ebosSimulator_);
@@ -497,6 +536,22 @@ namespace Opm {
     }
 
 
+    template<typename TypeTag>
+    void
+    BlackoilWellModel<TypeTag>::
+    computeTotalRatesForDof(RateVector& rate,
+                            unsigned elemIdx) const
+    {
+        rate = 0;
+
+        if (!is_cell_perforated_[elemIdx])
+            return;
+
+        for (const auto& well : well_container_)
+            well->addCellRates(rate, elemIdx);
+    }
+
+
     template<typename TypeTag>
     template <class Context>
     void
@@ -528,18 +583,10 @@ namespace Opm {
         ElementContext elemCtx(ebosSimulator_);
 
         const auto& gridView = ebosSimulator_.vanguard().gridView();
-        const auto& elemEndIt = gridView.template end</*codim=*/0>();
+
         OPM_BEGIN_PARALLEL_TRY_CATCH();
-
-        for (auto elemIt = gridView.template begin</*codim=*/0>();
-             elemIt != elemEndIt;
-             ++elemIt)
-        {
-            if (elemIt->partitionType() != Dune::InteriorEntity) {
-                continue;
-            }
-
-            elemCtx.updatePrimaryStencil(*elemIt);
+        for (const auto& elem : elements(gridView, Dune::Partitions::interior)) {
+            elemCtx.updatePrimaryStencil(elem);
             elemCtx.updatePrimaryIntensiveQuantities(/*timeIdx=*/0);
 
             const auto& fs = elemCtx.intensiveQuantities(/*spaceIdx=*/0, /*timeIdx=*/0).fluidState();
@@ -631,6 +678,12 @@ namespace Opm {
                         this->wellState().shutWell(w);
                         continue;
                     } else {
+                        if (!well_ecl.getAllowCrossFlow()) {
+                            // stopped wells where cross flow is not allowed
+                            // are not added to the well container
+                            this->wellState().shutWell(w);
+                            continue;
+                        }
                         // stopped wells are added to the container but marked as stopped
                         this->wellState().stopWell(w);
                         wellIsStopped = true;
@@ -669,7 +722,6 @@ namespace Opm {
                     case Well::ProducerCMode::RESV:
                         zero_rate_control = is_zero(prod_controls.resv_rate);
                         break;
-
                     default:
                         // Might still have zero rate controls, but is pressure controlled.
                         zero_rate_control = false;
@@ -812,7 +864,6 @@ namespace Opm {
             return;
         }
 
-
         updatePerforationIntensiveQuantities();
 
         if (iterationIdx == 0) {
@@ -828,7 +879,29 @@ namespace Opm {
             OPM_END_PARALLEL_TRY_CATCH_LOG(local_deferredLogger, "assemble() failed (It=0): ",
                                                terminal_output_, grid().comm());
         }
-        updateWellControls(local_deferredLogger, /* check group controls */ true);
+
+        const bool well_group_control_changed = assembleImpl(iterationIdx, dt, 0, local_deferredLogger);
+
+        // if group or well control changes we don't consider the
+        // case converged
+        last_report_.well_group_control_changed = well_group_control_changed;
+        last_report_.assemble_time_well += perfTimer.stop();
+    }
+
+
+
+
+
+    template<typename TypeTag>
+    bool
+    BlackoilWellModel<TypeTag>::
+    assembleImpl(const int iterationIdx,
+                 const double dt,
+                 const std::size_t recursion_level,
+                 DeferredLogger& local_deferredLogger)
+    {
+
+        auto [well_group_control_changed, network_changed, network_imbalance] = updateWellControls(local_deferredLogger);
 
         bool alq_updated = false;
         OPM_BEGIN_PARALLEL_TRY_CATCH();
@@ -853,10 +926,24 @@ namespace Opm {
             WellGroupHelpers::updateGuideRates(fieldGroup, schedule(), summaryState, this->phase_usage_, reportStepIdx, simulationTime,
                                         this->wellState(), this->groupState(), comm, &this->guideRate_, pot, local_deferredLogger);
         }
-        last_report_.converged = true;
-        last_report_.assemble_time_well += perfTimer.stop();
+
+        // Maybe do a recursive call to iterate network and well controls.
+        if (network_changed) {
+            if (shouldBalanceNetwork(reportStepIdx, iterationIdx)) {
+                const auto& balance = schedule()[reportStepIdx].network_balance();
+                // Iterate if not converged, and number of iterations is not yet max (NETBALAN item 3).
+                if (recursion_level < balance.pressure_max_iter() && network_imbalance > balance.pressure_tolerance()) {
+                    well_group_control_changed = assembleImpl(iterationIdx, dt, recursion_level + 1, local_deferredLogger);
+                }
+            }
+        }
+        return well_group_control_changed;
     }
 
+
+
+
+
     template<typename TypeTag>
     bool
     BlackoilWellModel<TypeTag>::
@@ -1003,8 +1090,11 @@ namespace Opm {
                 //   data if they are going to check the group rates in stage1
                 //   Another similar idea is to only communicate the rates to
                 //   process j = i + 1
-                Mpi::broadcast(comm, i, group_indexes, group_oil_rates,
-                               group_gas_rates, group_water_rates, group_alq_rates);
+#if HAVE_MPI
+                EclMpiSerializer ser(comm);
+                ser.broadcast(i, group_indexes, group_oil_rates,
+                              group_gas_rates, group_water_rates, group_alq_rates);
+#endif
                 if (comm.rank() != i) {
                     for (int j=0; j<num_rates_to_sync; j++) {
                         group_info.updateRate(group_indexes[j],
@@ -1081,10 +1171,6 @@ namespace Opm {
     BlackoilWellModel<TypeTag>::
     apply( BVector& r) const
     {
-        if ( ! localWellsActive() ) {
-            return;
-        }
-
         for (auto& well : well_container_) {
             well->apply(r);
         }
@@ -1097,17 +1183,11 @@ namespace Opm {
     BlackoilWellModel<TypeTag>::
     apply(const BVector& x, BVector& Ax) const
     {
-        // TODO: do we still need localWellsActive()?
-        if ( ! localWellsActive() ) {
-            return;
-        }
-
         for (auto& well : well_container_) {
             well->apply(x, Ax);
         }
     }
 
-#if HAVE_CUDA || HAVE_OPENCL
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
@@ -1145,7 +1225,6 @@ namespace Opm {
             }
         }
     }
-#endif
 
     // Ax = Ax - alpha * C D^-1 B x
     template<typename TypeTag>
@@ -1153,7 +1232,7 @@ namespace Opm {
     BlackoilWellModel<TypeTag>::
     applyScaleAdd(const Scalar alpha, const BVector& x, BVector& Ax) const
     {
-        if ( ! localWellsActive() ) {
+        if (this->well_container_.empty()) {
             return;
         }
 
@@ -1168,10 +1247,110 @@ namespace Opm {
         Ax.axpy( alpha, scaleAddRes_ );
     }
 
+    template<typename TypeTag>
+    void
+    BlackoilWellModel<TypeTag>::
+    addWellContributions(SparseMatrixAdapter& jacobian) const
+    {
+        for ( const auto& well: well_container_ ) {
+            well->addWellContributions(jacobian);
+        }
+    }
 
+    template<typename TypeTag>
+    void
+    BlackoilWellModel<TypeTag>::
+    addWellPressureEquations(PressureMatrix& jacobian, const BVector& weights,const bool use_well_weights) const
+    {
+        int nw =  this->numLocalWellsEnd();
+        int rdofs = local_num_cells_;
+        for ( int i = 0; i < nw; i++ ){
+            int wdof = rdofs + i; 
+            jacobian[wdof][wdof] = 1.0;// better scaling ?
+        }
 
+        for ( const auto& well : well_container_ ) {
+            well->addWellPressureEquations(jacobian, weights, pressureVarIndex, use_well_weights, this->wellState());
+        }
+    }
 
+    template<typename TypeTag>
+    int
+    BlackoilWellModel<TypeTag>::
+    numLocalWellsEnd() const
+    {
+        auto w = schedule().getWellsatEnd();
+        w.erase(std::remove_if(w.begin(), w.end(), not_on_process_), w.end());
+        return w.size();
+    }
 
+    template<typename TypeTag>
+    std::vector<std::vector<int>>
+    BlackoilWellModel<TypeTag>::
+    getMaxWellConnections() const
+    {
+        std::vector<std::vector<int>> wells;
+
+        auto schedule_wells = schedule().getWellsatEnd();
+        schedule_wells.erase(std::remove_if(schedule_wells.begin(), schedule_wells.end(), not_on_process_), schedule_wells.end());
+        wells.reserve(schedule_wells.size());
+
+        // initialize the additional cell connections introduced by wells.
+        for ( const auto& well : schedule_wells )
+        {
+            std::vector<int> compressed_well_perforations;
+            // All possible completions of the well
+            const auto& completionSet = well.getConnections();
+            compressed_well_perforations.reserve(completionSet.size());
+
+            for (const auto& connection: well.getConnections())
+            {
+                const int compressed_idx = compressedIndexForInterior(connection.global_index());
+                if ( compressed_idx >= 0 ) // Ignore completions in inactive/remote cells.
+                {
+                    compressed_well_perforations.push_back(compressed_idx);
+                }
+            }
+
+            // also include wells with no perforations in case
+            std::sort(compressed_well_perforations.begin(),
+                      compressed_well_perforations.end());
+
+            wells.push_back(compressed_well_perforations);
+        }
+        return wells;
+    }
+
+    template<typename TypeTag>
+    void
+    BlackoilWellModel<TypeTag>::
+    addWellPressureEquationsStruct(PressureMatrix& jacobian) const
+    {
+        int nw =  this->numLocalWellsEnd();
+        int rdofs = local_num_cells_;
+        for(int i=0; i < nw; i++){
+            int wdof = rdofs + i; 
+            jacobian.entry(wdof,wdof) = 1.0;// better scaling ?
+        }
+        std::vector<std::vector<int>> wellconnections = getMaxWellConnections();
+        for(int i=0; i < nw; i++){
+            const auto& perfcells = wellconnections[i];
+            for(int perfcell : perfcells){
+                int wdof = rdofs + i; 
+                jacobian.entry(wdof,perfcell) = 0.0;
+                jacobian.entry(perfcell, wdof) = 0.0;
+            }
+        }
+    }
+
+    template<typename TypeTag>
+    int
+    BlackoilWellModel<TypeTag>::
+    numLocalNonshutWells() const
+    {
+        return well_container_.size();
+    }
+    
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
@@ -1181,10 +1360,8 @@ namespace Opm {
         DeferredLogger local_deferredLogger;
         OPM_BEGIN_PARALLEL_TRY_CATCH();
         {
-            if (localWellsActive()) {
-                for (auto& well : well_container_) {
-                    well->recoverWellSolutionAndUpdateWellState(x, this->wellState(), local_deferredLogger);
-                }
+            for (auto& well : well_container_) {
+                well->recoverWellSolutionAndUpdateWellState(x, this->wellState(), local_deferredLogger);
             }
 
         }
@@ -1215,7 +1392,7 @@ namespace Opm {
     template<typename TypeTag>
     ConvergenceReport
     BlackoilWellModel<TypeTag>::
-    getWellConvergence(const std::vector<Scalar>& B_avg, bool checkGroupConvergence) const
+    getWellConvergence(const std::vector<Scalar>& B_avg, bool checkWellGroupControls) const
     {
 
         DeferredLogger local_deferredLogger;
@@ -1232,15 +1409,19 @@ namespace Opm {
                 local_report += report;
             }
         }
-        
+
         const Opm::Parallel::Communication comm = grid().comm();
         DeferredLogger global_deferredLogger = gatherDeferredLogger(local_deferredLogger, comm);
+        ConvergenceReport report = gatherConvergenceReport(local_report, comm);
+
+        // the well_group_control_changed info is already communicated
+        if (checkWellGroupControls) {
+            report.setWellGroupTargetsViolated(this->lastReport().well_group_control_changed);
+        }
+
         if (terminal_output_) {
             global_deferredLogger.logMessages();
         }
-
-        ConvergenceReport report = gatherConvergenceReport(local_report, comm);
-
         // Log debug messages for NaN or too large residuals.
         if (terminal_output_) {
             for (const auto& f : report.wellFailures()) {
@@ -1251,15 +1432,6 @@ namespace Opm {
                 }
             }
         }
-
-        if (checkGroupConvergence) {
-            const int reportStepIdx = ebosSimulator_.episodeIndex();
-            const Group& fieldGroup = schedule().getGroup("FIELD", reportStepIdx);
-            bool violated = checkGroupConstraints(fieldGroup,
-                                                  ebosSimulator_.episodeIndex(),
-                                                  global_deferredLogger);
-            report.setGroupConverged(!violated);
-        }
         return report;
     }
 
@@ -1283,62 +1455,78 @@ namespace Opm {
 
 
     template<typename TypeTag>
-    void
+    bool
     BlackoilWellModel<TypeTag>::
-    updateWellControls(DeferredLogger& deferred_logger, const bool checkGroupControls)
+    shouldBalanceNetwork(const int reportStepIdx, const int iterationIdx) const
+    {
+        const auto& balance = schedule()[reportStepIdx].network_balance();
+        if (balance.mode() == Network::Balance::CalcMode::TimeStepStart) {
+            return iterationIdx == 0;
+        } else if (balance.mode() == Network::Balance::CalcMode::NUPCOL) {
+            const int nupcol = schedule()[reportStepIdx].nupcol();
+            return iterationIdx < nupcol;
+        } else {
+            // We do not support any other rebalancing modes,
+            // i.e. TimeInterval based rebalancing is not available.
+            // This should be warned about elsewhere, so we choose to
+            // avoid spamming with a warning here.
+            return false;
+        }
+    }
+
+
+
+
+
+    template<typename TypeTag>
+    std::tuple<bool, bool, double>
+    BlackoilWellModel<TypeTag>::
+    updateWellControls(DeferredLogger& deferred_logger)
     {
         // Even if there are no wells active locally, we cannot
         // return as the DeferredLogger uses global communication.
         // For no well active globally we simply return.
-        if( !wellsActive() ) return ;
+        if( !wellsActive() ) return { false, false, 0.0 };
 
         const int episodeIdx = ebosSimulator_.episodeIndex();
         const int iterationIdx = ebosSimulator_.model().newtonMethod().numIterations();
         const auto& comm = ebosSimulator_.vanguard().grid().comm();
         updateAndCommunicateGroupData(episodeIdx, iterationIdx);
 
-        updateNetworkPressures(episodeIdx);
+        const auto [local_network_changed, local_network_imbalance]
+            = shouldBalanceNetwork(episodeIdx, iterationIdx) ?
+            updateNetworkPressures(episodeIdx) : std::make_pair(false, 0.0);
+        const bool network_changed = comm.sum(local_network_changed);
+        const double network_imbalance = comm.max(local_network_imbalance);
 
-        std::set<std::string> switched_wells;
-        std::set<std::string> switched_groups;
-
-        if (checkGroupControls) {
-            // Check group individual constraints.
-            bool changed_individual = updateGroupIndividualControls(deferred_logger, switched_groups,
-                                                        episodeIdx, iterationIdx);
-
-            if (changed_individual)
-                updateAndCommunicate(episodeIdx, iterationIdx, deferred_logger);
-
-            // Check group's constraints from higher levels.
-            bool changed_higher = updateGroupHigherControls(deferred_logger,
-                                                            episodeIdx,
-                                                            switched_groups);
-
-            if (changed_higher)
-                updateAndCommunicate(episodeIdx, iterationIdx, deferred_logger);
-
-            // Check wells' group constraints and communicate.
-            bool changed_well_group = false;
-            for (const auto& well : well_container_) {
-                const auto mode = WellInterface<TypeTag>::IndividualOrGroup::Group;
-                const bool changed_well = well->updateWellControl(ebosSimulator_, mode, this->wellState(), this->groupState(), deferred_logger);
-                if (changed_well) {
-                    switched_wells.insert(well->name());
-                    changed_well_group = changed_well || changed_well_group;
-                }
+        bool changed_well_group = false;
+        // Check group individual constraints.
+        const int nupcol = schedule()[episodeIdx].nupcol();
+        // don't switch group control when iterationIdx > nupcol
+        // to avoid oscilations between group controls
+        if (iterationIdx <= nupcol) {
+            const Group& fieldGroup = schedule().getGroup("FIELD", episodeIdx);
+            changed_well_group = updateGroupControls(fieldGroup, deferred_logger, episodeIdx, iterationIdx);
+        }
+        // Check wells' group constraints and communicate.
+        bool changed_well_to_group = false;
+        for (const auto& well : well_container_) {
+            const auto mode = WellInterface<TypeTag>::IndividualOrGroup::Group;
+            const bool changed_well = well->updateWellControl(ebosSimulator_, mode, this->wellState(), this->groupState(), deferred_logger);
+            if (changed_well) {
+                changed_well_to_group = changed_well || changed_well_to_group;
             }
-            changed_well_group = comm.sum(changed_well_group);
-            if (changed_well_group)
-                updateAndCommunicate(episodeIdx, iterationIdx, deferred_logger);
+        }
+
+        changed_well_to_group = comm.sum(changed_well_to_group);
+        if (changed_well_to_group) {
+            updateAndCommunicate(episodeIdx, iterationIdx, deferred_logger);
+            changed_well_group = true;
         }
 
         // Check individual well constraints and communicate.
         bool changed_well_individual = false;
         for (const auto& well : well_container_) {
-            if (switched_wells.count(well->name())) {
-                continue;
-            }
             const auto mode = WellInterface<TypeTag>::IndividualOrGroup::Individual;
             const bool changed_well = well->updateWellControl(ebosSimulator_, mode, this->wellState(), this->groupState(), deferred_logger);
             if (changed_well) {
@@ -1346,12 +1534,16 @@ namespace Opm {
             }
         }
         changed_well_individual = comm.sum(changed_well_individual);
-        if (changed_well_individual)
+        if (changed_well_individual) {
             updateAndCommunicate(episodeIdx, iterationIdx, deferred_logger);
+            changed_well_group = true;
+        }
 
         // update wsolvent fraction for REIN wells
         const Group& fieldGroup = schedule().getGroup("FIELD", episodeIdx);
         updateWsolvent(fieldGroup, episodeIdx,  this->nupcolWellState());
+
+        return { changed_well_group, network_changed, network_imbalance };
     }
 
 
@@ -1370,6 +1562,33 @@ namespace Opm {
         updateAndCommunicateGroupData(reportStepIdx, iterationIdx);
     }
 
+    template<typename TypeTag>
+    bool
+    BlackoilWellModel<TypeTag>::
+    updateGroupControls(const Group& group,
+                        DeferredLogger& deferred_logger,
+                        const int reportStepIdx,
+                        const int iterationIdx)
+    {
+        bool changed = false;
+        bool changed_hc = checkGroupHigherConstraints( group, deferred_logger, reportStepIdx);
+        if (changed_hc) {
+            changed = true;
+            updateAndCommunicate(reportStepIdx, iterationIdx, deferred_logger);
+        }
+        bool changed_individual =  updateGroupIndividualControl( group, deferred_logger, reportStepIdx);
+        if (changed_individual) {
+            changed = true;
+            updateAndCommunicate(reportStepIdx, iterationIdx, deferred_logger);
+        }
+        // call recursively down the group hierarchy
+        for (const std::string& groupName : group.groups()) {
+            bool changed_this = updateGroupControls( schedule().getGroup(groupName, reportStepIdx), deferred_logger, reportStepIdx,iterationIdx);
+            changed = changed || changed_this;
+        }
+        return changed;
+    }
+
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
@@ -1458,7 +1677,7 @@ namespace Opm {
                 <StandardWell<TypeTag>>(shutWell, reportStepIdx);
 
             wellPtr->init(&this->phase_usage_, this->depth_, this->gravity_,
-                          this->local_num_cells_, this->B_avg_);
+                          this->local_num_cells_, this->B_avg_, true);
 
             this->calculateProductivityIndexValues(wellPtr.get(), deferred_logger);
         }
@@ -1493,6 +1712,8 @@ namespace Opm {
             auto& events = this->wellState().well(well->indexOfWell()).events;
             if (events.hasEvent(WellState::event_mask)) {
                 well->updateWellStateWithTarget(ebosSimulator_, this->groupState(), this->wellState(), deferred_logger);
+                well->updatePrimaryVariables(this->wellState(), deferred_logger);
+                well->initPrimaryVariablesEvaluation();
                 // There is no new well control change input within a report step,
                 // so next time step, the well does not consider to have effective events anymore.
                 events.clearEvent(WellState::event_mask);
@@ -1510,9 +1731,6 @@ namespace Opm {
         updatePrimaryVariables(deferred_logger);
     }
 
-
-
-
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
@@ -1522,13 +1740,10 @@ namespace Opm {
         const auto& grid = ebosSimulator_.vanguard().grid();
         const auto& gridView = grid.leafGridView();
         ElementContext elemCtx(ebosSimulator_);
-        const auto& elemEndIt = gridView.template end</*codim=*/0, Dune::Interior_Partition>();
-        OPM_BEGIN_PARALLEL_TRY_CATCH();
 
-        for (auto elemIt = gridView.template begin</*codim=*/0, Dune::Interior_Partition>();
-             elemIt != elemEndIt; ++elemIt)
-        {
-            elemCtx.updatePrimaryStencil(*elemIt);
+        OPM_BEGIN_PARALLEL_TRY_CATCH();
+        for (const auto& elem : elements(gridView, Dune::Partitions::interior)) {
+            elemCtx.updatePrimaryStencil(elem);
             elemCtx.updatePrimaryIntensiveQuantities(/*timeIdx=*/0);
 
             const auto& intQuants = elemCtx.intensiveQuantities(/*spaceIdx=*/0, /*timeIdx=*/0);
@@ -1554,9 +1769,9 @@ namespace Opm {
 
         // compute global average
         grid.comm().sum(B_avg.data(), B_avg.size());
-        for(auto& bval: B_avg)
+        for (auto& bval : B_avg)
         {
-            bval/=global_num_cells_;
+            bval /= global_num_cells_;
         }
         B_avg_ = B_avg;
     }
@@ -1595,7 +1810,13 @@ namespace Opm {
     int
     BlackoilWellModel<TypeTag>::numComponents() const
     {
-        if (wellsActive()  && numPhases() < 3) {
+        // The numComponents here does not reflect the actual number of the components in the system.
+        // It more or less reflects the number of mass conservation equations for the well equations.
+        // For example, in the current formulation, we do not have the polymer conservation equation
+        // in the well equations. As a result, for an oil-water-polymer system, this function will return 2.
+        // In some way, it makes this function appear to be confusing from its name, and we need
+        // to revisit/revise this function again when extending the variants of system that flow can simulate.
+        if (numPhases() < 3) {
             return numPhases();
         }
         int numComp = FluidSystem::numComponents;
@@ -1610,29 +1831,24 @@ namespace Opm {
     void
     BlackoilWellModel<TypeTag>::extractLegacyDepth_()
     {
-        const auto& grid = ebosSimulator_.vanguard().grid();
-        const unsigned numCells = grid.size(/*codim=*/0);
-
-        depth_.resize(numCells);
-        for (unsigned cellIdx = 0; cellIdx < numCells; ++cellIdx) {
-            depth_[cellIdx] = UgGridHelpers::cellCenterDepth( grid, cellIdx );
+        const auto& eclProblem = ebosSimulator_.problem();
+        depth_.resize(local_num_cells_);
+        for (unsigned cellIdx = 0; cellIdx < local_num_cells_; ++cellIdx) {
+            depth_[cellIdx] = eclProblem.dofCenterDepth(cellIdx);
         }
     }
 
     template<typename TypeTag>
     void
     BlackoilWellModel<TypeTag>::
-    updatePerforationIntensiveQuantities() {
+    updatePerforationIntensiveQuantities()
+    {
         ElementContext elemCtx(ebosSimulator_);
         const auto& gridView = ebosSimulator_.gridView();
-        const auto& elemEndIt = gridView.template end</*codim=*/0, Dune::Interior_Partition>();
-        OPM_BEGIN_PARALLEL_TRY_CATCH();
-        for (auto elemIt = gridView.template begin</*codim=*/0, Dune::Interior_Partition>();
-             elemIt != elemEndIt;
-             ++elemIt)
-        {
 
-            elemCtx.updatePrimaryStencil(*elemIt);
+        OPM_BEGIN_PARALLEL_TRY_CATCH();
+        for (const auto& elem : elements(gridView, Dune::Partitions::interior)) {
+            elemCtx.updatePrimaryStencil(elem);
             int elemIdx = elemCtx.globalSpaceIndex(0, 0);
 
             if (!is_cell_perforated_[elemIdx]) {
@@ -1733,8 +1949,8 @@ namespace Opm {
                 connpos += well_perf_data_[i].size();
             }
             connpos *= np;
-            double weighted_temperature = 0.0;
-            double total_weight = 0.0;
+            std::array<double,2> weighted{0.0,0.0};
+            auto& [weighted_temperature, total_weight] = weighted;
 
             auto& well_info = local_parallel_well_info_[wellID].get();
             const int num_perf_this_well = well_info.communication().sum(well_perf_data_[wellID].size());
@@ -1765,8 +1981,7 @@ namespace Opm {
                 total_weight += weight_factor;
                 weighted_temperature += weight_factor * cellTemperatures;
             }
-            weighted_temperature = well_info.communication().sum(weighted_temperature);
-            total_weight = well_info.communication().sum(total_weight);
+            well_info.communication().sum(weighted.data(), 2);
             this->wellState().well(wellID).temperature = weighted_temperature/total_weight;
         }
     }
