@@ -49,6 +49,7 @@ namespace Opm
                      const std::vector<PerforationData>& perf_data)
     : Base(well, pw_info, time_step, param, rate_converter, pvtRegionIdx, num_components, num_phases, index_of_well, perf_data)
     , MSWEval(static_cast<WellInterfaceIndices<FluidSystem,Indices,Scalar>&>(*this))
+    , regularize_(false)
     , segment_fluid_initial_(this->numberOfSegments(), std::vector<double>(this->num_components_, 0.0))
     {
         // not handling solvent or polymer for now with multisegment well
@@ -71,6 +72,15 @@ namespace Opm
         if constexpr (Base::has_brine) {
             OPM_THROW(std::runtime_error, "brine is not supported by multisegment well yet");
         }
+
+        if constexpr (Base::has_watVapor) {
+            OPM_THROW(std::runtime_error, "water evaporation is not supported by multisegment well yet");
+        }
+
+        if(this->rsRvInj() > 0) {
+            OPM_THROW(std::runtime_error, "dissolved gas/ vapporized oil in injected oil/gas not supported by multisegment well yet."
+            << " \n See  (WCONINJE item 10 / WCONHIST item 8)");
+        }
     }
 
 
@@ -84,9 +94,10 @@ namespace Opm
          const std::vector<double>& depth_arg,
          const double gravity_arg,
          const int num_cells,
-         const std::vector< Scalar >& B_avg)
+         const std::vector< Scalar >& B_avg,
+         const bool changed_to_open_this_step)
     {
-        Base::init(phase_usage_arg, depth_arg, gravity_arg, num_cells, B_avg);
+        Base::init(phase_usage_arg, depth_arg, gravity_arg, num_cells, B_avg, changed_to_open_this_step);
 
         // TODO: for StandardWell, we need to update the perf depth here using depth_arg.
         // for MultisegmentWell, it is much more complicated.
@@ -276,7 +287,7 @@ namespace Opm
                 bhp_controlled_well = true;
             }
         }
-        if (thp_controlled_well || bhp_controlled_well) {
+        if (!this->changed_to_open_this_step_ && (thp_controlled_well || bhp_controlled_well)) {
 
             double total_rate = 0.0;
             const double sign = this->isInjector() ? 1.0:-1.0;
@@ -607,10 +618,10 @@ namespace Opm
 
         const double dFLimit = this->param_.dwell_fraction_max_;
         const double max_pressure_change = this->param_.max_pressure_change_ms_wells_;
-        this->MSWEval::updateWellState(dwells,
-                                       relaxation_factor,
-                                       dFLimit,
-                                       max_pressure_change);
+        this->MSWEval::updatePrimaryVariablesNewton(dwells,
+                                                    relaxation_factor,
+                                                    dFLimit,
+                                                    max_pressure_change);
 
         this->updateWellStateFromPrimaryVariables(well_state, getRefDensity(), deferred_logger);
         Base::calculateReservoirRates(well_state.well(this->index_of_well_));
@@ -731,9 +742,9 @@ namespace Opm
                     for (auto colB = this->duneB_[rowB].begin(), endB = this->duneB_[rowB].end(); colB != endB; ++colB) {
                         const auto col_index = colB.index();
                         OffDiagMatrixBlockWellType tmp1;
-                        Detail::multMatrixImpl(invDuneD[rowC][rowB], (*colB), tmp1, std::true_type());
+                        detail::multMatrixImpl(invDuneD[rowC][rowB], (*colB), tmp1, std::true_type());
                         typename SparseMatrixAdapter::MatrixBlock tmp2;
-                        Detail::multMatrixTransposedImpl((*colC), tmp1, tmp2, std::false_type());
+                        detail::multMatrixTransposedImpl((*colC), tmp1, tmp2, std::false_type());
                         jacobian.addToBlock(row_index, col_index, tmp2);
                     }
                 }
@@ -742,6 +753,74 @@ namespace Opm
     }
 
 
+   template<typename TypeTag>
+    void
+    MultisegmentWell<TypeTag>::
+    addWellPressureEquations(PressureMatrix& jacobian,
+                             const BVector& weights,
+                             const int pressureVarIndex,
+                             const bool /*use_well_weights*/,
+                             const WellState& well_state) const
+    {
+        // Add the pressure contribution to the cpr system for the well
+
+        // Add for coupling from well to reservoir
+        const auto seg_pressure_var_ind  = this->SPres; 
+        const int welldof_ind = this->duneC_.M() + this->index_of_well_;
+        if(not(this->isPressureControlled(well_state))){
+            for (size_t rowC = 0; rowC < this->duneC_.N(); ++rowC) {
+                for (auto colC = this->duneC_[rowC].begin(), endC = this->duneC_[rowC].end(); colC != endC; ++colC) {
+                    const auto row_index = colC.index();
+                    const auto& bw = weights[row_index];
+                    double matel = 0.0;
+                
+                    for(size_t i = 0; i< bw.size(); ++i){
+                        matel += bw[i]*(*colC)[seg_pressure_var_ind][i];
+                    }
+                    jacobian[row_index][welldof_ind] += matel;
+                
+                }
+            }
+        }
+        // make cpr weights for well by pure avarage of reservoir weights of the perforations
+        if(not(this->isPressureControlled(well_state))){        
+            auto well_weight = weights[0];
+            well_weight = 0.0;
+            int num_perfs = 0;
+            for (size_t rowB = 0; rowB < this->duneB_.N(); ++rowB) {
+                for (auto colB = this->duneB_[rowB].begin(), endB = this->duneB_[rowB].end(); colB != endB; ++colB) {
+                    const auto col_index = colB.index();
+                    const auto& bw = weights[col_index];
+                    well_weight += bw;
+                    num_perfs += 1;
+                }
+            }
+        
+            well_weight /= num_perfs;
+            assert(num_perfs>0);
+        
+            // Add for coupling from reservoir to well and caclulate diag elelement corresping to incompressible standard well
+            double diag_ell = 0.0;
+            for (size_t rowB = 0; rowB < this->duneB_.N(); ++rowB) {
+                const auto& bw = well_weight;
+                for (auto colB = this->duneB_[rowB].begin(), endB = this->duneB_[rowB].end(); colB != endB; ++colB) {
+                    const auto col_index = colB.index();               
+                    double matel = 0.0;
+                    for(size_t i = 0; i< bw.size(); ++i){
+                        matel += bw[i] *(*colB)[i][pressureVarIndex];
+                    }
+                    jacobian[welldof_ind][col_index] += matel;
+                    diag_ell -= matel;
+                }
+            }
+        
+            assert(diag_ell > 0.0);
+            jacobian[welldof_ind][welldof_ind] = diag_ell;
+        }else{
+            jacobian[welldof_ind][welldof_ind] = 1.0; // maybe we could have used diag_ell if calculated
+        }
+    }
+    
 
     template<typename TypeTag>
     template<class Value>
@@ -1070,7 +1149,7 @@ namespace Opm
         } else {
 
             const auto& paramsCell = materialLawManager->connectionMaterialLawParams(satid, cell_idx);
-            Eval relativePerms[3] = { 0.0, 0.0, 0.0 };
+            std::array<Eval,3> relativePerms = { 0.0, 0.0, 0.0 };
             MaterialLaw::relativePermeabilities(relativePerms, paramsCell, intQuants.fluidState());
 
             // reset the satnumvalue back to original
@@ -1122,7 +1201,7 @@ namespace Opm
         } else {
 
             const auto& paramsCell = materialLawManager->connectionMaterialLawParams(satid, cell_idx);
-            Scalar relativePerms[3] = { 0.0, 0.0, 0.0 };
+            std::array<Scalar,3> relativePerms = { 0.0, 0.0, 0.0 };
             MaterialLaw::relativePermeabilities(relativePerms, paramsCell, intQuants.fluidState());
 
             // reset the satnumvalue back to original
@@ -1164,12 +1243,20 @@ namespace Opm
         if ( bhp_limit_not_defaulted || !this->wellHasTHPConstraints(summaryState) ) {
             // if the BHP limit is not defaulted or the well does not have a THP limit
             // we need to check the BHP limit
+            double total_ipr_mass_rate = 0.0;
+            for (unsigned phaseIdx = 0; phaseIdx < FluidSystem::numPhases; ++phaseIdx)
+            {
+                if (!FluidSystem::phaseIsActive(phaseIdx)) {
+                    continue;
+                }
 
-            double ipr_rate = 0;
-            for (int p = 0; p < this->number_of_phases_; ++p) {
-                ipr_rate += this->ipr_a_[p] - this->ipr_b_[p] * bhp_limit;
+                const unsigned compIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::solventComponentIndex(phaseIdx));
+                const double ipr_rate = this->ipr_a_[compIdx] - this->ipr_b_[compIdx] * bhp_limit;
+
+                const double rho = FluidSystem::referenceDensity( phaseIdx, Base::pvtRegionIdx() );
+                total_ipr_mass_rate += ipr_rate * rho;
             }
-            if ( (this->isProducer() && ipr_rate < 0.) || (this->isInjector() && ipr_rate > 0.) ) {
+            if ( (this->isProducer() && total_ipr_mass_rate < 0.) || (this->isInjector() && total_ipr_mass_rate > 0.) ) {
                 this->operability_status_.operable_under_only_bhp_limit = false;
             }
 
@@ -1214,14 +1301,20 @@ namespace Opm
         std::fill(this->ipr_b_.begin(), this->ipr_b_.end(), 0.);
 
         const int nseg = this->numberOfSegments();
-        double seg_bhp_press_diff = 0;
-        double ref_depth = this->ref_depth_;
+        const double ref_depth = this->ref_depth_;
+        std::vector<double> seg_dp(nseg, 0.0);
         for (int seg = 0; seg < nseg; ++seg) {
             // calculating the perforation rate for each perforation that belongs to this segment
             const double segment_depth = this->segmentSet()[seg].depth();
-            const double dp = wellhelpers::computeHydrostaticCorrection(ref_depth, segment_depth, this->segment_densities_[seg].value(), this->gravity_);
-            ref_depth = segment_depth;
-            seg_bhp_press_diff += dp;
+            const int outlet_segment_index = this->segmentNumberToIndex(this->segmentSet()[seg].outletSegment());
+            const double segment_depth_outlet = seg == 0? ref_depth : this->segmentSet()[outlet_segment_index].depth();
+            double dp = wellhelpers::computeHydrostaticCorrection(segment_depth_outlet, segment_depth, this->segment_densities_[seg].value(), this->gravity_);
+            // we add the hydrostatic correction from the outlet segment
+            // in order to get the correction all the way to the bhp ref depth.
+            if (seg > 0) {
+                dp += seg_dp[outlet_segment_index];
+            }
+            seg_dp[seg] = dp;
             for (const int perf : this->segment_perforations_[seg]) {
             std::vector<Scalar> mob(this->num_components_, 0.0);
 
@@ -1232,7 +1325,7 @@ namespace Opm
             const auto& int_quantities = *(ebos_simulator.model().cachedIntensiveQuantities(cell_idx, /*timeIdx=*/ 0));
             const auto& fs = int_quantities.fluidState();
             // the pressure of the reservoir grid block the well connection is in
-                    // pressure difference between the segment and the perforation
+            // pressure difference between the segment and the perforation
             const double perf_seg_press_diff = this->gravity_ * this->segment_densities_[seg].value() * this->perforation_segment_depth_diffs_[perf];
             // pressure difference between the perforation and the grid cell
             const double cell_perf_press_diff = this->cell_perforation_pressure_diffs_[perf];
@@ -1249,7 +1342,7 @@ namespace Opm
             }
 
             // the pressure difference between the connection and BHP
-            const double h_perf = cell_perf_press_diff + perf_seg_press_diff + seg_bhp_press_diff;
+            const double h_perf = cell_perf_press_diff + perf_seg_press_diff + dp;
             const double pressure_diff = pressure_cell - h_perf;
 
             // do not take into consideration the crossflow here.
@@ -1261,15 +1354,12 @@ namespace Opm
             // the well index associated with the connection
             const double tw_perf = this->well_index_[perf]*ebos_simulator.problem().template rockCompTransMultiplier<double>(int_quantities, cell_idx);
 
-            // TODO: there might be some indices related problems here
-            // phases vs components
-            // ipr values for the perforation
             std::vector<double> ipr_a_perf(this->ipr_a_.size());
             std::vector<double> ipr_b_perf(this->ipr_b_.size());
-            for (int p = 0; p < this->number_of_phases_; ++p) {
-                const double tw_mob = tw_perf * mob[p] * b_perf[p];
-                ipr_a_perf[p] += tw_mob * pressure_diff;
-                ipr_b_perf[p] += tw_mob;
+            for (int comp_idx = 0; comp_idx < this->num_components_; ++comp_idx) {
+                const double tw_mob = tw_perf * mob[comp_idx] * b_perf[comp_idx];
+                ipr_a_perf[comp_idx] += tw_mob * pressure_diff;
+                ipr_b_perf[comp_idx] += tw_mob;
             }
 
             // we need to handle the rs and rv when both oil and gas are present
@@ -1292,10 +1382,9 @@ namespace Opm
                 ipr_b_perf[oil_comp_idx] += vap_oil_b;
             }
 
-            for (int p = 0; p < this->number_of_phases_; ++p) {
-                // TODO: double check the indices here
-                this->ipr_a_[this->ebosCompIdxToFlowCompIdx(p)] += ipr_a_perf[p];
-                this->ipr_b_[this->ebosCompIdxToFlowCompIdx(p)] += ipr_b_perf[p];
+            for (size_t comp_idx = 0; comp_idx < ipr_a_perf.size(); ++comp_idx) {
+                this->ipr_a_[comp_idx] += ipr_a_perf[comp_idx];
+                this->ipr_b_[comp_idx] += ipr_b_perf[comp_idx];
             }
             }
         }
@@ -1368,7 +1457,6 @@ namespace Opm
         if (!this->isOperableAndSolvable() && !this->wellIsStopped()) return true;
 
         const int max_iter_number = this->param_.max_inner_iter_ms_wells_;
-        const WellState well_state0 = well_state;
 
         {
             // getWellFiniteResiduals returns false for nan/inf residuals
@@ -1386,14 +1474,17 @@ namespace Opm
         bool converged = false;
         int stagnate_count = 0;
         bool relax_convergence = false;
+        this->regularize_ = false;
         for (; it < max_iter_number; ++it, ++debug_cost_counter_) {
 
             assembleWellEqWithoutIteration(ebosSimulator, dt, inj_controls, prod_controls, well_state, group_state, deferred_logger);
 
             const BVectorWell dx_well = mswellhelpers::applyUMFPack(this->duneD_, this->duneDSolver_, this->resWell_);
 
-            if (it > this->param_.strict_inner_iter_wells_)
+            if (it > this->param_.strict_inner_iter_wells_) {
                 relax_convergence = true;
+                this->regularize_ = true;
+            }
 
             const auto report = getWellConvergence(well_state, Base::B_avg_, deferred_logger, relax_convergence);
             if (report.converged()) {
@@ -1454,6 +1545,8 @@ namespace Opm
                     sstr << " well " << this->name() << " observes oscillation in inner iteration " << it << "\n";
                 }
                 sstr << " relaxation_factor is " << relaxation_factor << " now\n";
+
+                this->regularize_ = true;
                 deferred_logger.debug(sstr.str());
             }
             updateWellState(dx_well, well_state, deferred_logger, relaxation_factor);
@@ -1524,6 +1617,7 @@ namespace Opm
         auto& ws = well_state.well(this->index_of_well_);
         ws.dissolved_gas_rate = 0;
         ws.vaporized_oil_rate = 0;
+        ws.vaporized_wat_rate = 0;
 
         // for the black oil cases, there will be four equations,
         // the first three of them are the mass balance equations, the last one is the pressure equations.
@@ -1543,7 +1637,7 @@ namespace Opm
                 // Add a regularization_factor to increase the accumulation term
                 // This will make the system less stiff and help convergence for
                 // difficult cases
-                const Scalar regularization_factor =  this->param_.regularization_factor_ms_wells_;
+                const Scalar regularization_factor =  this->regularize_? this->param_.regularization_factor_wells_ : 1.0;
                 // for each component
                 for (int comp_idx = 0; comp_idx < this->num_components_; ++comp_idx) {
                     const EvalWell accumulation_term = regularization_factor * (segment_surface_volume * this->surfaceVolumeFraction(seg, comp_idx)
@@ -1561,10 +1655,10 @@ namespace Opm
                     const EvalWell segment_rate = this->getSegmentRateUpwinding(seg, comp_idx) * this->well_efficiency_factor_;
 
                     const int seg_upwind = this->upwinding_segments_[seg];
-                    // segment_rate contains the derivatives with respect to GTotal in seg,
+                    // segment_rate contains the derivatives with respect to WQTotal in seg,
                     // and WFrac and GFrac in seg_upwind
                     this->resWell_[seg][comp_idx] -= segment_rate.value();
-                    this->duneD_[seg][seg][comp_idx][GTotal] -= segment_rate.derivative(GTotal + Indices::numEq);
+                    this->duneD_[seg][seg][comp_idx][WQTotal] -= segment_rate.derivative(WQTotal + Indices::numEq);
                     if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
                         this->duneD_[seg][seg_upwind][comp_idx][WFrac] -= segment_rate.derivative(WFrac + Indices::numEq);
                     }
@@ -1582,10 +1676,10 @@ namespace Opm
                         const EvalWell inlet_rate = this->getSegmentRateUpwinding(inlet, comp_idx) * this->well_efficiency_factor_;
 
                         const int inlet_upwind = this->upwinding_segments_[inlet];
-                        // inlet_rate contains the derivatives with respect to GTotal in inlet,
+                        // inlet_rate contains the derivatives with respect to WQTotal in inlet,
                         // and WFrac and GFrac in inlet_upwind
                         this->resWell_[seg][comp_idx] += inlet_rate.value();
-                        this->duneD_[seg][inlet][comp_idx][GTotal] += inlet_rate.derivative(GTotal + Indices::numEq);
+                        this->duneD_[seg][inlet][comp_idx][WQTotal] += inlet_rate.derivative(WQTotal + Indices::numEq);
                         if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
                             this->duneD_[seg][inlet_upwind][comp_idx][WFrac] += inlet_rate.derivative(WFrac + Indices::numEq);
                         }
