@@ -94,6 +94,7 @@ namespace Opm
         using PolymerModule =  BlackOilPolymerModule<TypeTag>;
         using FoamModule = BlackOilFoamModule<TypeTag>;
         using BrineModule = BlackOilBrineModule<TypeTag>;
+        using typename Base::PressureMatrix;
 
         // number of the conservation equations
         static constexpr int numWellConservationEq = Indices::numPhases + Indices::numSolvents;
@@ -121,6 +122,7 @@ namespace Opm
         using Eval = typename StdWellEval::Eval;
         using EvalWell = typename StdWellEval::EvalWell;
         using BVectorWell = typename StdWellEval::BVectorWell;
+        using DiagMatrixBlockWellType = typename StdWellEval::DiagMatrixBlockWellType;
 
         StandardWell(const Well& well,
                      const ParallelWellInfo& pw_info,
@@ -137,7 +139,8 @@ namespace Opm
                           const std::vector<double>& depth_arg,
                           const double gravity_arg,
                           const int num_cells,
-                          const std::vector< Scalar >& B_avg) override;
+                          const std::vector< Scalar >& B_avg,
+                          const bool changed_to_open_this_step) override;
 
 
         virtual void initPrimaryVariablesEvaluation() const override;
@@ -178,7 +181,13 @@ namespace Opm
                                              WellState& well_state,
                                              DeferredLogger& deferred_logger) const override;
 
-        virtual void  addWellContributions(SparseMatrixAdapter& mat) const override;
+        virtual void addWellContributions(SparseMatrixAdapter& mat) const override;
+
+        virtual void addWellPressureEquations(PressureMatrix& mat,
+                                              const BVector& x,
+                                              const int pressureVarIndex,
+                                              const bool use_well_weights,
+                                              const WellState& well_state) const override;
 
         // iterate well equations with the specified control until converged
         bool iterateWellEqWithControl(const Simulator& ebosSimulator,
@@ -242,6 +251,8 @@ namespace Opm
 
 
     protected:
+        bool regularize_;
+
         // xw = inv(D)*(rw - C*x)
         void recoverSolutionWell(const BVector& x, BVectorWell& xw) const;
 
@@ -257,6 +268,7 @@ namespace Opm
                                                          std::vector<double>& b_perf,
                                                          std::vector<double>& rsmax_perf,
                                                          std::vector<double>& rvmax_perf,
+                                                         std::vector<double>& rvwmax_perf,
                                                          std::vector<double>& surf_dens_perf) const;
 
         void computeWellConnectionDensitesPressures(const Simulator& ebosSimulator,
@@ -264,6 +276,7 @@ namespace Opm
                                                     const std::vector<double>& b_perf,
                                                     const std::vector<double>& rsmax_perf,
                                                     const std::vector<double>& rvmax_perf,
+                                                    const std::vector<double>& rvwmax_perf,
                                                     const std::vector<double>& surf_dens_perf,
                                                     DeferredLogger& deferred_logger);
 
@@ -280,6 +293,7 @@ namespace Opm
                                  std::vector<EvalWell>& cq_s,
                                  double& perf_dis_gas_rate,
                                  double& perf_vap_oil_rate,
+                                 double& perf_vap_wat_rate,
                                  DeferredLogger& deferred_logger) const;
 
         void computePerfRateScalar(const IntensiveQuantities& intQuants,
@@ -297,6 +311,7 @@ namespace Opm
                              const Value& bhp,
                              const Value& rs,
                              const Value& rv,
+                             const Value& rvw,
                              std::vector<Value>& b_perfcells_dense,
                              const double Tw,
                              const int perf,
@@ -306,6 +321,7 @@ namespace Opm
                              std::vector<Value>& cq_s,
                              double& perf_dis_gas_rate,
                              double& perf_vap_oil_rate,
+                             double& perf_vap_wat_rate,
                              DeferredLogger& deferred_logger) const;
 
         void computeWellRatesWithBhpIterations(const Simulator& ebosSimulator,
