@@ -34,13 +34,11 @@
 #include <opm/simulators/wells/ParallelWellInfo.hpp>
 #include <opm/simulators/wells/WellInterfaceIndices.hpp>
 #include <opm/simulators/wells/WellState.hpp>
+#include <opm/simulators/linalg/bda/WellContributions.hpp>
 
 #include <cassert>
 #include <cmath>
 
-#if HAVE_CUDA || HAVE_OPENCL
-#include <opm/simulators/linalg/bda/WellContributions.hpp>
-#endif
 
 
 namespace Opm
@@ -195,12 +193,19 @@ getQs(const int comp_idx) const
             if (Indices::enableSolvent && comp_idx == Indices::contiSolventEqIdx) { // solvent
                 inj_frac = baseif_.wsolvent();
             } else if (comp_idx == int(Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx))) {
-                inj_frac = Indices::enableSolvent ? 1.0 - baseif_.wsolvent() : 1.0;
+                inj_frac = 1.0 - baseif_.rsRvInj();
+                if (Indices::enableSolvent) {
+                    inj_frac -= baseif_.wsolvent();
+                }
+            } else if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx) && comp_idx == int(Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx))) {
+                inj_frac = baseif_.rsRvInj();
             }
             break;
         case InjectorType::OIL:
             if (comp_idx == int(Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx))) {
-                inj_frac = 1.0;
+                inj_frac = 1.0 - baseif_.rsRvInj();
+            } else if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx) && comp_idx == int(Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx))) {
+                inj_frac = baseif_.rsRvInj();
             }
             break;
         case InjectorType::MULTI:
@@ -316,9 +321,9 @@ updatePrimaryVariables(const WellState& well_state, DeferredLogger& deferred_log
                 if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)) {
                     auto phase = baseif_.wellEcl().getInjectionProperties().injectorType;
                     if (phase == InjectorType::GAS) {
-                        primary_variables_[GFrac] = 1.0;
+                        primary_variables_[GFrac] = (1.0 - baseif_.rsRvInj());
                         if constexpr (Indices::enableSolvent) {
-                            primary_variables_[GFrac] = 1.0 - baseif_.wsolvent();
+                            primary_variables_[GFrac] = 1.0 - baseif_.rsRvInj() - baseif_.wsolvent();
                             primary_variables_[SFrac] = baseif_.wsolvent();
                         }
                     } else {
@@ -426,7 +431,7 @@ assembleControlEq(const WellState& well_state,
     // TODO: we should use a different index system for the well equations
     this->resWell_[0][Bhp] = control_eq.value();
     for (int pv_idx = 0; pv_idx < numWellEq_; ++pv_idx) {
-        this->invDuneD_[0][0][Bhp][pv_idx] = control_eq.derivative(pv_idx + Indices::numEq);
+        this->duneD_[0][0][Bhp][pv_idx] = control_eq.derivative(pv_idx + Indices::numEq);
     }
 }
 
@@ -568,6 +573,13 @@ updateThp(WellState& well_state,
         return;
     }
 
+    // For THP controlled wells, we know the thp value
+    bool thp_controlled = baseif_.isInjector() ? ws.injection_cmode == Well::InjectorCMode::THP:
+                                              ws.production_cmode == Well::ProducerCMode::THP;
+    if (thp_controlled) {
+        return;
+    }
+
     // the well is under other control types, we calculate the thp based on bhp and rates
     std::vector<double> rates(3, 0.0);
 
@@ -831,6 +843,7 @@ computeConnectionDensities(const std::vector<double>& perfComponentRates,
                            const std::vector<double>& b_perf,
                            const std::vector<double>& rsmax_perf,
                            const std::vector<double>& rvmax_perf,
+                           const std::vector<double>& rvwmax_perf,
                            const std::vector<double>& surf_dens_perf,
                            DeferredLogger& deferred_logger)
 {
@@ -939,7 +952,7 @@ computeConnectionDensities(const std::vector<double>& perfComponentRates,
         // Compute volume ratio.
         x = mix;
 
-        // Subtract dissolved gas from oil phase and vapporized oil from gas phase
+        // Subtract dissolved gas from oil phase and vapporized oil from gas phase and vaporized water from gas phase
         if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx) && FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
             const unsigned gaspos = Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
             const unsigned oilpos = Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx);
@@ -971,7 +984,37 @@ computeConnectionDensities(const std::vector<double>& perfComponentRates,
                     x[oilpos] = (mix[oilpos] - mix[gaspos]*rv)/d;
                 }
             }
+            if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
+                //matrix system: (mix[oilpos] = q_os, x[oilpos] = bo*q_or, etc...)
+                //┌             ┐   ┌                ┐  ┌           ┐
+                //│mix[oilpos]  │   | 1     Rv     0 |  |x[oilpos]  |
+                //│mix[gaspos]  │ = │ Rs    1      0 │  │x[gaspos]  │
+                //│mix[waterpos]│   │ 0     Rvw    1 │  │x[waterpos │
+                //└             ┘   └                ┘  └           ┘
+                const unsigned waterpos = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+                double rvw = 0.0;
+                if (!rvwmax_perf.empty() && mix[gaspos] > 1e-12) {
+                    rvw = std::min(mix[waterpos]/mix[gaspos], rvwmax_perf[perf]);
+                }
+                if (rvw > 0.0) {
+                    // Subtract water in gas from water mixture
+                    x[waterpos] = mix[waterpos] - x[gaspos] * rvw;
+                }
+            }
+        } else if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx) && FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
+            //no oil
+            const unsigned gaspos = Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
+            const unsigned waterpos = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+            double rvw = 0.0;
+            if (!rvwmax_perf.empty() && mix[gaspos] > 1e-12) {
+                rvw = std::min(mix[waterpos]/mix[gaspos], rvwmax_perf[perf]);
+            }
+            if (rvw > 0.0) {
+               // Subtract water in gas from water mixture
+               x[waterpos] = mix[waterpos] - mix[gaspos] * rvw;
+            }
         }
+
         double volrat = 0.0;
         for (int component = 0; component < num_comp; ++component) {
             volrat += x[component] / b_perf[perf*num_comp+ component];
@@ -1018,16 +1061,16 @@ init(std::vector<double>& perf_depth,
     //[A C^T    [x    =  [ res
     // B D] x_well]      res_well]
     // set the size of the matrices
-    this->invDuneD_.setSize(1, 1, 1);
+    this->duneD_.setSize(1, 1, 1);
     this->duneB_.setSize(1, num_cells, baseif_.numPerfs());
     this->duneC_.setSize(1, num_cells, baseif_.numPerfs());
 
-    for (auto row=this->invDuneD_.createbegin(), end = this->invDuneD_.createend(); row!=end; ++row) {
+    for (auto row=this->duneD_.createbegin(), end = this->duneD_.createend(); row!=end; ++row) {
         // Add nonzeros for diagonal
         row.insert(row.index());
     }
     // the block size is run-time determined now
-    this->invDuneD_[0][0].resize(numWellEq_, numWellEq_);
+    this->duneD_[0][0].resize(numWellEq_, numWellEq_);
 
     for (auto row = this->duneB_.createbegin(), end = this->duneB_.createend(); row!=end; ++row) {
         for (int perf = 0 ; perf < baseif_.numPerfs(); ++perf) {
@@ -1065,13 +1108,12 @@ init(std::vector<double>& perf_depth,
         this->Bx_[i].resize(numWellEq_);
     }
 
-    this->invDrw_.resize( this->invDuneD_.N() );
-    for (unsigned i = 0; i < this->invDuneD_.N(); ++i) {
+    this->invDrw_.resize( this->duneD_.N() );
+    for (unsigned i = 0; i < this->duneD_.N(); ++i) {
         this->invDrw_[i].resize(numWellEq_);
     }
 }
 
-#if HAVE_CUDA || HAVE_OPENCL
 template<class FluidSystem, class Indices, class Scalar>
 void
 StandardWellEval<FluidSystem,Indices,Scalar>::
@@ -1120,7 +1162,6 @@ addWellContribution(WellContributions& wellContribs) const
     }
     wellContribs.addMatrix(WellContributions::MatrixType::B, colIndices.data(), nnzValues.data(), this->duneB_.nonzeroes());
 }
-#endif
 
 #define INSTANCE(A,...) \
 template class StandardWellEval<BlackOilFluidSystem<double,A>,__VA_ARGS__,double>;
