@@ -134,13 +134,6 @@ InterpData findInterpData(const double value_in, const std::vector<double>& valu
         }
     }
 
-    // Disallow extrapolation with higher factor than 3.0.
-    // The factor 3.0 has been chosen because it works well
-    // with certain testcases, and may not be optimal.
-    if (retval.factor_ > 3.0) {
-        retval.factor_ = 3.0;
-    }
-
     return retval;
 }
 
@@ -345,12 +338,19 @@ VFPEvaluation bhp(const VFPProdTable& table,
                   const double  liquid,
                   const double  vapour,
                   const double  thp,
-                  const double  alq)
+                  const double  alq,
+                  const double  explicit_wfr,
+                  const double  explicit_gfr,
+                  const bool    use_vfpexplicit)
 {
     //Find interpolation variables
     double flo = detail::getFlo(table, aqua, liquid, vapour);
     double wfr = detail::getWFR(table, aqua, liquid, vapour);
     double gfr = detail::getGFR(table, aqua, liquid, vapour);
+    if (use_vfpexplicit) {
+        wfr = explicit_wfr;
+        gfr = explicit_gfr;
+    }
 
     //First, find the values to interpolate between
     //Recall that flo is negative in Opm, so switch sign.
