@@ -21,19 +21,29 @@
 #include <config.h>
 #include <opm/simulators/wells/WellGroupHelpers.hpp>
 
+#include <opm/input/eclipse/Schedule/Schedule.hpp>
 #include <opm/input/eclipse/Schedule/Group/GConSump.hpp>
 #include <opm/input/eclipse/Schedule/Group/GConSale.hpp>
+#include <opm/input/eclipse/Schedule/Group/GPMaint.hpp>
+#include <opm/input/eclipse/Schedule/Group/Group.hpp>
+
+#include <opm/material/fluidsystems/BlackOilFluidSystem.hpp>
+
 #include <opm/simulators/utils/DeferredLogger.hpp>
 #include <opm/simulators/utils/DeferredLoggingErrorHelpers.hpp>
+#include <opm/simulators/utils/ParallelCommunication.hpp>
+
+#include <opm/simulators/wells/GroupState.hpp>
+#include <opm/simulators/wells/RegionAverageCalculator.hpp>
 #include <opm/simulators/wells/TargetCalculator.hpp>
 #include <opm/simulators/wells/VFPProdProperties.hpp>
 #include <opm/simulators/wells/WellState.hpp>
-#include <opm/simulators/wells/WellContainer.hpp>
 
 #include <algorithm>
 #include <cassert>
 #include <set>
 #include <stack>
+#include <stdexcept>
 
 namespace {
     Opm::GuideRate::RateVector
@@ -655,6 +665,86 @@ namespace WellGroupHelpers
     }
 
 
+    template <class RegionalValues>
+    void updateGpMaintTargetForGroups(const Group& group,
+                                      const Schedule& schedule,
+                                      const RegionalValues& regional_values,
+                                      const int reportStepIdx,
+                                      const double dt,
+                                      const WellState& well_state,
+                                      GroupState& group_state)
+    {
+        for (const std::string& groupName : group.groups()) {
+            const Group& groupTmp = schedule.getGroup(groupName, reportStepIdx);
+            updateGpMaintTargetForGroups(groupTmp, schedule, regional_values, reportStepIdx, dt, well_state, group_state);
+        }
+        const auto& gpm = group.gpmaint();
+        if (!gpm)
+            return;
+
+        const auto [name, number] = *gpm->region();
+        const double error = gpm->pressure_target() - regional_values.pressure(number);
+        double current_rate = 0.0;
+        const auto& pu = well_state.phaseUsage();
+        double sign = 1.0;
+        switch (gpm->flow_target()) {
+            case GPMaint::FlowTarget::RESV_PROD:
+            {
+                current_rate = -group_state.injection_vrep_rate(group.name());
+                sign = -1.0;
+                break;
+            }
+            case GPMaint::FlowTarget::RESV_OINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Liquid])
+                    current_rate = group_state.injection_reservoir_rates(group.name())[pu.phase_pos[BlackoilPhases::Liquid]];
+
+                break;
+            }
+            case GPMaint::FlowTarget::RESV_WINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Aqua])
+                    current_rate = group_state.injection_reservoir_rates(group.name())[pu.phase_pos[BlackoilPhases::Aqua]];
+
+                break;
+            }
+            case GPMaint::FlowTarget::RESV_GINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Vapour])
+                    current_rate = group_state.injection_reservoir_rates(group.name())[pu.phase_pos[BlackoilPhases::Vapour]];
+                break;
+            }
+            case GPMaint::FlowTarget::SURF_OINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Liquid])
+                    current_rate = group_state.injection_surface_rates(group.name())[pu.phase_pos[BlackoilPhases::Liquid]];
+
+                break;
+            }
+            case GPMaint::FlowTarget::SURF_WINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Aqua])
+                    current_rate = group_state.injection_surface_rates(group.name())[pu.phase_pos[BlackoilPhases::Aqua]];
+
+                break;
+            }
+            case GPMaint::FlowTarget::SURF_GINJ:
+            {
+                if (pu.phase_used[BlackoilPhases::Vapour])
+                    current_rate = group_state.injection_surface_rates(group.name())[pu.phase_pos[BlackoilPhases::Vapour]];
+
+                break;
+            }
+            default:
+                throw std::invalid_argument("Invalid Flow target type in GPMAINT");
+        }
+
+        auto& gpmaint_state = group_state.gpmaint(group.name());
+        double rate = gpm->rate(gpmaint_state, current_rate, error, dt);
+        group_state.update_gpmaint_target(group.name(), std::max(0.0, sign * rate));
+    }
+
+
 
 
     std::map<std::string, double>
@@ -766,7 +856,10 @@ namespace WellGroupHelpers
                                                               rates[BlackoilPhases::Liquid],
                                                               rates[BlackoilPhases::Vapour],
                                                               up_press,
-                                                              alq);
+                                                              alq,
+                                                              0.0, //explicit_wfr
+                                                              0.0, //explicit_gfr
+                                                              false); //use_expvfp we dont support explicit lookup
 #define EXTRA_DEBUG_NETWORK 0
 #if EXTRA_DEBUG_NETWORK
                     std::ostringstream oss;
@@ -1180,82 +1273,46 @@ namespace WellGroupHelpers
         const size_t num_ancestors = chain.size() - 1;
         // we need to find out the level where the current well is applied to the local reduction 
         size_t local_reduction_level = 0;
-        for (size_t ii = 0; ii < num_ancestors; ++ii) {
-            if ((ii == 0) || guideRate->has(chain[ii])) {
+        for (size_t ii = 1; ii < num_ancestors; ++ii) {
+            const int num_gr_ctrl = groupControlledWells(schedule,
+                                                         wellState,
+                                                         group_state,
+                                                         reportStepIdx,
+                                                         chain[ii],
+                                                         "",
+                                                         /*is_producer*/ true,
+                                                         /*injectionPhaseNotUsed*/ Phase::OIL);
+            if (guideRate->has(chain[ii]) && num_gr_ctrl > 0) {
                 local_reduction_level = ii;
             }
         }
-
         // check whether guide rate is violated
-        if (local_reduction_level > 0) {
-            const auto& guided_group = chain[local_reduction_level];
-            const double grefficiency
+        for (size_t ii = 1; ii < num_ancestors; ++ii) {
+            if (guideRate->has(chain[ii])) {
+                const auto& guided_group = chain[ii];
+                const double grefficiency
                             = schedule.getGroup(guided_group, reportStepIdx).getGroupEfficiencyFactor();
-            const double currentRateFraction = grefficiency * localCurrentRate(guided_group) / (localCurrentRate(chain[local_reduction_level-1]));
-            const double guiderateFraction = localFraction(guided_group);
-            // we add a factor here to avoid switching due to numerical instability
-            const double factor = 1.01;
-            if (currentRateFraction > (guiderateFraction * factor)) {
-                return std::make_pair(true, guiderateFraction/currentRateFraction);
+                const double currentRateFraction = grefficiency * localCurrentRate(guided_group) / (localCurrentRate(chain[ii-1]));
+                const double guiderateFraction = localFraction(guided_group);
+                // we add a factor here to avoid switching due to numerical instability
+                const double factor = 1.01;
+                if (currentRateFraction > (guiderateFraction * factor)) {
+                    return std::make_pair(true, guiderateFraction/currentRateFraction);
+                }
             }
         }
-
         double target = orig_target;
         for (size_t ii = 0; ii < num_ancestors; ++ii) {
             if ((ii == 0) || guideRate->has(chain[ii])) {
                 // Apply local reductions only at the control level
                 // (top) and for levels where we have a specified
                 // group guide rate.
-                target -= localReduction(chain[ii]);
-
+                if (local_reduction_level >= ii ) {
+                    target -= localReduction(chain[ii]);
+                }
                 // Add my reduction back at the level where it is included in the local reduction
-                if (local_reduction_level == ii )
+                if (local_reduction_level == ii ) {
                     target += current_rate * efficiencyFactor;
-
-                for (size_t iii = ii + 1; iii < num_ancestors; ++iii) {
-                    // Not final level. Add sub-level reduction back, if
-                    // it was not under individual control and nonzero due to having no group-controlled
-                    // wells. We may need to look several levels down the hierarchy to find groups without
-                    // group control wells
-                    const Group::ProductionCMode& subGroupControl = group_state.production_control(chain[iii]);
-                    const bool individual_control = (subGroupControl != Group::ProductionCMode::FLD
-                                                     && subGroupControl != Group::ProductionCMode::NONE);
-                    // The sub group is on individual control. No adjustments needed.
-                    if (individual_control) {
-                        break;
-                    }
-
-                    // Note that we make this call without setting
-                    // the current well to be always included, because we
-                    // want to know the situation that applied to the
-                    // calculation of reductions.
-                    const int num_gr_ctrl = groupControlledWells(schedule,
-                                                                 wellState,
-                                                                 group_state,
-                                                                 reportStepIdx,
-                                                                 chain[iii],
-                                                                 "",
-                                                                 /*is_producer*/ true,
-                                                                 /*injectionPhaseNotUsed*/ Phase::OIL);
-                    if (num_gr_ctrl == 0) {
-                        // We found a sub wells with no group controlled wells. We now need to adapt the reduction rate
-                        // to reflect what would have happen if the current well under consideration would have been
-                        // under group control. I.e. we first remove the efficient sub_rate from the reduction rate
-                        // (since the reduction rate is removed already from the target we need to add it to the target)
-                        const double sub_efficiency
-                            = schedule.getGroup(chain[iii], reportStepIdx).getGroupEfficiencyFactor();
-                        const double sub_rate = tcalc.calcModeRateFromRates(group_state.production_rates(chain[iii]));
-                        target += sub_efficiency * sub_rate;
-                        // than we remove the local reduction from the target if it does not have a guide rate
-                        if (!guideRate->has(chain[iii])) {
-                            target -= sub_efficiency * localReduction(chain[iii]);
-                            // this local reduction rate may also need adjustments since it may be computed based on the
-                            // assumption that its subgroup dont have a group control wells. I.e we need to move down the
-                            // hierarchy.
-                        } else {
-                            break;
-                        }
-                    }
                 }
             }
             target *= localFraction(chain[ii + 1]);
@@ -1358,23 +1415,33 @@ namespace WellGroupHelpers
         const size_t num_ancestors = chain.size() - 1;
         // we need to find out the level where the current well is applied to the local reduction
         size_t local_reduction_level = 0;
-        for (size_t ii = 0; ii < num_ancestors; ++ii) {
-            if ((ii == 0) || guideRate->has(chain[ii], injectionPhase)) {
+        for (size_t ii = 1; ii < num_ancestors; ++ii) {
+            const int num_gr_ctrl = groupControlledWells(schedule,
+                                                         wellState,
+                                                                 group_state,
+                                                                 reportStepIdx,
+                                                                 chain[ii],
+                                                                 "",
+                                                                 /*is_producer*/ false,
+                                                                 injectionPhase);
+            if (guideRate->has(chain[ii], injectionPhase) && num_gr_ctrl > 0) {
                 local_reduction_level = ii;
             }
         }
 
         // check whether guide rate is violated
-        if (local_reduction_level > 0) {
-            const auto& guided_group = chain[local_reduction_level];
-            const double grefficiency
+        for (size_t ii = 1; ii < num_ancestors; ++ii) {
+            if (guideRate->has(chain[ii], injectionPhase)) {
+                const auto& guided_group = chain[ii];
+                const double grefficiency
                             = schedule.getGroup(guided_group, reportStepIdx).getGroupEfficiencyFactor();
-            const double currentRateFraction = grefficiency * localCurrentRate(guided_group) / (localCurrentRate(chain[local_reduction_level-1]));
-            const double guiderateFraction = localFraction(guided_group);
-            // we add a factor here to avoid switching due to numerical instability
-            const double factor = 1.01;
-            if (currentRateFraction > (guiderateFraction * factor)) {
-                return std::make_pair(true, guiderateFraction/currentRateFraction);
+                const double currentRateFraction = grefficiency * localCurrentRate(guided_group) / (localCurrentRate(chain[ii-1]));
+                const double guiderateFraction = localFraction(guided_group);
+                // we add a factor here to avoid switching due to numerical instability
+                const double factor = 1.01;
+                if (currentRateFraction > (guiderateFraction * factor)) {
+                    return std::make_pair(true, guiderateFraction/currentRateFraction);
+                }
             }
         }
 
@@ -1384,57 +1451,13 @@ namespace WellGroupHelpers
                 // Apply local reductions only at the control level
                 // (top) and for levels where we have a specified
                 // group guide rate.
-                target -= localReduction(chain[ii]);
+                if (local_reduction_level >= ii ) {
+                    target -= localReduction(chain[ii]);
+                }
 
                 // Add my reduction back at the level where it is included in the local reduction
-                if (local_reduction_level == ii )
+                if (local_reduction_level == ii ) {
                     target += current_rate * efficiencyFactor;
-
-                for (size_t iii = ii + 1; iii < num_ancestors; ++iii) {
-                    // Not final level. Add sub-level reduction back, if
-                    // it was not under individual control and nonzero due to having no group-controlled
-                    // wells. We may need to look several levels down the hierarchy to find groups without
-                    // group control wells
-                    const Group::InjectionCMode& subGroupControl = group_state.injection_control(chain[iii], injectionPhase);
-                    const bool individual_control = (subGroupControl != Group::InjectionCMode::FLD
-                                                     && subGroupControl != Group::InjectionCMode::NONE);
-                    // The sub group is on individual control. No adjustments needed.
-                    if (individual_control) {
-                        break;
-                    }
-
-                    // Note that we make this call without setting
-                    // the current well to be always included, because we
-                    // want to know the situation that applied to the
-                    // calculation of reductions.
-                    const int num_gr_ctrl = groupControlledWells(schedule,
-                                                                 wellState,
-                                                                 group_state,
-                                                                 reportStepIdx,
-                                                                 chain[iii],
-                                                                 "",
-                                                                 /*is_producer*/ false,
-                                                                 injectionPhase);
-                    if (num_gr_ctrl == 0) {
-                        // We found a sub wells with no group controlled wells. We now need to adapt the reduction rate
-                        // to reflect what would have happen if the well under consideration would have been
-                        // under group control. I.e. we first remove the efficient sub_rate from the reduction rate
-                        // (since the reduction rate is removed already from the target we need to add it to the target)
-                        const double sub_efficiency
-                            = schedule.getGroup(chain[iii], reportStepIdx).getGroupEfficiencyFactor();
-                        const double sub_rate
-                            = tcalc.calcModeRateFromRates(group_state.injection_surface_rates(chain[iii]));
-                        target += sub_efficiency * sub_rate;
-                        // than we remove the local reduction from the target if it does not have a guide rate
-                        if (!guideRate->has(chain[iii], injectionPhase)) {
-                            target -= sub_efficiency * localReduction(chain[iii]);
-                            // this local reduction rate may also need adjustments since it may be computed based on the
-                            // assumption that its subgroup dont have group control wells. I.e we need to move down the
-                            // hierarchy.
-                        } else {
-                            break;
-                        }
-                    }
                 }
             }
             target *= localFraction(chain[ii + 1]);
@@ -1585,45 +1608,47 @@ namespace WellGroupHelpers
         }
     }
 
- #define INSTANCE_WELLGROUP_HELPERS(...) \
-    template \
-    void updateGuideRateForProductionGroups<Dune::CollectiveCommunication<__VA_ARGS__>>(const Group& group,\
-                                                              const Schedule& schedule, \
-                                                              const PhaseUsage& pu, \
-                                                              const int reportStepIdx, \
-                                                              const double& simTime, \
-                                                              WellState& wellState, \
-                                                              const GroupState& group_state, \
-                                                              const Dune::CollectiveCommunication<__VA_ARGS__>& comm, \
-                                                              GuideRate* guideRate, \
-                                                              std::vector<double>& pot); \
-    template \
-    void updateGuideRatesForWells<Dune::CollectiveCommunication<__VA_ARGS__>>(const Schedule& schedule, \
-                                                              const PhaseUsage& pu, \
-                                                              const int reportStepIdx, \
-                                                              const double& simTime, \
-                                                              const WellState& wellState, \
-                                                              const Dune::CollectiveCommunication<__VA_ARGS__>& comm, \
-                                                              GuideRate* guideRate); \
-    template \
-    void updateGuideRates<Dune::CollectiveCommunication<__VA_ARGS__>>(const Group& group, \
-                                                                      const Schedule& schedule, \
-                                                                      const SummaryState& summary_state, \
-                                                                      const PhaseUsage& pu, \
-                                                                      const int report_step, \
-                                                                      const double sim_time, \
-                                                                      WellState& well_state, \
-                                                                      const GroupState& group_state, \
-                                                                      const Dune::CollectiveCommunication<__VA_ARGS__>& comm,\
-                                                                      GuideRate* guide_rate, \
-                                                                      std::vector<double>& pot,\
-                                                                      Opm::DeferredLogger& deferred_logger);
+ using AvgP = RegionAverageCalculator::AverageRegionalPressure<BlackOilFluidSystem<double>,std::vector<int>>;
+ template void WellGroupHelpers::updateGpMaintTargetForGroups<AvgP>(const Group&,
+                                                                    const Schedule&,
+                                                                    const AvgP&,
+                                                                    int,
+                                                                    double,
+                                                                    const WellState&,
+                                                                    GroupState&);
 
-#if HAVE_MPI
-    INSTANCE_WELLGROUP_HELPERS(MPI_Comm)
-#else
-    INSTANCE_WELLGROUP_HELPERS(Dune::No_Comm)
-#endif
+template
+void updateGuideRateForProductionGroups<Parallel::Communication>(const Group& group,
+                                                                 const Schedule& schedule,
+                                                                 const PhaseUsage& pu,
+                                                                 const int reportStepIdx,
+                                                                 const double& simTime,
+                                                                 WellState& wellState,
+                                                                 const GroupState& group_state,
+                                                                 const Parallel::Communication& comm,
+                                                                 GuideRate* guideRate,
+                                                                 std::vector<double>& pot);
+template
+void updateGuideRatesForWells<Parallel::Communication>(const Schedule& schedule,
+                                                       const PhaseUsage& pu,
+                                                       const int reportStepIdx,
+                                                       const double& simTime,
+                                                       const WellState& wellState,
+                                                       const Parallel::Communication& comm,
+                                                       GuideRate* guideRate);
+template
+void updateGuideRates<Parallel::Communication>(const Group& group,
+                                               const Schedule& schedule,
+                                               const SummaryState& summary_state,
+                                               const PhaseUsage& pu,
+                                               const int report_step,
+                                               const double sim_time,
+                                               WellState& well_state,
+                                               const GroupState& group_state,
+                                               const Parallel::Communication& comm,
+                                               GuideRate* guide_rate,
+                                               std::vector<double>& pot,
+                                               DeferredLogger&);
 
 } // namespace WellGroupHelpers
 
