@@ -118,7 +118,16 @@ activeProductionConstraint(const SingleWellState& ws,
     if (controls.hasControl(Well::ProducerCMode::LRAT) && currentControl != Well::ProducerCMode::LRAT) {
         double current_rate = -ws.surface_rates[pu.phase_pos[BlackoilPhases::Liquid]];
         current_rate -= ws.surface_rates[pu.phase_pos[BlackoilPhases::Aqua]];
-        if (controls.liquid_rate < current_rate)
+
+        bool skip = false;
+        if (controls.liquid_rate == controls.oil_rate) {
+            const double current_water_rate = ws.surface_rates[pu.phase_pos[BlackoilPhases::Aqua]];
+            if (std::abs(current_water_rate) < 1e-12) {
+                skip = true;
+                deferred_logger.debug("LRAT_ORAT_WELL", "Well " + this->name() + " The LRAT target is equal the ORAT target and the water rate is zero, skip checking LRAT");
+            }
+        }
+        if (!skip && controls.liquid_rate < current_rate)
             return Well::ProducerCMode::LRAT;
     }
 
@@ -163,7 +172,7 @@ activeProductionConstraint(const SingleWellState& ws,
     if (controls.hasControl(Well::ProducerCMode::THP) && currentControl != Well::ProducerCMode::THP) {
         const auto& thp = getTHPConstraint(summaryState);
         double current_thp = ws.thp;
-        if (thp > current_thp) {
+        if (thp > current_thp && !ws.trivial_target) {
             // If WVFPEXP item 4 is set to YES1 or YES2
             // switching to THP is prevented if the well will
             // produce at a higher rate with THP control
@@ -182,7 +191,7 @@ activeProductionConstraint(const SingleWellState& ws,
                 return Well::ProducerCMode::THP;
             } else {
                 this->operability_status_.thp_limit_violated_but_not_switched = true;
-                deferred_logger.debug("NOT_SWITCHING_TO_THP",
+                deferred_logger.info("NOT_SWITCHING_TO_THP",
                 "The THP limit is violated for producer " +
                 this->name() +
                 ". But the rate will increase if switched to THP. " +
@@ -716,6 +725,11 @@ updateWellTestStateEconomic(const SingleWellState& ws,
         return;
     }
 
+    if (this->isInjector()) {
+        deferred_logger.warning("ECON_LIMITS_INJECTOR_" + this->name(), this->name() + " is an injector, the production economic limits for this well will be ignored.\n");
+        return;
+    }
+
     // flag to check if the mim oil/gas rate limit is violated
     bool rate_limit_violated = false;
 
@@ -1120,6 +1134,10 @@ getGroupProductionTargetRate(const Group& group,
     const auto& rates = ws.surface_rates;
     const auto current_rate = -tcalc.calcModeRateFromRates(rates); // Switch sign since 'rates' are negative for producers.
     double scale = 1.0;
+    if (target_rate == 0.0) {
+        return 0.0;
+    }
+
     if (current_rate > 1e-14)
         scale = target_rate/current_rate;
     return scale;
