@@ -55,8 +55,8 @@ WellInterfaceGeneric::WellInterfaceGeneric(const Well& well,
       , number_of_phases_(num_phases)
       , index_of_well_(index_of_well)
       , perf_data_(&perf_data)
-      , ipr_a_(number_of_phases_)
-      , ipr_b_(number_of_phases_)
+      , ipr_a_(num_components)
+      , ipr_b_(num_components)
 {
     assert(well.name()==pw_info.name());
     assert(std::is_sorted(perf_data.begin(), perf_data.end(),
@@ -173,6 +173,11 @@ double WellInterfaceGeneric::wsolvent() const
     return wsolvent_;
 }
 
+double WellInterfaceGeneric::rsRvInj() const
+{
+    return well_ecl_.getInjectionProperties().rsRvInj;
+}
+
 bool WellInterfaceGeneric::wellHasTHPConstraints(const SummaryState& summaryState) const
 {
     if (dynamic_thp_limit_) {
@@ -342,6 +347,11 @@ void WellInterfaceGeneric::setDynamicThpLimit(const double thp_limit)
     dynamic_thp_limit_ = thp_limit;
 }
 
+std::optional<double> WellInterfaceGeneric::getDynamicThpLimit() const
+{
+    return dynamic_thp_limit_;
+}
+
 void WellInterfaceGeneric::updatePerforatedCell(std::vector<bool>& is_cell_perforated)
 {
 
@@ -409,6 +419,17 @@ bool WellInterfaceGeneric::isOperableAndSolvable() const
     return operability_status_.isOperableAndSolvable();
 }
 
+bool WellInterfaceGeneric::useVfpExplicit() const
+{
+    const auto& wvfpexp = well_ecl_.getWVFPEXP();
+    return ((wvfpexp.explicit_lookup() && !changedToOpenThisStep())|| operability_status_.use_vfpexplicit);
+}
+
+bool WellInterfaceGeneric::thpLimitViolatedButNotSwitched() const
+{
+    return operability_status_.thp_limit_violated_but_not_switched;
+}
+
 double WellInterfaceGeneric::getALQ(const WellState& well_state) const
 {
     return well_state.getALQ(name());
@@ -419,19 +440,19 @@ void WellInterfaceGeneric::reportWellSwitching(const SingleWellState& ws, Deferr
     if (well_control_log_.empty())
         return;
 
-    std::string msg = "    Well " + name()
-        + " control mode changed from ";
-    for (const std::string& from : well_control_log_) {
-        msg += from + "->";
-    }
+    std::string from = well_control_log_[0];
     std::string to;
     if (isInjector()) {
         to = Well::InjectorCMode2String(ws.injection_cmode);
     } else {
         to = Well::ProducerCMode2String(ws.production_cmode);
     }
-    msg += to;
-    deferred_logger.info(msg);
+    // only report the final switching
+    if (from != to) {
+        std::string msg = "    Well " + name()
+        + " control mode changed from " + from + " to " + to;
+        deferred_logger.info(msg);
+    }
 }
 
 std::optional<double>
@@ -448,11 +469,13 @@ bhpMax(const std::function<double(const double)>& fflo,
     double high = maxPerfPress + 1.0 * unit::barsa;
     double f_low = fflo(low);
     double f_high = fflo(high);
-    deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
-                           "  low = " + std::to_string(low) +
-                           "  high = " + std::to_string(high) +
-                           "  f(low) = " + std::to_string(f_low) +
-                           "  f(high) = " + std::to_string(f_high));
+    if constexpr (extraBhpAtThpLimitProdOutput) {
+        deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
+                              "  low = " + std::to_string(low) +
+                              "  high = " + std::to_string(high) +
+                              "  f(low) = " + std::to_string(f_low) +
+                              "  f(high) = " + std::to_string(f_high));
+    }
     int adjustments = 0;
     const int max_adjustments = 10;
     const double adjust_amount = 5.0 * unit::barsa;
@@ -501,12 +524,14 @@ bhpMax(const std::function<double(const double)>& fflo,
             return std::nullopt;
         }
     }
-    deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
-                          "  low = " + std::to_string(low) +
-                          "  high = " + std::to_string(high) +
-                          "  f(low) = " + std::to_string(f_low) +
-                          "  f(high) = " + std::to_string(f_high) +
-                          "  bhp_max = " + std::to_string(bhp_max));
+    if constexpr (extraBhpAtThpLimitProdOutput) {
+        deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
+                              "  low = " + std::to_string(low) +
+                              "  high = " + std::to_string(high) +
+                              "  f(low) = " + std::to_string(f_low) +
+                              "  f(high) = " + std::to_string(f_high) +
+                              "  bhp_max = " + std::to_string(bhp_max));
+    }
     return bhp_max;
 }
 
@@ -526,11 +551,13 @@ bisectBracket(const std::function<double(const double)>& eq,
     double eq_high = eq(high);
     double eq_low = eq(low);
     const double eq_bhplimit = eq_low;
-    deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
-                          "  low = " + std::to_string(low) +
-                          "  high = " + std::to_string(high) +
-                          "  eq(low) = " + std::to_string(eq_low) +
-                          "  eq(high) = " + std::to_string(eq_high));
+    if constexpr (extraBhpAtThpLimitProdOutput) {
+        deferred_logger.debug("computeBhpAtThpLimitProd(): well = " + this->name() +
+                              "  low = " + std::to_string(low) +
+                              "  high = " + std::to_string(high) +
+                              "  eq(low) = " + std::to_string(eq_low) +
+                              "  eq(high) = " + std::to_string(eq_high));
+    }
     if (eq_low * eq_high > 0.0) {
         // Failed to bracket the zero.
         // If this is due to having two solutions, bisect until bracketed.
@@ -565,9 +592,9 @@ bisectBracket(const std::function<double(const double)>& eq,
             }
         } else { // eq_low * eq_high > 0.0
             // Still failed bracketing!
-            const double limit = 3.0 * unit::barsa;
+            const double limit = 0.1 * unit::barsa;
             if (std::min(abs_low, abs_high) < limit) {
-                // Return the least bad solution if less off than 3 bar.
+                // Return the least bad solution if less off than 0.1 bar.
                 deferred_logger.warning("FAILED_ROBUST_BHP_THP_SOLVE_BRACKETING_FAILURE",
                                         "Robust bhp(thp) not solved precisely for well " + this->name());
                 approximate_solution = abs_low < abs_high ? low : high;
@@ -651,10 +678,14 @@ computeBhpAtThpLimitProdCommon(const std::function<std::vector<double>(const dou
     const double vfp_ref_depth = table.getDatumDepth();
     const double thp_limit = this->getTHPConstraint(summary_state);
     const double dp = wellhelpers::computeHydrostaticCorrection(this->refDepth(), vfp_ref_depth, rho, this->gravity());
+
     auto fbhp = [this, &controls, thp_limit, dp, alq_value](const std::vector<double>& rates) {
         assert(rates.size() == 3);
+        const auto& wfr =  this->vfpProperties()->getExplicitWFR(controls.vfp_table_number, this->indexOfWell());
+        const auto& gfr = this->vfpProperties()->getExplicitGFR(controls.vfp_table_number, this->indexOfWell());
+        const bool use_vfpexp = this->useVfpExplicit();
         return this->vfpProperties()->getProd()
-        ->bhp(controls.vfp_table_number, rates[Water], rates[Oil], rates[Gas], thp_limit, alq_value) - dp;
+        ->bhp(controls.vfp_table_number, rates[Water], rates[Oil], rates[Gas], thp_limit, alq_value, wfr, gfr, use_vfpexp) - dp;
     };
 
     // Make the flo() function.
