@@ -73,11 +73,13 @@ namespace Opm
          const std::vector<double>& /* depth_arg */,
          const double gravity_arg,
          const int /* num_cells */,
-         const std::vector< Scalar >& B_avg)
+         const std::vector< Scalar >& B_avg,
+         const bool changed_to_open_this_step)
     {
         this->phase_usage_ = phase_usage_arg;
         this->gravity_ = gravity_arg;
         B_avg_ = B_avg;
+        this->changed_to_open_this_step_ = changed_to_open_this_step;
     }
 
 
@@ -317,6 +319,10 @@ namespace Opm
         updatePrimaryVariables(well_state_copy, deferred_logger);
         initPrimaryVariablesEvaluation();
 
+        if (this->isProducer()) {
+            gliftBeginTimeStepWellTestUpdateALQ(simulator, well_state_copy, deferred_logger);
+        }
+
         WellTestState welltest_state_temp;
 
         bool testWell = true;
@@ -332,6 +338,7 @@ namespace Opm
                 return;
             }
 
+
             updateWellOperability(simulator, well_state_copy, deferred_logger);
             if ( !this->isOperableAndSolvable() ) {
                 const auto msg = fmt::format("WTEST: Well {} is not operable (physical)", this->name());
@@ -452,7 +459,39 @@ namespace Opm
         // keep a copy of the original well state
         const WellState well_state0 = well_state;
         const double dt = ebosSimulator.timeStepSize();
-        const bool converged = iterateWellEquations(ebosSimulator, dt, well_state, group_state, deferred_logger);
+        bool converged = iterateWellEquations(ebosSimulator, dt, well_state, group_state, deferred_logger);
+
+        // Newly opened wells with THP control sometimes struggles to
+        // converge due to bad initial guess. Or due to the simple fact
+        // that the well needs to change to another control.
+        // We therefore try to solve the well with BHP control to get
+        // an better initial guess.
+        // If the well is supposed to operate under THP control
+        // "updateWellControl" will switch it back to THP later.
+        if (!converged) {
+            auto& ws = well_state.well(this->indexOfWell());
+            bool thp_control = false;
+            if (this->well_ecl_.isInjector()) {
+                thp_control = ws.injection_cmode == Well::InjectorCMode::THP;
+                if (thp_control) {
+                    ws.injection_cmode = Well::InjectorCMode::BHP;
+                    this->well_control_log_.push_back(Well::InjectorCMode2String(Well::InjectorCMode::THP));
+                }
+            } else {
+                thp_control = ws.production_cmode == Well::ProducerCMode::THP;
+                if (thp_control) {
+                    ws.production_cmode = Well::ProducerCMode::BHP;
+                    this->well_control_log_.push_back(Well::ProducerCMode2String(Well::ProducerCMode::THP));
+                }
+            }
+            if (thp_control) {
+                const std::string msg = std::string("The newly opened well ") + this->name()
+                    + std::string(" with THP control did not converge during inner iterations, we try again with bhp control");
+                deferred_logger.debug(msg);
+                converged = this->iterateWellEquations(ebosSimulator, dt, well_state, group_state, deferred_logger);
+            }
+        }
+
         if (!converged) {
             const int max_iter = param_.max_welleq_iter_;
             deferred_logger.debug("Compute initial well solution for well " + this->name() + ". Failed to converge in "
@@ -531,7 +570,33 @@ namespace Opm
         assembleWellEqWithoutIteration(ebosSimulator, dt, inj_controls, prod_controls, well_state, group_state, deferred_logger);
     }
 
-
+    template<typename TypeTag>
+    bool
+    WellInterface<TypeTag>::isPressureControlled(const WellState& well_state) const
+    {
+         bool thp_controlled_well = false;
+        bool bhp_controlled_well = false;
+        const auto& ws = well_state.well(this->index_of_well_);
+        if (this->isInjector()) {
+            const Well::InjectorCMode& current = ws.injection_cmode;
+            if (current == Well::InjectorCMode::THP) {
+                thp_controlled_well = true;
+            }
+            if (current == Well::InjectorCMode::BHP) {
+                bhp_controlled_well = true;
+            }
+        } else {
+            const Well::ProducerCMode& current = ws.production_cmode;
+            if (current == Well::ProducerCMode::THP) {
+                thp_controlled_well = true;
+            }
+            if (current == Well::ProducerCMode::BHP) {
+                bhp_controlled_well = true;
+            }
+        }
+        bool ispressureControlled =  (bhp_controlled_well || thp_controlled_well);
+        return ispressureControlled;
+    }
 
     template<typename TypeTag>
     void
@@ -584,9 +649,57 @@ namespace Opm
         }
 
         updateWellOperability(ebos_simulator, well_state, deferred_logger);
+        if (!this->operability_status_.isOperableAndSolvable()) {
+            this->operability_status_.use_vfpexplicit = true;
+            deferred_logger.debug("EXPLICIT_LOOKUP_VFP",
+                                "well not operable, trying with explicit vfp lookup: " + this->name());
+            updateWellOperability(ebos_simulator, well_state, deferred_logger);
+        }
     }
 
-
+    template<typename TypeTag>
+    void
+    WellInterface<TypeTag>::
+    gliftBeginTimeStepWellTestUpdateALQ(const Simulator& ebos_simulator,
+                          WellState& well_state,
+                          DeferredLogger& deferred_logger)
+    {
+        const auto& summary_state = ebos_simulator.vanguard().summaryState();
+        const auto& well_name = this->name();
+        if (!this->wellHasTHPConstraints(summary_state)) {
+            const std::string msg = fmt::format("GLIFT WTEST: Well {} does not have THP constraints", well_name);
+            deferred_logger.info(msg);
+            return;
+        }
+        const auto& well_ecl = this->wellEcl();
+        const auto& schedule = ebos_simulator.vanguard().schedule();
+        auto report_step_idx = ebos_simulator.episodeIndex();
+        const auto& glo = schedule.glo(report_step_idx);
+        if (!glo.has_well(well_name)) {
+            const std::string msg = fmt::format(
+                "GLIFT WTEST: Well {} : Gas Lift not activated: "
+                "WLIFTOPT is probably missing. Skipping.", well_name);
+            deferred_logger.info(msg);
+            return;
+        }
+        const auto& gl_well = glo.well(well_name);
+        auto& max_alq_optional = gl_well.max_rate();
+        double max_alq;
+        if (max_alq_optional) {
+            max_alq = *max_alq_optional;
+        }
+        else {
+            const auto& controls = well_ecl.productionControls(summary_state);
+            const auto& table = this->vfpProperties()->getProd()->getTable(controls.vfp_table_number);
+            const auto& alq_values = table.getALQAxis();
+            max_alq = alq_values.back();
+        }
+        well_state.setALQ(well_name, max_alq);
+        const std::string msg = fmt::format(
+            "GLIFT WTEST: Well {} : Setting ALQ to max value: {}",
+            well_name, max_alq);
+        deferred_logger.info(msg);
+    }
 
     template<typename TypeTag>
     void
@@ -597,20 +710,20 @@ namespace Opm
     {
         this->operability_status_.resetOperability();
 
-        bool thp_controled = this->isInjector() ? well_state.well(this->index_of_well_).injection_cmode == Well::InjectorCMode::THP:
+        bool thp_controlled = this->isInjector() ? well_state.well(this->index_of_well_).injection_cmode == Well::InjectorCMode::THP:
                                               well_state.well(this->index_of_well_).production_cmode == Well::ProducerCMode::THP;
-        bool bhp_controled = this->isInjector() ? well_state.well(this->index_of_well_).injection_cmode == Well::InjectorCMode::BHP:
+        bool bhp_controlled = this->isInjector() ? well_state.well(this->index_of_well_).injection_cmode == Well::InjectorCMode::BHP:
                                               well_state.well(this->index_of_well_).production_cmode == Well::ProducerCMode::BHP;
 
         // Operability checking is not free
         // Only check wells under BHP and THP control
-        bool check_thp = thp_controled || this->operability_status_.thp_limit_violated_but_not_switched;
-        if (check_thp || bhp_controled) {
+        bool check_thp = thp_controlled || this->operability_status_.thp_limit_violated_but_not_switched;
+        if (check_thp || bhp_controlled) {
             updateIPR(ebos_simulator, deferred_logger);
             checkOperabilityUnderBHPLimit(well_state, ebos_simulator, deferred_logger);
         }
         // we do some extra checking for wells under THP control.
-        if (thp_controled) {
+        if (check_thp) {
             checkOperabilityUnderTHPLimit(ebos_simulator, well_state, deferred_logger);
         }
     }
@@ -673,7 +786,16 @@ namespace Opm
             switch(current) {
             case Well::InjectorCMode::RATE:
             {
-                ws.surface_rates[phasePos] = controls.surface_rate;
+                ws.surface_rates[phasePos] = (1.0 - this->rsRvInj()) * controls.surface_rate;
+                if(this->rsRvInj() > 0) {
+                    if (injectorType == InjectorType::OIL && FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)) {
+                        ws.surface_rates[pu.phase_pos[BlackoilPhases::Vapour]] = controls.surface_rate * this->rsRvInj();
+                    } else if (injectorType == InjectorType::GAS && FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
+                        ws.surface_rates[pu.phase_pos[BlackoilPhases::Liquid]] = controls.surface_rate * this->rsRvInj();
+                    } else {
+                        OPM_DEFLOG_THROW(std::runtime_error, "Expected OIL or GAS as type for injectors when RS/RV (item 10) is non-zero "  + this->name(), deferred_logger );
+                    }
+                }
                 break;
             }
 
@@ -691,6 +813,7 @@ namespace Opm
                 auto rates = ws.surface_rates;
                 double bhp = this->calculateBhpFromThp(well_state, rates, well, summaryState, this->getRefDensity(), deferred_logger);
                 ws.bhp = bhp;
+                ws.thp = this->getTHPConstraint(summaryState);
 
                 // if the total rates are negative or zero
                 // we try to provide a better intial well rate
@@ -740,6 +863,13 @@ namespace Opm
             }
 
             }
+            // for wells with zero injection rate, if we assign exactly zero rate,
+            // we will have to assume some trivial composition in the wellbore.
+            // here, we use some small value (about 0.01 m^3/day ~= 1.e-7) to initialize
+            // the zero rate target, then we can use to retain the composition information
+            // within the wellbore from the previous result, and hopefully it is a good
+            // initial guess for the zero rate target.
+            ws.surface_rates[phasePos] = std::max(1.e-7, ws.surface_rates[phasePos]);
         }
         //Producer
         else
@@ -908,6 +1038,7 @@ namespace Opm
                 this->adaptRatesForVFP(rates);
                 double bhp = this->calculateBhpFromThp(well_state, rates, well, summaryState, this->getRefDensity(), deferred_logger);
                 ws.bhp = bhp;
+                ws.thp = this->getTHPConstraint(summaryState);
 
                 // if the total rates are negative or zero
                 // we try to provide a better intial well rate
@@ -937,6 +1068,9 @@ namespace Opm
                     for (int p = 0; p<np; ++p) {
                         ws.surface_rates[p] *= scale;
                     }
+                    ws.trivial_target = false;
+                } else {
+                    ws.trivial_target = true;
                 }
                 break;
             }
@@ -1008,8 +1142,9 @@ namespace Opm
         // if more than one nonzero rate.
         auto& ws = well_state.well(this->index_of_well_);
         int nonzero_rate_index = -1;
+        const double floating_point_error_epsilon = 1e-14;
         for (int p = 0; p < this->number_of_phases_; ++p) {
-            if (ws.surface_rates[p] != 0.0) {
+            if (std::abs(ws.surface_rates[p]) > floating_point_error_epsilon) {
                 if (nonzero_rate_index == -1) {
                     nonzero_rate_index = p;
                 } else {
