@@ -26,6 +26,8 @@
 #include <opm/input/eclipse/Schedule/Schedule.hpp>
 #include <opm/simulators/wells/ParallelWellInfo.hpp>
 
+#include <opm/simulators/utils/ParallelCommunication.hpp>
+
 #include <algorithm>
 #include <cassert>
 #include <numeric>
@@ -63,9 +65,10 @@ void WellState::initSingleProducer(const Well& well,
     const double temp = 273.15 + 15.56;
 
     auto& ws = this->wells_.add(well.name(), SingleWellState{well.name(), well_info, true, pressure_first_connection, well_perf_data, pu, temp});
-    if ( ws.perf_data.empty())
-        return;
 
+    // the rest of the code needs to executed even if ws.perf_data is empty
+    // as this does not say anything for the whole well if it is distributed.
+    // Hence never ever return here!
     if (well.getStatus() == Well::Status::OPEN) {
         ws.status = Well::Status::OPEN;
     }
@@ -85,9 +88,10 @@ void WellState::initSingleInjector(const Well& well,
     const double temp = inj_controls.temperature;
 
     auto& ws = this->wells_.add(well.name(), SingleWellState{well.name(), well_info, false, pressure_first_connection, well_perf_data, pu, temp});
-    if ( ws.perf_data.empty())
-        return;
 
+    // the rest of the code needs to executed even if ws.perf_data is empty
+    // as this does not say anything for the whole well if it is distributed.
+    // Hence never ever return here!
     if (well.getStatus() == Well::Status::OPEN) {
         ws.status = Well::Status::OPEN;
     }
@@ -144,11 +148,6 @@ void WellState::init(const std::vector<double>& cellPressures,
     const auto& pu = this->phaseUsage();
     const int np = pu.num_phases;
 
-    int nperf = 0;
-    for (const auto& wpd : well_perf_data) {
-        nperf += wpd.size();
-    }
-
     {
         const auto& wg_events = schedule[report_step].wellgroup_events();
         for (const auto& ecl_well : wells_ecl) {
@@ -407,6 +406,7 @@ WellState::report(const int* globalCellIdxMap,
 
         well.rates.set(rt::dissolved_gas, ws.dissolved_gas_rate);
         well.rates.set(rt::vaporized_oil, ws.vaporized_oil_rate);
+        well.rates.set(rt::vaporized_water, ws.vaporized_wat_rate);
 
         {
             auto& curr = well.current_control;
@@ -822,8 +822,8 @@ WellState::parallelWellInfo(std::size_t well_index) const
     return ws.parallel_info;
 }
 
-template void WellState::updateGlobalIsGrup<ParallelWellInfo::Communication>(const ParallelWellInfo::Communication& comm);
-template void WellState::communicateGroupRates<ParallelWellInfo::Communication>(const ParallelWellInfo::Communication& comm);
+template void WellState::updateGlobalIsGrup<Parallel::Communication>(const Parallel::Communication& comm);
+template void WellState::communicateGroupRates<Parallel::Communication>(const Parallel::Communication& comm);
 } // namespace Opm
 
 
