@@ -26,16 +26,19 @@
 
 #include <ebos/equil/equilibrationhelpers.hh>
 #include <ebos/eclproblem.hh>
-#include <opm/models/utils/start.hh>
-
-#include <opm/simulators/wells/BlackoilWellModel.hpp>
-#include <opm/simulators/flow/BlackoilModelParametersEbos.hpp>
+#include <ebos/eclgenericvanguard.hh>
 
 #include <opm/grid/UnstructuredGrid.h>
 #include <opm/grid/GridManager.hpp>
+#include <opm/grid/cpgrid/GridHelpers.hpp>
 
 #include <opm/input/eclipse/Units/Units.hpp>
 
+#include <opm/models/utils/start.hh>
+
+#include <opm/simulators/flow/BlackoilModelParametersEbos.hpp>
+#include <opm/simulators/wells/BlackoilWellModel.hpp>
+
 #if HAVE_DUNE_FEM
 #include <dune/fem/misc/mpimanager.hh>
 #else
@@ -78,13 +81,27 @@ namespace TTag {
 struct TestEquilTypeTag {
     using InheritsFrom = std::tuple<FlowModelParameters, EclBaseProblem, BlackOilModel>;
 };
+struct TestEquilVapwatTypeTag {
+    using InheritsFrom = std::tuple<FlowModelParameters, EclBaseProblem, BlackOilModel>;
+};
 }
 
 template<class TypeTag>
 struct EclWellModel<TypeTag, TTag::TestEquilTypeTag> {
     using type = BlackoilWellModel<TypeTag>;
 };
-
+template<class TypeTag>
+struct EnableEvaporation<TypeTag, TTag::TestEquilTypeTag> {
+    static constexpr bool value = true;
+};
+template<class TypeTag>
+struct EclWellModel<TypeTag, TTag::TestEquilVapwatTypeTag> {
+    using type = BlackoilWellModel<TypeTag>;
+};
+template<class TypeTag>
+struct EnableEvaporation<TypeTag, TTag::TestEquilVapwatTypeTag> {
+    static constexpr bool value = true;
+};
 } // namespace Opm::Properties
 
 template <class TypeTag>
@@ -103,7 +120,9 @@ initSimulator(const char *filename)
 
     Opm::setupParameters_<TypeTag>(/*argc=*/sizeof(argv)/sizeof(argv[0]), argv, /*registerParams=*/false);
 
-    return std::unique_ptr<Simulator>(new Simulator);
+    Opm::EclGenericVanguard::readDeck(filename);
+
+    return std::make_unique<Simulator>();
 }
 
 template <class GridView>
@@ -195,7 +214,7 @@ static Opm::EquilRecord mkEquilRecord( double datd, double datp,
                                        double zwoc, double pcow_woc,
                                        double zgoc, double pcgo_goc )
 {
-    return Opm::EquilRecord( datd, datp, zwoc, pcow_woc, zgoc, pcgo_goc, true, true, 0);
+    return Opm::EquilRecord( datd, datp, zwoc, pcow_woc, zgoc, pcgo_goc, true, true, 0, true);
 }
 
 template <typename Simulator>
@@ -216,7 +235,6 @@ struct EquilFixture {
         Dune::MPIHelper::instance(argc, argv);
 #endif
         Opm::EclGenericVanguard::setCommunication(std::make_unique<Opm::Parallel::Communication>());
-        using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
         Opm::BlackoilModelParametersEbos<TypeTag>::registerParameters();
         Opm::Parameters::registerParam<TypeTag, bool>("EnableTerminalOutput",
                                                       "EnableTerminalOutput",
@@ -224,6 +242,18 @@ struct EquilFixture {
                                                       "Dummy added for the well model to compile.");
         Opm::registerAllParameters_<TypeTag>();
     }
+
+    using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
+    using FluidSystem = Opm::GetPropType<TypeTag, Opm::Properties::FluidSystem>;
+    using Grid = Opm::GetPropType<TypeTag, Opm::Properties::Grid>;
+    using GridView = Opm::GetPropType<TypeTag, Opm::Properties::GridView>;
+    using ElementMapper = Opm::GetPropType<TypeTag, Opm::Properties::ElementMapper>;
+    using CartesianIndexMapper = Dune::CartesianIndexMapper<Grid>;
+    using Initializer = Opm::EQUIL::DeckDependent::InitialStateComputer<FluidSystem,
+                                                                        Grid,
+                                                                        GridView,
+                                                                        ElementMapper,
+                                                                        CartesianIndexMapper>;
 };
 
 }
@@ -248,6 +278,7 @@ BOOST_AUTO_TEST_CASE(PhasePressure)
         record,
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0
     };
@@ -304,24 +335,28 @@ BOOST_AUTO_TEST_CASE(CellSubset)
         Opm::EQUIL::EquilReg(record[0],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[0],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[1],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[1],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
     };
@@ -407,24 +442,28 @@ BOOST_AUTO_TEST_CASE(RegMapping)
         Opm::EQUIL::EquilReg(record[0],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[0],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[1],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         ,
         Opm::EQUIL::EquilReg(record[1],
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
+        std::make_shared<Opm::EQUIL::Miscibility::NoMixing>(),
         trivialSaltVdTable,
         0)
         };
@@ -497,10 +536,11 @@ BOOST_AUTO_TEST_CASE(DeckAllDead)
     const auto& eclipseState = simulator->vanguard().eclState();
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
-
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 10.0);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                    eclipseState, 
+                                    simulator->vanguard().grid(),
+                                    simulator->vanguard().gridView(),
+                                    simulator->vanguard().cartesianMapper(), 10.0);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -521,8 +561,6 @@ BOOST_AUTO_TEST_CASE(CapillaryInversion)
     // Test setup.
     using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
     using FluidSystem = Opm::GetPropType<TypeTag, Opm::Properties::FluidSystem>;
-    using MaterialLaw = Opm::GetPropType<TypeTag, Opm::Properties::MaterialLaw>;
-    using MaterialLawManager = typename Opm::GetProp<TypeTag, Opm::Properties::MaterialLaw>::EclMaterialLawManager;
 
     auto simulator = initSimulator<TypeTag>("equil_capillary.DATA");
 
@@ -536,7 +574,7 @@ BOOST_AUTO_TEST_CASE(CapillaryInversion)
         const std::vector<double> s = { 0.2, 0.2, 0.2, 0.466666666666, 0.733333333333, 1.0, 1.0, 1.0, 1.0 };
         BOOST_REQUIRE_EQUAL(pc.size(), s.size());
         for (size_t i = 0; i < pc.size(); ++i) {
-            const double s_computed = Opm::EQUIL::satFromPc<FluidSystem, MaterialLaw, MaterialLawManager>(*simulator->problem().materialLawManager(), phase, cell, pc[i], increasing);
+            const double s_computed = Opm::EQUIL::satFromPc<FluidSystem>(*simulator->problem().materialLawManager(), phase, cell, pc[i], increasing);
             BOOST_CHECK_CLOSE(s_computed, s[i], reltol);
         }
     }
@@ -549,7 +587,7 @@ BOOST_AUTO_TEST_CASE(CapillaryInversion)
         const std::vector<double> s = { 0.8, 0.8, 0.8, 0.533333333333, 0.266666666666, 0.0, 0.0, 0.0, 0.0 };
         BOOST_REQUIRE_EQUAL(pc.size(), s.size());
         for (size_t i = 0; i < pc.size(); ++i) {
-            const double s_computed = Opm::EQUIL::satFromPc<FluidSystem, MaterialLaw, MaterialLawManager>(*simulator->problem().materialLawManager(), phase, cell, pc[i], increasing);
+            const double s_computed = Opm::EQUIL::satFromPc<FluidSystem>(*simulator->problem().materialLawManager(), phase, cell, pc[i], increasing);
             BOOST_CHECK_CLOSE(s_computed, s[i], reltol);
         }
     }
@@ -562,7 +600,7 @@ BOOST_AUTO_TEST_CASE(CapillaryInversion)
         const std::vector<double> s = { 0.2, 0.333333333333, 0.6, 0.866666666666, 1.0 };
         BOOST_REQUIRE_EQUAL(pc.size(), s.size());
         for (size_t i = 0; i < pc.size(); ++i) {
-            const double s_computed = Opm::EQUIL::satFromSumOfPcs<FluidSystem, MaterialLaw, MaterialLawManager>(*simulator->problem().materialLawManager(), water, gas, cell, pc[i]);
+            const double s_computed = Opm::EQUIL::satFromSumOfPcs<FluidSystem>(*simulator->problem().materialLawManager(), water, gas, cell, pc[i]);
             BOOST_CHECK_CLOSE(s_computed, s[i], reltol);
         }
     }
@@ -577,9 +615,11 @@ BOOST_AUTO_TEST_CASE(DeckWithCapillary)
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
 
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 10.0);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState, 
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 10.0);
 
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
@@ -617,9 +657,11 @@ BOOST_AUTO_TEST_CASE(DeckWithCapillaryOverlap)
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
 
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 9.80665);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -678,9 +720,11 @@ BOOST_AUTO_TEST_CASE(DeckWithLiveOil)
     const UnstructuredGrid& grid = *(gm.c_grid());
 
     // Initialize the fluid system
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 9.80665);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -746,18 +790,20 @@ BOOST_AUTO_TEST_CASE(DeckWithLiveOil)
     }
 }
 
-BOOST_AUTO_TEST_CASE(DeckWithLiveGas)
+BOOST_AUTO_TEST_CASE(DeckWithWetGas)
 {
     using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
     using FluidSystem = Opm::GetPropType<TypeTag, Opm::Properties::FluidSystem>;
-    auto simulator = initSimulator<TypeTag>("equil_livegas.DATA");
+    auto simulator = initSimulator<TypeTag>("equil_wetgas.DATA");
     const auto& eclipseState = simulator->vanguard().eclState();
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
 
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 9.80665);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -826,6 +872,67 @@ BOOST_AUTO_TEST_CASE(DeckWithLiveGas)
     }
 }
 
+BOOST_AUTO_TEST_CASE(DeckWithHumidWetGas)
+{
+    using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
+    using FluidSystem = Opm::GetPropType<TypeTag, Opm::Properties::FluidSystem>;
+    FluidSystem::setEnableVaporizedWater(true);
+    auto simulator = initSimulator<TypeTag>("equil_humidwetgas.DATA");
+    const auto& eclipseState = simulator->vanguard().eclState();
+    Opm::GridManager gm(eclipseState.getInputGrid());
+    const UnstructuredGrid& grid = *(gm.c_grid());
+
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
+    const auto& pressures = comp.press();
+    BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
+    BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
+
+    const int first = 0, last = grid.number_of_cells - 1;
+    const double reltol = 1.0e-1;
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::waterPhaseIdx][first], 1.480599988e7, reltol);  
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::waterPhaseIdx][last],  1.549297524e7, reltol);
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::oilPhaseIdx][first], 1.489599988e7, reltol);
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::oilPhaseIdx][last],  1.550297524e7, reltol);
+
+    const auto& sats = comp.saturation();
+    std::vector<double> s_opm[3]; 
+    s_opm[FluidSystem::waterPhaseIdx] = { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.35838026, 0.64069098, 0.9154626,    1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
+    s_opm[FluidSystem::oilPhaseIdx] =   { 0,   0,   0,   0,   0,   0,   0,   0,          0,      0.02738364, 0,  0, 0, 0, 0, 0, 0, 0, 0, 0 };
+    s_opm[FluidSystem::gasPhaseIdx] =   { 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.64161973, 0.359309012, 0.057153701, 0,       0, 0, 0, 0, 0, 0, 0, 0, 0 };
+    for (int phase = 0; phase < 3; ++phase) {
+        BOOST_REQUIRE_EQUAL(sats[phase].size(), s_opm[phase].size());
+        for (size_t i = 0; i < s_opm[phase].size(); ++i) {
+            BOOST_CHECK_CLOSE(sats[phase][i], s_opm[phase][i], 100.*reltol);
+        }
+        std::cout << std::endl;
+    }
+
+    const auto& rv = comp.rv();
+    const std::vector<double> rv_opm { 
+        0.00024837999651755729, 0.00024869285236692635, 0.00024900604366769004, 0.00024931957094322978,
+        0.00024963343471801471, 0.00024994763551760586, 0.00025026217386865733, 0.00025057705029892072,
+        0.00025089226533724643, 0.00025120780158539152, 0.00025105, 0.00025105,
+        0.00025105, 0.00025105, 0.00025105, 0.00025105,
+        0.00025105, 0.00025105, 0.00025105, 0.00025105};
+
+    const auto& rvw = comp.rvw();
+    const std::vector<double> rvw_opm {  
+        0.00024837999651755729, 0.00024869285236692635, 0.00024900604366769004, 0.00024931957094322978,
+        0.00024963343471801471, 0.00024994763551760586, 0.00025026217386865733, 0.00025057705029892072,
+        0.00025089226533724643, 0.00025120780158539152, 0.00025236969680655122, 0.00025384953117447344,
+        0.00025532939474124625, 0.00025680928750801825, 0.00025828920947593858, 0.00025976916064615645,
+        0.00026124914101982041, 0.00026272915059807997, 0.0002642091893820838, 0.00026568925737298143};
+
+    for (size_t i = 0; i < rv_opm.size(); ++i) {
+        BOOST_CHECK_CLOSE(rv[i], rv_opm[i], reltol);
+        BOOST_CHECK_CLOSE(rvw[i], rvw_opm[i], reltol);
+    }
+}
+
 BOOST_AUTO_TEST_CASE(DeckWithRSVDAndRVVD)
 {
     using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
@@ -835,9 +942,11 @@ BOOST_AUTO_TEST_CASE(DeckWithRSVDAndRVVD)
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
 
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 9.80665);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -935,9 +1044,11 @@ BOOST_AUTO_TEST_CASE(DeckWithPBVDAndPDVD)
     Opm::GridManager gm(eclipseState.getInputGrid());
     const UnstructuredGrid& grid = *(gm.c_grid());
 
-    Opm::EQUIL::DeckDependent::InitialStateComputer<TypeTag> comp(*simulator->problem().materialLawManager(),
-                                                                  eclipseState, simulator->vanguard().gridView(),
-                                                                  simulator->vanguard().cartesianMapper(), 9.80665);
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
     const auto& pressures = comp.press();
     BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
     BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
@@ -1018,6 +1129,77 @@ BOOST_AUTO_TEST_CASE(DeckWithPBVDAndPDVD)
     }
 }
 
+BOOST_AUTO_TEST_CASE(DeckWithRSVDAndRVVDAndRVWVD)
+{
+    using TypeTag = Opm::Properties::TTag::TestEquilTypeTag;
+    using FluidSystem = Opm::GetPropType<TypeTag, Opm::Properties::FluidSystem>;
+    FluidSystem::setEnableVaporizedWater(true);
+    auto simulator = initSimulator<TypeTag>("equil_rsvd_and_rvvd_and_rvwvd.DATA");
+    const auto& eclipseState = simulator->vanguard().eclState();
+    Opm::GridManager gm(eclipseState.getInputGrid());
+    const UnstructuredGrid& grid = *(gm.c_grid());
+
+    EquilFixture::Initializer comp(*simulator->problem().materialLawManager(),
+                                   eclipseState,
+                                   simulator->vanguard().grid(),
+                                   simulator->vanguard().gridView(),
+                                   simulator->vanguard().cartesianMapper(), 9.80665);
+    const auto& pressures = comp.press();
+    BOOST_REQUIRE_EQUAL(pressures.size(), 3U);
+    BOOST_REQUIRE_EQUAL(int(pressures[0].size()), grid.number_of_cells);
+
+    const int first = 0, last = grid.number_of_cells - 1;
+    const double reltol = 1.0e-4;
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::waterPhaseIdx][first], 1.483359963e7, reltol);  
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::waterPhaseIdx][last],  1.549297524e7, reltol);
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::oilPhaseIdx][first], 1.492359963e7, reltol);
+    BOOST_CHECK_CLOSE(pressures[FluidSystem::oilPhaseIdx][last],  1.550297524e7, reltol);
+
+    const auto& sats = comp.saturation();
+    std::vector<double> s_opm[3]; 
+    s_opm[FluidSystem::waterPhaseIdx] = { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.32527877578987319, 0.62976875867666171, 0.918795223850500588, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
+    s_opm[FluidSystem::oilPhaseIdx]   = { 0,   0,   0,   0,   0,   0,   0,   0,          0, 0.054786199472198836, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+    s_opm[FluidSystem::gasPhaseIdx]   = { 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.674721224210102681, 0.37023124132333829, 0.026418562022795279,    0,          0, 0, 0, 0, 0, 0, 0, 0, 0 };
+
+    for (int phase = 0; phase < 3; ++phase) {
+        BOOST_REQUIRE_EQUAL(sats[phase].size(), s_opm[phase].size());
+        for (size_t i = 0; i < s_opm[phase].size(); ++i) {
+            BOOST_CHECK_CLOSE(sats[phase][i], s_opm[phase][i], reltol);
+        }
+        std::cout << std::endl;
+    }
+
+    const auto& rs = comp.rs();
+    const std::vector<double> rs_opm { // opm
+        74.617998198796087,74.652774471604374, 74.687905898686935, 74.723393674854691,
+        74.759238999357947, 74.795443075905553, 74.832007112684167, 74.892422092838459, 
+        74.986801564438935, 75.088917653469338, 52.5, 57.5, 
+        62.5, 67.5, 72.5, 76.528193441026076,
+        76.774856836636729, 77.021525099679991, 77.268198230347295, 77.514876228830232};
+
+    const auto& rv = comp.rv();
+    const std::vector<double> rv_opm { 
+        2.5000000000000002e-06, 7.5000000000000002e-06, 1.2500000000000001e-05, 1.7500000000000002e-05,
+        2.2500000000000001e-05, 2.7500000000000004e-05, 3.2500000000000004e-05, 3.7500000000000003e-05,
+        4.2500000000000003e-05, 0.00025116322680309166, 5.2500000000000002e-05, 5.7500000000000002e-05,
+        6.2500000000000001e-05, 6.7500000000000001e-05, 7.25e-05, 7.75e-05,
+        8.25e-05, 8.7500000000000013e-05, 9.2500000000000012e-05, 9.7499999999999998e-05};
+
+    const auto& rvw = comp.rvw();
+    const std::vector<double> rvw_opm {  
+        0.00024920798919277656, 0.00024941664682962629, 0.00024962743539212165, 0.00024984036204912818,
+        0.00025005543399614773, 0.00025027265845543336, 0.000250492042676105, 0.00025071359393426718,
+        0.00025093731953312241, 0.00025116322680309166, 0.00025236969680655122, 0.00025384953117447344,
+        0.00025532939474124625, 0.00025680928750801825, 0.00025828920947593858, 0.00025976916064615645,
+        0.00026124914101982041, 0.00026272915059807997, 0.0002642091893820838, 0.00026568925737298143};
+
+    for (size_t i = 0; i < rv_opm.size(); ++i) {
+        BOOST_CHECK_CLOSE(rs[i], rs_opm[i], reltol);
+        BOOST_CHECK_CLOSE(rv[i], rv_opm[i], reltol);
+        BOOST_CHECK_CLOSE(rvw[i], rvw_opm[i], reltol);
+    }
+}
+
 BOOST_AUTO_TEST_CASE(DeckWithSwatinit)
 {
 #if 0
